// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga;

import static org.makagiga.common.UI._;

import java.io.File;
import java.io.IOException;

import org.makagiga.common.*;
import org.makagiga.editors.Editor;
import org.makagiga.editors.EditorExport;
import org.makagiga.editors.EditorPlugin;
import org.makagiga.tree.Tree;
import org.makagiga.tree.TreeList;
import org.makagiga.tree.TreeListItem;

public final class ImportExport {

	// public

	public static void exportFile(final EditorPlugin.FileType type) {
		Editor editor = Tabs.getInstance().getSelectedTab();

		if (editor == null)
			return;

// !!!: Disable "PNG" export menu item [0.9.5]
		if (!(editor instanceof EditorExport)) {
			MMessage.info(null, _("Export is not supported by this file."));

			return;
		}

		MFileChooser fileChooser = createFileChooser(true, type, editor);

		if (!fileChooser.saveDialog())
			return;

		String path = fileChooser.getSelectedPath();

		MLogger.info("EXPORT: Type=\"%s\", Path=\"%s\"", type, path);

		// selected file exists?
		if (FS.exists(path) && !MMessage.confirmFileOverwrite(null, path))
			return;

		EditorExport.class.cast(editor).beginExport();
		try {
			if (type.isInternal() && type.is(".png"))
				UI.createScreenshot(editor.getCore(), "png", path);
			else
				EditorExport.class.cast(editor).exportFile(type, path);
		}
		catch (Exception exception) {
			editor.showError(exception, _("Export failed"));
		}
		finally {
			EditorExport.class.cast(editor).endExport();
		}
	}

	public static void importFile(final EditorPlugin.FileType type) {
		MFileChooser fileChooser = createFileChooser(false, type, null);

		if (!fileChooser.openDialog())
			return;

		String source = fileChooser.getSelectedPath();

		TreeListItem parentItem = Tree.getInstance().getSelectedFolderItem();

		if (parentItem == null)
			return;

		MetaInfo metaInfo = parentItem.getData();

		String name = TK.escapeURL(new File(source).getName());
		String destination = FS.makePath(metaInfo.getPath(), name);

		boolean overwrite = FS.exists(destination);
// !!!: file names in "confirmFileOverwrite" are confusing [0.9.5]
		if (overwrite && !MMessage.confirmFileOverwrite(null, destination))
			return;

		MLogger.info("IMPORT: Type=\"%s\", Source=\"%s\", Destination=\"%s\"", type, source, destination);

		// copy data file!!! share with TreeFS [0.9.3]
		if (FS.copyFile(source, destination)) {
			// copy properties file
			source += ".properties";
			if (FS.exists(source))
				FS.copyFile(source, destination + ".properties");
			// add tree item
			final TreeList treeList = TreeList.getInstance();
			if (overwrite) {
				//!!! [0.9.3]
			}
			else {
				TreeListItem fileItem = TreeListItem.createFileItem(
					destination,
					EditorPlugin.findPluginForType(type.toString())
				);//!!!test [0.9.3]
				treeList.addItem(parentItem, fileItem);
				treeList.focusItem(fileItem);
				// open imported file
				Tabs.getInstance().openEditor(fileItem.getData());
			}
		}
		else {
			MMessage.error(
				null,
				_("Could not copy\n\"{0}\"\nto\n\"{1}\""),
				source,
				destination
			);
		}
	}

	// private

	private static MFileChooser createFileChooser(final boolean export, final EditorPlugin.FileType type, final Editor editor) {
		MFileChooser fileChooser = new MFileChooser(null);
		fileChooser.addFilter(type.toString(), type.getDescription());
		fileChooser.setConfigKey(export ? ("export" + type) : ("import" + type));
		fileChooser.setFileSelectionMode(MFileChooser.FILES_ONLY);
		if (editor != null)
			fileChooser.setSelectedPath(editor.metaInfo.toString() + type);
		if (export) {
			fileChooser.setTitle(_("Export"));
		}
		else {
			fileChooser.setTitle(_("Import"));
// TODO: 2.0: fileChooser.setMultiSelectionEnabled(true);
		}

		return fileChooser;
	}

	// import/export actions

	public static abstract class ImportExportAction extends AppAction {

		// protected

		protected EditorPlugin.FileType type;

		// public

		ImportExportAction(final EditorPlugin.FileType type) {
			super(String.format("%s [%s]", type.getDescription(), type));
			this.type = type;
		}

	}

	public static final class ImportAction extends ImportExportAction {

		// public

		ImportAction(final EditorPlugin.FileType type) {
			super(type);
		}

		public void onAction() {
			importFile(type);
		}

	}

	public static final class ExportAction extends ImportExportAction {

		// public

		ExportAction(final EditorPlugin.FileType type) {
			super(type);
		}

		public void onAction() {
			exportFile(type);
		}

	}

}
