// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga;

// TODO: 2.0: icon theme plugin
// TODO: 2.0: metal theme plugin

import static java.awt.Color.*;
import static java.awt.event.KeyEvent.*;

import static org.makagiga.common.UI._;

import java.awt.CardLayout;
import java.awt.Image;
import java.awt.print.PrinterException;
import javax.swing.BorderFactory;
import javax.swing.ToolTipManager;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.MenuEvent;

import org.makagiga.QuickStart;
import org.makagiga.common.*;
import org.makagiga.console.Console;
import org.makagiga.editors.*;
import org.makagiga.internetsearch.InternetSearch;
import org.makagiga.plugins.PluginInfo;
import org.makagiga.plugins.PluginManager;
import org.makagiga.plugins.PluginMenu;
import org.makagiga.settings.AppSettings;
import org.makagiga.settings.GeneralSettings;
import org.makagiga.settings.PluginSettings;
import org.makagiga.tools.*;
import org.makagiga.tree.Tree;
import org.makagiga.tree.TreeList;

// TODO: 2.0: view modes: minimal (no tool bars, status bar, etc), normal, max
// TODO: 2.0: set active editor content as wallpaper
/** The main window. */
public final class MainWindow extends MMainWindow implements Thread.UncaughtExceptionHandler {

	// private

	private boolean quickStart;
	private Config config;
	private static final int DEFAULT_WIDTH = 800;
	private static final int DEFAULT_HEIGHT = 600;
	private static final int MIN_WIDTH = 320;
	private static final int MIN_HEIGHT = 200;
	private int viewMenuItemCount;
	private MLabel editorStatusText;
	private static MainWindow _instance;
	private MMenu bookmarksMenu;
	private MMenu closeMenu;
	private MMenu editMenu;
	private MMenu exportMenu;
	private MMenu importMenu;
	private MMenu viewMenu;
	private MPanel editorPanel;
	private MPanel mainPanel;
	private MSplitPane mainSplitPane;
// !!!: option: toggle editor tool bar [0.9.5]
	private MToolBar editorToolBar;

	// private actions
	// file
	private PrintAction printAction;
	private QuitAction quitAction;
	// edit
	private FindAction findAction;
	private FindNextAction findNextAction;
	// view
	private ToggleDesignerAction toggleDesignerAction;
	private AppToggleAction showHideSidebarAction;

	// public

	public void backup() {
		new Backup(this);
	}

	public MMenu getBookmarksMenu() { return bookmarksMenu; }

	public synchronized MToolBar getEditorToolBar() { return editorToolBar; }

	public MMenu getExportMenu() { return exportMenu; }

	public MMenu getImportMenu() { return importMenu; }

	public static MainWindow getInstance() {
		if (_instance == null)
			new MainWindow(); // set "_instance" in constructor

		return _instance;
	}

	public void help() {
		final Browser browser = Browser.getInstance();
		// generate help
		String path = FS.makeConfigPath("help", FS.CREATE_DIR);
		try {
			UI.setWaitCursor(browser, true);
			Widget2Help w2h = new Widget2Help();

			w2h.addShortcut(_("Common Keyboard Shortcuts"));
			w2h.addShortcut(_("Move between the Tree and Tabs"), VK_F6);
			w2h.addShortcut(_("Move to splitter bar (press Tab to cancel)"), VK_F8);
			w2h.addShortcut(_("Activate main menu"), VK_F10);
// !!!: context menus [0.9.5]
			w2h.addShortcut(_("Activate context menu"), VK_F10, AppAction.SHIFT_MASK);
			w2h.addShortcut(_("Todo Editor"));
			w2h.addShortcut(_("Edit table cell"), VK_F2);
			w2h.addShortcut(_("Move to previous widget"), VK_TAB, AppAction.CTRL_MASK + AppAction.SHIFT_MASK);

			w2h.setMenuBar(getMMenuBar());
			w2h.setOutputDir(path);
			w2h.setTitle(AppInfo.getTitle());
			w2h.generate();
		}
		finally {
			UI.setWaitCursor(browser, false);
		}
		// load page
		browser.goTo(FS.toURL(path + "/index.html"));
		// show browser
		Sidebar.goTo(Sidebar.Tab.BROWSER);
	}

	public boolean isQuickStart() { return quickStart; }

	public void setQuickStart(final boolean value) {
		quickStart = value;
		CardLayout.class.cast(mainPanel.getLayout()).show(
			mainPanel,
			quickStart ? "QuickStart" : "Editor"
		);
	}

	public synchronized void saveSession(final int flags) {
		MLogger.info("Saving session...");

		// splitters
		config.write("splitter.main.location", mainSplitPane.getDividerLocation());

		// tabs
		MetaInfo metaInfo = Tabs.getInstance().getCurrentMetaInfo();
		config.write("tabs.current", (metaInfo == null) ? "" : metaInfo.getPath());

		// view
		config.write("view.sidebar", Sidebar.getInstance().isVisible());

		// window
		config.write("window.width", getWidth());
		config.write("window.height", getHeight());
		config.write("window.x", getX());
		config.write("window.y", getY());
		config.saveToFile();

// FIXME: 2.0: "setCursor" sometimes does not work (JAVA BUG?)
// MUSTANG: do not try to save if disk is full (JAVA BUG #4057701)
		Tabs.getInstance().saveAllEditors(flags);
		TreeList.getInstance().saveAllItems();
	}

	/**
	 * Sets the editor status text to @p value.
	 * Use @c null value to hide the editor status bar.
	 */
	public void setEditorStatusText(final String value) {
		if (TK.isEmpty(value)) {
			Editor editor = Tabs.getInstance().getSelectedTab();
			String s =
				(editor == null)
				? null
				: editor.info.get().toString();
			editorStatusText.setText(s);
			editorStatusText.setToolTipText(UI.makeHTML(s, UI.ESCAPE));
			editorStatusText.setVisible(s != null);
		}
		else {
			editorStatusText.setText(value);
			editorStatusText.setToolTipText(UI.makeHTML(value, UI.ESCAPE));
			editorStatusText.setVisible(true);
		}
	}

	public void setSidebarVisible(final boolean value) {
		Sidebar.getInstance().setVisible(value);
		mainSplitPane.setOneTouchExpandable(value);
// !!!: hot keys "freeze" after "F9" press [0.9.3]
		if (value) {
			if (mainSplitPane.getDividerLocation() < 10)
				mainSplitPane.setDividerLocation(0.3f);
			mainSplitPane.setDividerSize((Integer)UIManager.get("SplitPane.dividerSize"));
		}
		else {
			mainSplitPane.setDividerSize(0);
		}
		showHideSidebarAction.setSelected(value);
		QuickStart.getInstance().setShowSidebarButtonVisible(!value);
	}

	/**
	 * The default uncaught exception handler.
	 * @param t A thread
	 * @param e A throwable exception
	 */
	public void uncaughtException(Thread t, Throwable e) {
		MLogger.exception(e);
	}

// !!!: plugin [0.9.5]
	public void update() {
		final String HOST = AppInfo.www.get();
		String url = String.format(
			HOST + "/update.php?internalversion=%s&fullversion=%s&os=%s",
			// internal version (e.g. (int)0x000900 == "0.9.0")
			"0x" + AppInfo.internalVersion.toHexString(),
			// nice version (e.g. "0.9.0 Beta")
			TK.escapeURL(AppInfo.niceVersion.get()),
			// internal os name (e.g. "linux")
			TK.escapeURL(OS.getInternalName())
		);
		Sidebar.goTo(Sidebar.Tab.BROWSER);
		Browser.getInstance().goTo(url);
	}

	// protected

	@Override
	protected synchronized boolean onQuit() {
		MLogger.info("Quit...");
		saveSession(Tabs.NORMAL_SAVE);

		// total shut down
		return true;
	}

	@Override
	protected synchronized void onShutDown() {
		MLogger.info("Shut down hook...");
		saveSession(Tabs.QUIET_SAVE);
		forceQuit();
	}

	// private

	private MainWindow() {
		_instance = this;
		Thread.currentThread().setDefaultUncaughtExceptionHandler(this);

		// init splash screen
		Image logo = FS.getImage("makagiga");
		if (!Args.isSet("no-splash") && GeneralSettings.readSplashScreenEnabled()) {
			Image scaledLogo = logo.getScaledInstance(256, 256, Image.SCALE_FAST);
			MSplashScreen.start(WHITE, scaledLogo);
		}

		setIconImage(logo);

		initPlugins();
		initActions();
		initMenuBar();
		initStatusBar();

		// init components

		final Tabs tabs = Tabs.getInstance();

		// main splitter
		mainSplitPane = new MSplitPane(MSplitPane.HORIZONTAL_SPLIT);

		// sidebar
		final Sidebar sidebar = Sidebar.getInstance();
		mainSplitPane.setLeftComponent(sidebar);

		// editor panel
		editorPanel = MPanel.createBorderPanel();
		editorPanel.addCenter(tabs);
// !!!: EditorQuickFind.getInstance() - init on demand
		editorPanel.addSouth(EditorQuickFind.getInstance());
		editorPanel.setVisible(false); // CardLayout will show it

		// main panel
		mainPanel = new MPanel(new CardLayout());
		mainPanel.add(editorPanel, "Editor");
		mainPanel.add(QuickStart.getInstance(), "QuickStart");
		mainSplitPane.setRightComponent(mainPanel);
		addCenter(mainSplitPane);

// TODO: 2.0: invoke "MetaInfo.getToolTipText()" on demand
		MetaInfo.setToolTipsEnabled(true);
		restoreSession();

		MStatusBar.ready();
		MSplashScreen.stop();

		// focus current tab before "setVisible(true)"
		String currentTab = config.read("tabs.current");
		tabs.selectEditorAt(tabs.findEditor(currentTab));
		updateState();
		setVisible(true); // show main window

		Editor editor = Editor.currentEditor.get();
		if (editor == null)
			// set focus to the tree list
			TreeList.getInstance().requestFocusInWindow();
		else
			// set focus to the active editor
			editor.setFocus(); // after "setVisible(true)"

		// add tab change event handler
		tabs.addChangeListener(new ChangeListener() {
			public void stateChanged(final ChangeEvent e) {
				updateState();
			}
		} );

		initAutosave();
		
/* TODO: 2.0: force minimum window size (JAVA BUG #4320050)
		addComponentListener(new ComponentAdapter() {
			@Override
			public void componentResized(final ComponentEvent e) {
				int w = Math.max(DEFAULT_WIDTH / 2, getWidth());
				int h = Math.max(DEFAULT_HEIGHT / 2, getHeight());
				setSize(w, h);
			}
		} );
*/

		// end time demo
		if (Args.isSet("time-demo")) {
			MLogger.info("END TIME DEMO");
			quit();
		}
	}

	private void applySettings() {
		setMinimizeOnClose(GeneralSettings.readMinimizeOnClose());
		ToolTipManager.sharedInstance().setEnabled(GeneralSettings.readToolTipsEnabled());
	}

	private void initPlugins() {
		final PluginManager plugins = PluginManager.getInstance();
		plugins.registerInternal("org/makagiga/editors/image");
		plugins.registerInternal("org/makagiga/editors/text");
		plugins.registerInternal("org/makagiga/editors/todo");
	}

	private void initActions() {
		// file
		printAction = new PrintAction();
		quitAction = new QuitAction();
		// view
		toggleDesignerAction = new ToggleDesignerAction();
		showHideSidebarAction = new ShowHideSidebarAction();
	}

	private void initAutosave() {
		new MTimer(MTimer.Format.MINUTES, 10) {
			public void onTimeout() {
				if (!GeneralSettings.readAutosave())
					return;

				try {
					stop(); // stop autosave
					String s = _("Autosave...");
					MStatusBar.setText(
						s,
						MStatusBar.AUTOHIDE
					);
					MLogger.info(s);
					saveSession(Tabs.QUIET_SAVE);
				}
				finally {
					restart();
					MStatusBar.ready();
				}
			}
		}.start();
	}

// TODO: 2.0: import clipboard content
	private void initImportMenu(final MMenu menu) {
		EditorPlugin.FileType[] types;
		for (PluginInfo i : PluginManager.getInfos(EditorPlugin.TYPE)) {
			types = EditorPlugin.fromInfo(i).importTypes.get();
			if (!TK.isEmpty(types)) {
				MMenu importSubmenu = new MMenu(i.toString(), i.getIcon());
				menu.add(importSubmenu);
				for (EditorPlugin.FileType t : types)
					importSubmenu.add(new ImportExport.ImportAction(t));
			}
		}
	}

	private synchronized void initMenuBar() {
		final Tabs tabs = Tabs.getInstance();

		MMenuBar menuBar = new MMenuBar();

		// file menu
		MMenu fileMenu = new MMenu(_("&File"));

		MMenu treeMenu = new MMenu(_("&Tree"), "tree");
		treeMenu.setHTMLHelp(_("From this menu, you can modify all the files and folders."));
		treeMenu.addMenuListener(new MMenuAdapter() {
			@Override
			public void menuDeselected(final MenuEvent e) {
				MMenu m = (MMenu)e.getSource();
				if (!m.isEmpty())
					m.removeAll();
			}
			@Override
			public void menuSelected(final MenuEvent e) {
				MMenu m = (MMenu)e.getSource();
				if (m.isEmpty())
					Tree.getInstance().createTreeMenu(m, false);
			}
		} );
		fileMenu.add(treeMenu);

		fileMenu.add(new QuickSaveAction());
		fileMenu.add(printAction);
		fileMenu.addSeparator();

		importMenu = new MMenu(_("&Import"));
		importMenu.setHTMLHelp(_("Imports a specified file to the Tree."));
		importMenu.addMenuListener(new MMenuAdapter() {
			@Override
			public void menuSelected(final MenuEvent e) {
				MMenu m = (MMenu)e.getSource();
				if (m.isEmpty())
					initImportMenu(m);
			}
		} );
		fileMenu.add(importMenu);

		exportMenu = new MMenu(_("&Export"));
		exportMenu.setHTMLHelp(_("Exports the content of the active editor.<br>For example, you can save a \"todo\" table in the HTML format."));
		fileMenu.add(exportMenu);

		fileMenu.addSeparator();
		fileMenu.add(tabs.getLocateAction());
		closeMenu = tabs.createCloseMenu();
		fileMenu.add(closeMenu);
		fileMenu.add(quitAction);
		menuBar.add(fileMenu);

		// edit menu (used by editor plugins)
		editMenu = new MMenu(_("&Edit"));
		menuBar.add(editMenu);

		// view menu (used by editor plugins)
		viewMenu = new MMenu(_("&View"));
		viewMenu.add(toggleDesignerAction);
		viewMenu.addSeparator();
		viewMenu.addCheckBox(showHideSidebarAction);
		viewMenuItemCount = viewMenu.getItemCount();
		menuBar.add(viewMenu);

		// bookmarks menu
		bookmarksMenu = new MMenu(_("&Bookmarks"));
		bookmarksMenu.addMenuListener(new MMenuAdapter() {
			@Override
			public void menuSelected(final MenuEvent e) {
				bookmarksMenu.removeAll();
				// add bookmarks to the menu
				Bookmarks.getInstance().updateMenu((MMenu)e.getSource());
				// no bookmarks?
				if (bookmarksMenu.isEmpty())
					bookmarksMenu.add(_("No Bookmarks")).setEnabled(false);
			}
		} );
		menuBar.add(bookmarksMenu);

		// tools menu
		MMenu toolsMenu = new MMenu(_("&Tools"));
		toolsMenu.add(new SearchAction());
		toolsMenu.add(new InternetSearchAction());
		toolsMenu.addSeparator();
		toolsMenu.add(new BackupAction());
		toolsMenu.add(new UpdateAction());
		toolsMenu.add(new ConsoleAction());
		toolsMenu.addSeparator();
		toolsMenu.add(new ConfigurePluginsAction());
		toolsMenu.add(new ConfigureAction());
		menuBar.add(toolsMenu);

		// help menu
		MMenu helpMenu = new MMenu(_("&Help"));
		helpMenu.add(new HelpAction());
		helpMenu.add(new LiveTutorialAction());
		helpMenu.addSeparator();
		helpMenu.add(new AboutAction());
		menuBar.add(helpMenu);

		setMMenuBar(menuBar);
	}

	private void initStatusBar() {
		editorStatusText = new MLabel();
		editorStatusText.setBorder(BorderFactory.createLoweredBevelBorder());
		editorStatusText.setMnemonicEnabled(false);
		getStatusBar().add(editorStatusText);
	}

	private void restoreSession() {
		config = Config.getGlobal(); // default config
		if (config.isError())
			MLogger.error("Configuration file error: \"%s\"", config.path);

		// "SAFE" MODE

		if (Args.isSet("safe-mode")) {
			setBounds(0, 0, DEFAULT_WIDTH, DEFAULT_HEIGHT);

			return;
		}

		// "UNSAFE" MODE

		applySettings();

		setLocationByPlatform(true); // use workspace size
		setBounds(
			config.readInt("window.x", 0),
			config.readInt("window.y", 0),
			Math.max(MIN_WIDTH, config.readInt("window.width", DEFAULT_WIDTH)),
			Math.max(MIN_HEIGHT, config.readInt("window.height", DEFAULT_HEIGHT))
		);

		// splitter
		mainSplitPane.setDividerLocation(config.readInt("splitter.main.location", -1));

		// view
		setSidebarVisible(config.read("view.sidebar", true));
		showHideSidebarAction.setSelected(Sidebar.getInstance().isVisible());

		// reopen last files, expand tree, etc.
		TreeList.getInstance().restoreAll();
	}

	private synchronized void updateState() {
		Designer designer = null;
		final Sidebar sidebar = Sidebar.getInstance();
		final Tabs tabs = Tabs.getInstance();
		final TreeList treeList = TreeList.getInstance();

		editMenu.removeAll();
		viewMenu.removeFrom(viewMenuItemCount);

		// update tree and tabs
		tabs.updateState();
		// update close menu
		AppAction closeAction = tabs.getCloseAction();
		closeMenu.setEnabled(closeAction.isEnabled());

		treeList.focusItem(tabs.getCurrentMetaInfo());

		// update editor
		Editor editor = tabs.getSelectedTab();
		if (editor != null) {
			// update editor menus
			if (editMenu != null) {
				if (editor instanceof PluginMenu) {
					PluginMenu pluginMenu = (PluginMenu)editor;
					pluginMenu.updateMenu(PluginMenu.EDIT_MENU, editMenu);
				}
			}

			if (viewMenu != null) {
				if (editor instanceof PluginMenu) {
					PluginMenu pluginMenu = (PluginMenu)editor;
					viewMenu.addSeparator();
					pluginMenu.updateMenu(PluginMenu.VIEW_MENU, viewMenu);
				}
			}

			// update editor designer
			if (editor instanceof EditorDesigner) {
				designer = EditorDesigner.class.cast(editor).getEditorDesigner();
				EditorDesigner.class.cast(editor).updateEditorDesigner();
			}
			else {
				designer = null;
			}

			// update editor tool bar
			if (editorToolBar == null) {
				editorToolBar = new MToolBar();
				editorPanel.addNorth(editorToolBar);
			}
			else {
				editorToolBar.removeAll();
			}
			if (editor instanceof PluginMenu) {
				PluginMenu pluginMenu = (PluginMenu)editor;
				pluginMenu.updateToolBar(PluginMenu.EDITOR_TOOL_BAR, editorToolBar);
			}
			editorToolBar.addStretch();
			editorToolBar.add(tabs.getLocateAction());
			editorToolBar.add(tabs.getCloseAction());
			editorToolBar.repaint();

			editorStatusText.setVisible(true);
			setEditorStatusText(editor.getStatusText());

			// update actions
			editor.updateActions();
			if (editor instanceof EditorSearch) {
				editMenu.addSeparator();
				// find
				if (findAction == null)
					findAction = new FindAction();
				findAction.setEnabled(true);
				editMenu.add(findAction);
				// find next
				if (findNextAction == null)
					findNextAction = new FindNextAction();
				findNextAction.setEnabled(true);
				editMenu.add(findNextAction);
			}
			else {
				// find
				if (findAction != null)
					findAction.setEnabled(false);
				// find next
				if (findNextAction != null)
					findNextAction.setEnabled(false);
			}
			printAction.setEnabled(editor instanceof EditorPrint);

			// update components
			if (!(editor instanceof EditorSearch))
				EditorQuickFind.getInstance().setActive(false);
		}
		else {
			editorStatusText.setVisible(false);
			setEditorStatusText(null);

			// update actions
			if (findAction != null)
				findAction.setEnabled(false);
			if (findNextAction != null)
				findNextAction.setEnabled(false);
			printAction.setEnabled(false);

			// update components
			EditorQuickFind.getInstance().setVisible(false);
		}
		editMenu.setEnabled(!editMenu.isEmpty());

		// update sidebar
		if (designer == null) {
			sidebar.setEnabledAt(Sidebar.Tab.DESIGNER.ordinal(), false);
			if (sidebar.isDesignerSelected())
				sidebar.goTo(Sidebar.Tab.TREE);
			toggleDesignerAction.setEnabled(false);
		}
		else {
			sidebar.setComponentAt(Sidebar.Tab.DESIGNER.ordinal(), designer);
			sidebar.setEnabledAt(Sidebar.Tab.DESIGNER.ordinal(), true);
			designer.repaint(); // required
			if (editor != null)
				designer.setReadOnly(editor.metaInfo.get().isLocked());
			toggleDesignerAction.setEnabled(true);
		}

		// update export menu
		exportMenu.removeAll();
		// add built-in "PNG" export
		exportMenu.add(new ImportExport.ExportAction(new EditorPlugin.FileType(".png", _("PNG Image"), true)));
		if (editor == null) {
			exportMenu.setEnabled(false);
		}
		else {
			exportMenu.setEnabled(true);
			// add plugin export types
			EditorPlugin.FileType[] types = EditorPlugin.fromInfo(editor.info.get()).exportTypes.get();
			if (!TK.isEmpty(types)) {
				exportMenu.addSeparator();
				for (EditorPlugin.FileType i : types)
					exportMenu.add(new ImportExport.ExportAction(i));
			}
		}

		updateTitle();
		setQuickStart(tabs.isEmpty());
	}

	private void updateTitle() {
		MetaInfo metaInfo = Tabs.getInstance().getCurrentMetaInfo();
		setTitle(
			(metaInfo == null)
			? AppInfo.fullName.get()
			: (metaInfo.getNicePath() + " - " + AppInfo.fullName)
		);
	}

	// private classes

	private static final class AboutAction extends AppAction {

		// public

		AboutAction() {
			super(_("&About"), "makagiga");
			setHTMLHelp(_("Displays the <a href=\"makagiga:about\">information</a> about this application."));
		}

		public void onAction() {
			about();
		}

	}

	private final class BackupAction extends AppAction {

		// public

		BackupAction() {
			super(_("&Backup..."), "backup");
			setHTMLHelp(_("Use this command to <a href=\"makagiga:backup\">create a backup</a>."));
		}

		public void onAction() {
			backup();
		}

	}

	private final class ConfigureAction extends AppAction {

		// public

		public ConfigureAction() {
			super(_("&Options..."), "ui/configure");
			setHTMLHelp(_("Application settings (autosave, look and feel, browser, etc)."));
		}

		public void onAction() {
			new AppSettings(MainWindow.this);
			applySettings();
		}

	}

	private static final class ConfigurePluginsAction extends AppAction {

		// public

		public ConfigurePluginsAction() {
			super(_("&Plugins"), "plugin");
			setHTMLHelp(_("Configure installed plugins."));
		}

		public void onAction() {
			new PluginSettings(MainWindow.getInstance());
		}

	}

	private static final class ConsoleAction extends AppAction {

		// public

		ConsoleAction() {
			super(_("&Console"), "console", VK_F12, CTRL_MASK + SHIFT_MASK);
			setHTMLHelp(_("Opens the <b>Console</b>."));
		}

		public void onAction() {
			Console.showWindow();
		}

	}

	private static final class FindAction extends AppAction {

		// public

		FindAction() {
			super(_("&Find..."), "find", VK_F, CTRL_MASK);
			setHTMLHelp(_("Search for text in the active editor."));
		}

		public void onAction() {
			EditorQuickFind.getInstance().setActive(true);
		}

	}

	private static final class FindNextAction extends AppAction {

		// public

		FindNextAction() {
			super(_("Find &Next..."), null, VK_F3);
			setHTMLHelp(_("Search for the next matching text in the active editor."));
		}

		public void onAction() {
			EditorQuickFind.getInstance().findNext();
		}

	}

	private final class HelpAction extends AppAction {

		// public

		HelpAction() {
			super(_("&Help"), "help", VK_F1);
			setHTMLHelp(_("Displays the <b>Help</b> in the <b>Browser</b>."));
		}

		public void onAction() {
			setSidebarVisible(true);
			help();
		}

	}

// !!!: icon
	private final class InternetSearchAction extends AppAction {

		// public

		public InternetSearchAction() {
			super(_("&Internet Search..."), null, VK_K, CTRL_MASK);
			setHTMLHelp(_("Internet search.<br><b>Tip:</b> Select text (or item) in the active editor<br>to search for it."));
		}

		public void onAction() {
			Editor editor = Editor.currentEditor.get();
			String defaultText =
				(editor instanceof EditorSearch)
				? EditorSearch.class.cast(editor).getDefaultSearchText()
				: "";
			new InternetSearch(MainWindow.this, defaultText);
		}

	}

	private static final class LiveTutorialAction extends AppAction {

		// public

		public LiveTutorialAction() {
			super(_("Toggle &Live Tutorial"), null, VK_F1, SHIFT_MASK);
			setHTMLHelp(_("Toggles the <b>Live Tutorial</b>."));
		}

		public void onAction() {
			LiveTutorial.toggle();
		}

	}

	private static final class PrintAction extends AppAction {

		// public

		public PrintAction() {
			super(_("&Print..."), "print", VK_P, CTRL_MASK);
			setEnabled(false);
			setHTMLHelp(_("Prints the content of the open file."));
		}

		public void onAction() {
			Editor editor = Tabs.getInstance().getSelectedTab();

			if (!(editor instanceof EditorPrint))
				return;

			if (!MMessage.confirmWithIcon(
				null,
				"print",
				_("Print \"{0}\"?"),
				editor.metaInfo
			))
				return;

			try {
				if (!EditorPrint.class.cast(editor).printDocument())
					MStatusBar.setError(_("Print cancelled"));
			}
			catch (PrinterException exception) {
				MMessage.fileError(
					null,
					_("Print failed"),
					exception.getLocalizedMessage()
				);
			}
		}

	}

	private final class QuickSaveAction extends AppAction {

		// public

		QuickSaveAction() {
			super(_("&Save All"), "saveall", VK_S, CTRL_MASK);
			setHTMLHelp(_("Saves all the modified files.<br>Use the <b>Autosave</b> option to save files automatically every 10 minutes."));
		}

		public void onAction() {
			saveSession(Tabs.NORMAL_SAVE);
		}

	}

	private final class QuitAction extends AppAction {

		// public

		QuitAction() {
			super(_("&Quit"), "quit", VK_Q, CTRL_MASK);
			setHTMLHelp(_("Quits the application.<br>All the modified files are saved automatically."));
		}

		public void onAction() {
			quit();
		}

	}

	private static final class SearchAction extends AppAction {

		// public

		SearchAction() {
			super(_("&Find Files..."), "find", VK_F7);
			setHTMLHelp(_("Search for files."));
		}

		public void onAction() {
			Sidebar.goTo(Sidebar.Tab.SEARCH);
		}

	}

	private final class ShowHideSidebarAction extends AppToggleAction {

		// public

		ShowHideSidebarAction() {
			super(_("&Sidebar"), null, VK_F9);
			setHTMLHelp(_("Toggles the <b>Sidebar</b>."));
		}

		public void onToggle(final boolean isSelected) {
			setSidebarVisible(isSelected);
		}

	}

	private static final class ToggleDesignerAction extends AppAction {

		// public

		public ToggleDesignerAction() {
			super(_("Toggle &Designer"), "designer", VK_F4);
			setHTMLHelp(_("Toggles the <b>Designer</b>."));
		}

		public void onAction() {
			final Sidebar sidebar = Sidebar.getInstance();
			if (sidebar.isDesignerSelected())
				sidebar.selectLastTab();
			else
				Sidebar.goTo(Sidebar.Tab.DESIGNER);
		}

	}

	private final class UpdateAction extends AppAction {

		// public

		UpdateAction() {
			super(_("&Update"), "update");
			setHTMLHelp(_("Launches the Internet <a href=\"makagiga:update\">update</a>."));
		}

		public void onAction() {
			setSidebarVisible(true);
			update();
		}

	}

}
