// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga;

import static java.awt.event.KeyEvent.*;

import static org.makagiga.common.UI._;

import java.awt.Color;
import java.awt.event.MouseEvent;
import java.util.Iterator;

import org.makagiga.common.*;
import org.makagiga.editors.Editor;
import org.makagiga.editors.EditorConfig;
import org.makagiga.editors.EditorDesigner;
import org.makagiga.editors.EditorPlugin;
import org.makagiga.tree.Tree;
import org.makagiga.tree.TreeList;

// MUSTANG: tab "close" icon/button
/** The tabs. */
public final class Tabs extends MTabbedPane<Editor> {

	// public

	// "closeEditorAt" flags
	public static final int REMOVE_TAB = 1 << 0; /**< Remove tab from the MTabbedPane. */
	public static final int SAVE_TO_FILE = 1 << 1; /**< Save editor before remove. */

	// "saveEditor" flags
	public static final int NORMAL_SAVE = 1 << 0; /**< Normal save. */
	public static final int QUIET_SAVE = 1 << 1; /**< Quiet save (no message boxes). */

	// private

	private static Tabs _instance;

	// private actions
	private final CloseAction closeAction = new CloseAction();
	private final CloseAllAction closeAllAction = new CloseAllAction();
	private final CloseAllOthersAction closeAllOthersAction = new CloseAllOthersAction();
	private final LocateAction locateAction = new LocateAction();

	// public

	/**
	 * Closes editor at the specified index.
	 * @param index A tab index
	 * @param flags Flags (@ref REMOVE_TAB, @ref SAVE_TO_FILE)
	 */
	public void closeEditorAt(final int index, final int flags) {
		Editor editor = getTabAt(index);

		if (editor == null)
			return;

		boolean removeTab = (flags & REMOVE_TAB) != 0;
		boolean saveToFile = (flags & SAVE_TO_FILE) != 0;

		if (saveToFile)
			saveEditor(editor, NORMAL_SAVE);

		editor.metaInfo.get().open.no(); // mark as closed

		if (removeTab) {
			removeTabAt(index);
			fireStateChanged();
		}

		editor.destroy();
	}
	
	public MMenu createCloseMenu() {
		MMenu menu = new MMenu(_("&Close"), "ui/close");
		menu.add(closeAction);
		menu.add(closeAllAction);
		menu.add(closeAllOthersAction);
		
		return menu;
	}

	/**
	 * Finds editor with @p metaInfo.
	 * @return An editor index, or -1 if editor could not be found
	 * @see getEditorAt
	 */
	public int findEditor(final MetaInfo metaInfo) {
		return (metaInfo == null) ? -1 : findEditor(metaInfo.getPath());
	}

	/**
	 * Finds editor with @p path.
	 * @return An editor index, or -1 if editor could not be found
	 * @see getEditorAt
	 */
	public int findEditor(final String path) {
		if (TK.isEmpty(path))
			return -1;

		Editor editor;
		MTabIterator i = tabIterator();
		while (i.hasNext()) {
			editor = i.next();

			if ((editor != null) && editor.metaInfo.get().getPath().equals(path))
				return i.getIndex();
		}

		return -1;
	}
	
	public CloseAction getCloseAction() { return closeAction; }

	/**
	 * Returns the active editor's meta info (current tab).
	 */
	public MetaInfo getCurrentMetaInfo() {
		Editor editor = getSelectedTab();

		if (editor == null)
			return null;

		return editor.metaInfo.get();
	}

	/**
	 * Creates and returns the instance of the tabs.
	 */
	public static Tabs getInstance() {
		if (_instance == null)
			new Tabs(); // set "_instance" in constructor

		return _instance;
	}

	/**
	 * Returns the "Locate" action.
	 */
	public LocateAction getLocateAction() { return locateAction; }

	/**
	 * Opens editor with @p metaInfo.
	 * See source for details.
	 * @see saveEditor
	 */
	public void openEditor(final MetaInfo metaInfo) {
		if (metaInfo == null) {
			MLogger.trace();
			MLogger.error("Null meta info");

			return;
		}

// TODO: 2.0: autodownload/suggest a missing plugin
		if (!metaInfo.isSupported()) {
			MMessage.fileError(
				null,
				_("This file is not supported. Plugin missing?"),
				metaInfo.getNicePath()
			);
			metaInfo.open.no();

			return;
		}

		int editorIndex = findEditor(metaInfo); // file already open?
		if (editorIndex != -1) {
			// select already open file
			selectEditorAt(editorIndex);

			return;
		}

		// open new file
		Editor editor;
		try {
			editor = EditorPlugin.createEditorForPath(metaInfo.getPath());
		}
		catch (Exception exception) {
			MLogger.exception(exception);
			MMessage.fileError(
				null,
				_("Could not initialize editor"),
				exception
			);
			metaInfo.open.no();

			return;
		}

		// check editor
		if (editor == null) {
			MMessage.fileError(
				null,
				_("Could not open file"),
				metaInfo.getNicePath()
			);
			metaInfo.open.no();

			return;
		}

		// check editor core
		if (editor.getCore() == null) {
			MMessage.fileError(
				null,
				_("Could not initialize editor"),
				editor.getClass().getName() + ": getCore() == null"
			);
			metaInfo.open.no();

			return;
		}

		editor.metaInfo.set(metaInfo);

		// load file
		MLogger.info("Loading: \"%s\"", metaInfo.getPath());
		MStatusBar.setText(_("Loading..."));
		UI.setWaitCursor(MainWindow.getInstance(), true);
		try {
			editor.loadFromFile();
			metaInfo.setLastOpenTime();
			editor.modified.no();
			editor.setReadOnly(metaInfo.isLocked());
			addTab(null, editor);
			setTabInfo(metaInfo);
			setSelectedComponent(editor);
			editor.setFocus();
			metaInfo.open.yes(); // mark as open
// FIXME: 2.0: NullPointerException after "TreeList.getInstance().updateCellSize()"
// see also: "makagiga.tree.TreeList.java" and its "addKeyListener" (?)
			TreeList.getInstance().updateUI();

			// load config
			if (editor instanceof EditorConfig) {
				EditorConfig.class.cast(editor).loadConfig(
					Config.getGlobal(), // global
					metaInfo.getConfig() // local
				);
			}

			// update designer
			if (editor instanceof EditorDesigner)
				EditorDesigner.class.cast(editor).updateEditorDesigner();

			editor.updateStatusText();
		}
		catch (Exception exception) {
			MLogger.exception(exception);
			metaInfo.open.no(); // mark as closed
			editor.showError(exception, _("Could not open file"));
		}
		finally {
			MStatusBar.ready();
			UI.setWaitCursor(MainWindow.getInstance(), false);
		}
	}

	/**
	 * Saves all open editors (all tabs).
	 */
	public void saveAllEditors(final int flags) {
		MTabIterator i = tabIterator();
		while (i.hasNext())
			saveEditor(getTabAt(i.nextIndex()), flags);
	}

	/**
	 * Saves @p editor.
	 * See source for details.
	 * @see openEditor
	 */
	public void saveEditor(final Editor editor, final int flags) {
		if (editor == null)
			return;

		boolean quiet = (flags & QUIET_SAVE) != 0;

		MetaInfo metaInfo = editor.metaInfo.get();
		// save config
		editor.syncConfig();
		// save file
		if (editor.modified.get() || editor.alwaysSave.get()) {
			MLogger.info("Saving: \"%s\"", metaInfo.getPath());
			try {
				if (!quiet)
					UI.setWaitCursor(MainWindow.getInstance(), true);
				editor.saveToFile();
				editor.modified.no();
			}
			catch (Exception exception) {
				MLogger.exception(exception);
				if (!quiet)
					editor.showError(exception, _("Could not save file"));
			}
			finally {
				if (!quiet) {
					editor.updateStatusText();
					UI.setWaitCursor(MainWindow.getInstance(), false);
				}
			}
		}
	}

	/**
	 * Selects and focuses an editor at @p index.
	 */
	public void selectEditorAt(final int index) {
		Editor editor = getTabAt(index);

		if (editor == null)
			return;

		setSelectedIndex(index);
		editor.setFocus();
	}

	/**
	 * Updates the tab info with @p metaInfo.
	 * Updated tab infos:
	 * - background color
	 * - foreground color
	 * - icon
	 * - tab name
	 * - "[Locked]" text if editor is locked
	 * - tool tip text
	 * - editor color
	 */
	public void setTabInfo(final MetaInfo metaInfo) {
		int i = findEditor(metaInfo);
		if (i != -1) {
			Color color = metaInfo.getColor();
			if (i != getSelectedIndex()) {
				setBackgroundAt(i, color);
				setForegroundAt(i, UI.getXorColor(color));
			}
			setIconAt(i, metaInfo.getIcon());
// TODO: 2.0: option: "setTitleAt(i, metaInfo.getNicePath());"
			if (metaInfo.isLocked())
				setTitleAt(i, metaInfo + " [" + _("Locked") + "]");
			else
				setTitleAt(i, metaInfo.toString());
			setToolTipTextAt(i, metaInfo.getToolTipText());
			Editor editor = getTabAt(i);
			if (editor != null)
				editor.setMainColor(color);
		}
	}

	/**
	 * Updates tabs state (sets active editor, updates actions and colors).
	 */
	public void updateState() {
		Editor editor = getSelectedTab();
		Editor.currentEditor.set(editor);
		boolean isEditor = editor != null;
		locateAction.setEnabled(isEditor);
		closeAction.setEnabled(isEditor);
		closeAllAction.setEnabled(isEditor);
		closeAllOthersAction.setEnabled(isEditor && (getTabCount() > 1));
		updateColors();
	}

	// private

	private Tabs() {
		_instance = this;
		addMouseListener(new MMouseAdapter() {
			@Override
			public void popupTrigger(final MouseEvent e) {
				popupContextMenu(e.getX(), e.getY());
			}
		} );
	}

	/**
	 * Closes the active editor (tab).
	 */
	private void close() {
		closeEditorAt(getSelectedIndex(), REMOVE_TAB | SAVE_TO_FILE);
		if (isEmpty() && Sidebar.getInstance().isDesignerSelected()) {
			Sidebar.goTo(Sidebar.Tab.TREE);
		}
		else {
			if (!(getSelectedTab() instanceof EditorDesigner) && !Sidebar.getInstance().isSearchSelected())
				Sidebar.goTo(Sidebar.Tab.TREE);
		}
		TreeList.getInstance().updateUI();
	}

	/**
	 * Closes all editors (tabs).
	 */
	private void closeAll() {
		if (isEmpty())
			return;

		MTabIterator i = tabIterator();
		while (i.hasNext())
			closeEditorAt(i.nextIndex(), SAVE_TO_FILE);
		removeAll();
		fireStateChanged();
		if (Sidebar.getInstance().isDesignerSelected())
			Sidebar.goTo(Sidebar.Tab.TREE);
		TreeList.getInstance().updateUI();
	}

	private void closeAllOthers() {
		if (getTabCount() < 2)
			return;

		int currentEditor = getSelectedIndex();
		for (int i = getTabCount() - 1; i >= 0; --i) {
			if (i != currentEditor)
				closeEditorAt(i, REMOVE_TAB | SAVE_TO_FILE);
		}
		TreeList.getInstance().updateUI();
	}

	/**
	 * Displays tab context menu at @p x, @p y position.
	 */
	private void popupContextMenu(final int x, final int y) {
		// select tab under mouse
		int i = indexAtLocation(x, y);

		if (i == -1)
			return;

		try {
			setSelectedIndex(i);
		}
		catch (IndexOutOfBoundsException exception) {
			return;
		}
		// create and show popup menu
		createCloseMenu().showPopup(this, x, y);
	}

	/**
	 * Updates tabs colors.
	 */
	private void updateColors() {
		Editor editor;
		int index;
		MetaInfo metaInfo;
		MTabIterator i = tabIterator();
		while (i.hasNext()) {
			editor = i.next();
			index = i.getIndex();
			if (editor != null) {
				if (index == getSelectedIndex()) {
					// set default colors
					setBackgroundAt(index, null);
					setForegroundAt(index, null);
				}
				else {
					// set colors from the meta info
					Color color = editor.metaInfo.get().getColor();
					setBackgroundAt(index, color);
					setForegroundAt(index, UI.getXorColor(color));
				}
			}
		}
	}

	// private classes

	private final class CloseAction extends AppAction {

		// public

		CloseAction() {
			super(_("&Close Current"), "ui/close", VK_W, CTRL_MASK);
			setEnabled(false);
			setHTMLHelp(_("Closes the current tab.<br>Modified file is saved automatically."));
		}

		public void onAction() {
			close();
		}

	}

	private final class CloseAllAction extends AppAction {

		// public

		CloseAllAction() {
			super(_("Close &All"), null, VK_W, CTRL_MASK + SHIFT_MASK);
			setEnabled(false);
			setHTMLHelp(_("Closes all tabs.<br>All the modified files are saved automatically."));
		}

		public void onAction() {
			closeAll();
		}

	}

	private final class CloseAllOthersAction extends AppAction {

		// public

		CloseAllOthersAction() {
			super(_("Close All &Others"));
			setEnabled(false);
			setHTMLHelp(_("Closes all other tabs.<br>All the modified files are saved automatically."));
		}

		public void onAction() {
			closeAllOthers();
		}

	}

	private static final class LocateAction extends AppAction {

		// public

		LocateAction() {
			super(_("&Locate In Tree"), "locate", VK_L, CTRL_MASK | SHIFT_MASK);
			setEnabled(false);
			setHTMLHelp(_("Locates the currently edited file in the <b>Tree</b>."));
		}

		public void onAction() {
			MainWindow.getInstance().setSidebarVisible(true);
			Sidebar.goTo(Sidebar.Tab.TREE);
		}

	}

}
