// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.KeyStroke;

/**
 * An application action.
 *
 * EXAMPLE:
 * @code
 * ...
 * // Create File|Quit menu item ..
 * makagiga.common.MMenuBar menuBar = new makagiga.common.MMenuBar();
 * makagiga.common.MMenu fileMenu = new makagiga.common.MMenu("&File");
 * fileMenu.add(new QuitAction());
 * menuBar.add(fileMenu);
 * ...
 * private static final class QuitAction extends AppAction {
 *   QuitAction() {
 *     super(
 *       _("&Quit"),
 *       "quit", // icon name
 *       KeyEvent.VK_Q, AppAction.CTRL_MASK // Ctrl+Q shortcut
 *     );
 *     setHTMLHelp(_("A <b>detailed</b> action description."));
 *   }
 *   public void onAction() {
 *     // menu item clicked ..
 *     makagiga.common.TK.quit();
 *   }
 * }
 * ...
 * @endcode
 */
public abstract class AppAction extends AbstractAction {

	// public

	public static final int ALT_MASK = InputEvent.ALT_MASK;
	public static final int CTRL_MASK = InputEvent.CTRL_MASK;
	public static final int SHIFT_MASK = InputEvent.SHIFT_MASK;

	// private

	private char mnemonic;

	// public

	/**
	 * Constructs a new action.
	 */
	public AppAction() { }

	/**
	 * Constructs a new action.
	 * @param text A text to display
	 */
	public AppAction(final String text) {
		this(text, null, 0, 0);
	}

	/**
	 * Constructs a new action.
	 * @param text A text to display
	 * @param iconName An icon name (@c null = no icon)
	 */
	public AppAction(final String text, final String iconName) {
		this(text, iconName, 0, 0);
	}

	/**
	 * Constructs a new action.
	 * @param text A text to display
	 * @param iconName An icon name (@c null = no icon)
	 * @param keyCode A key code (e.g. @c KeyEvent.VK_X)
	 */
	public AppAction(final String text, final String iconName, final int keyCode) {
		this(text, iconName, keyCode, 0);
	}

	/**
	 * Constructs a new action.
	 * @param text A text to display
	 * @param iconName An icon name (@c null = no icon)
	 * @param keyCode A key code (e.g. @c KeyEvent.VK_X)
	 * @param modifiers Modifiers keys (e.g. @c AppAction.CTRL_MASK)
	 */
	public AppAction(final String text, final String iconName, final int keyCode, final int modifiers) {
		setName(text);

		if (keyCode != 0)
			setAcceleratorKey(keyCode, modifiers);

		setIcon(iconName);
	}

	/**
	 * Invoked when an action occurs.
	 * This function calls @ref onAction.
	 * @param e An action event
	 * @see onAction
	 */
	public void actionPerformed(final ActionEvent e) {
		onAction();
	}

	public KeyStroke getAcceleratorKey() {
		return (KeyStroke)getValue(Action.ACCELERATOR_KEY);
	}

	public void setAcceleratorKey(final int keyCode, final int modifiers) {
		putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(keyCode, modifiers));
	}

	/**
	 * Returns the HTML help, or @c null if no help.
	 */
	public String getHTMLHelp() {
		return (String)getValue(LONG_DESCRIPTION);
	}

	/**
	 * Sets HTML help to @p value.
	 * @b NOTE: Required HTML tags wil be added automatically.
	 */
	public void setHTMLHelp(final String value) {
		putValue(LONG_DESCRIPTION, TK.isEmpty(value) ? null : UI.makeHTML(value));
	}

	public Icon getIcon() {
		return (Icon)getValue(SMALL_ICON);
	}

	public void setIcon(final Icon value) {
		putValue(SMALL_ICON, value);
	}

	public void setIcon(final String name) {
		putValue(SMALL_ICON, (name == null) ? null : FS.getIcon(name));
	}

	/**
	 * Returns the mnemonic character, or @c 0 if action has no mnemonic.
	 */
	public char getMnemonic() { return mnemonic; }

	public abstract void onAction();

	public void setDescription(final String value) {
		putValue(SHORT_DESCRIPTION, value);
	}

	public void setName(final String value) {
		if (TK.isEmpty(value)) {
			mnemonic = 0;
			putValue(NAME, null);
		}
		else {
			mnemonic = UI.getMnemonic(value);
			String s = (mnemonic == 0) ? value : UI.removeMnemonic(value);
			putValue(NAME, s);
		}
	}

	public void setToolTipText(final String value) {
		putValue(LONG_DESCRIPTION, value);
	}

}
