// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.Color;
import java.io.*;
import java.util.Date;
import java.util.Locale;
import java.util.Properties;

/** A configuration file. */
public class Config {

	// public

	/**
	 * The configuration file name.
	 */
	public final StringProperty path = new StringProperty();

	// flags
	public static final int READ_ONLY = 1 << 0;
	public static final int RESOURCE = 1 << 1;

	/**
	 * I18N support.
	 * If this flag is set, you can use the following syntax:
	 * @code
	 * ...
	 * # English text
	 * String.text=Hello
	 * # Polish text
	 * String.text[pl]=Witaj
	 * ...
	 * @endcode
	 */
	public static final int USE_I18N = 1 << 2;

	// private

	private boolean loadError;
	private boolean modified;
	private boolean readOnly;
	private boolean resource;
	private boolean useI18N;
	private static Config global;
	private Properties properties;
	private static String defaultComment;
	private static final String LANG_KEY = "[" + Locale.getDefault().getLanguage() + "]";

	// public

	/**
	 * Constructs a new configuration file.
	 * Configuration values are loaded from the specified file.
	 * @param path A configuration file to load
	 * @param flags Flags (e.g. @ref RESOURCE)
	 * @see isError Use this function to check for load error
	 */
	public Config(final String path, final int flags) {
		readOnly = (flags & READ_ONLY) != 0;
		resource = (flags & RESOURCE) != 0;
		useI18N = (flags & USE_I18N) != 0;
		this.path.set(path);
		loadError = !loadFromFile();
	}

	/**
	 * Constructs a new configuration file.
	 * Configuration values are loaded from the specified file.
	 * @param path A configuration file to load
	 * @see isError Use this function to check for load error
	 */
	public Config(final String path) {
		this(path, 0);
	}

	/**
	 * Constructs a new configuration file.
	 * Configuration values are loaded from the default configuration file.
	 * @see isError Use this function to check for load error
	 */
	public Config() {
		this(FS.makeConfigPath(AppInfo.internalName + ".properties"), 0);
	}

	/**
	 * Returns the global configuration, or @c null if no global configuration.
	 * @see makeGlobal
	 */
	public static Config getGlobal() { return global; }

	/**
	 * Returns @c false if configuration file could not be loaded.
	 */
	public boolean isError() {
		return loadError || (properties == null) || (path == null);
	}

	/**
	 * Makes this configuration the global configuration.
	 * Only one global configuration can be set.
	 * @see getGlobal
	 */
	public void makeGlobal() { global = this; }

	/**
	 * Reads a @c String value from the configuration file.
	 * The default value is @c null.
	 * @param key A configuration key
	 */
	public String read(final String key) {
		return read(key, null);
	}

	/**
	 * Reads a @c boolean value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value. This value is returned if @p key does not exist.
	 */
	public boolean read(final String key, final boolean defaultValue) {
		String value = readRaw("Boolean", key, Boolean.toString(defaultValue));

		return Boolean.parseBoolean(value);
	}

	/**
	 * Reads a @c String value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value. This value is returned if @p key does not exist.
	 */
	public String read(final String key, final String defaultValue) {
		return readRaw("String", key, defaultValue);
	}

	/**
	 * Reads a @c Color value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value. This value is returned if @p key does not exist, or if value is invalid.
	 */
	public Color readColor(final String key, final Color defaultValue) {
		try {
			String value = readRaw("Color", key, null);

			if (TK.isEmpty(value))
				return defaultValue;

			if ((value.length() != 7) || (value.charAt(0) != '#'))
				throw new Exception();

			return new Color(Integer.parseInt(value.substring(1), 16));
		}
		catch (Exception exception) {
			MLogger.error("Config.readColor: Invalid color (Color.%s, %s)", key, path);

			return defaultValue;
		}
	}

	/**
	 * Reads a @c Date value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value. This value is returned if @p key does not exist, or if value is invalid.
	 */
	public Date readDate(final String key, final long defaultValue) {
		try {
			String value = readRaw("Date", key, null);

			if (value == null)
				return new Date(defaultValue);

			return new Date(Long.parseLong(value));
		}
		catch (Exception exception) {
			MLogger.error("Config.readDate: Invalid value (Date.%s, %s)", key, path);

			return new Date(defaultValue);
		}
	}

	/**
	 * Reads an @c int value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value. This value is returned if @p key does not exist, or if value is invalid.
	 */
	public int readInt(final String key, final int defaultValue) {
		try {
			String value = readRaw("Integer", key, null);

			if (value == null)
				return defaultValue;

			return Integer.parseInt(value);
		}
		catch (Exception exception) {
			MLogger.error("Config.readInt: Invalid value (Integer.%s, %s)", key, path);

			return defaultValue;
		}
	}

	/**
	 * Reads an @c int value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value
	 * @param minValue A minimal value
	 * @param maxValue A maximal value
	 * This value is returned if @p key does not exist,
	 * value is invalid,
	 * value is less than @p minValue,
	 * or if value is greater than @p maxValue.
	 */
	public int readInt(final String key, final int defaultValue, final int minValue, final int maxValue) {
		int value = readInt(key, defaultValue);

		if (value < minValue || value > maxValue)
			return defaultValue;

		return value;
	}

	/**
	 * Reads a @c long value from the configuration file.
	 * @param key A configuration key
	 * @param defaultValue A default value. This value is returned if @p key does not exist, or if value is invalid.
	 */
	public long readLong(final String key, final long defaultValue) {
		try {
			return Long.parseLong(readRaw("Integer", key, ""));
		}
		catch (Exception exception) {
			MLogger.error("Config.readLong: Invalid value (Integer.%s, %s)", key, path);

			return defaultValue;
		}
	}

	public void removeBoolean(final String key) {
		removeRaw("Boolean." + key);
	}

	public void removeString(final String key) {
		removeRaw("String." + key);
	}

	/**
	 * Saves the configuration to the file.
	 * You can change target file name using @ref path.
	 * @return @c true if successful; otherwise @c false
	 */
	public boolean saveToFile() {
		if (readOnly)
			return false;

		if (!modified)
			return true;

		if ((path == null) || (properties == null))
			return false;

		BufferedOutputStream output = null;
		try {
			output = new BufferedOutputStream(new FileOutputStream(path.get()));
			properties.store(output, defaultComment);

			modified = false;

			return true;
		}
		catch (IOException exception) {
			MLogger.exception(exception);

			return false;
		}
		finally {
			FS.close(output);
		}
	}

	/**
	 * Sets the default comment to @p value.
	 * The comment value is stored in the configuration file as a comment.
	 */
	public static void setDefaultComment(final String value) { defaultComment = value; }

	/**
	 * Writes a @c boolean value.
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final boolean value) {
		writeRaw("Boolean", key, Boolean.toString(value));
	}

	/**
	 * Writes a @c BooleanProperty value.
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final BooleanProperty value) {
		write(key, value.get());
	}

	/**
	 * Writes a @c Color value.
	 * The value is stored in the HTML triplet format (e.g. #ff0000 = red color, #rrggbb).
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final Color value) {
		writeRaw("Color", key, (value == null) ? "" : TK.toHTMLTriplet(value));
	}

	/**
	 * Writes a @c ColorProperty value.
	 * The value is stored in the HTML triplet format (e.g. #ff0000 = red color, #rrggbb).
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final ColorProperty value) {
		write(key, value.get());
	}

	/**
	 * Writes a @c Date value.
	 * The date is stored as the @c long type.
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final Date value) {
		writeRaw("Date", key, (value == null) ? "0" : Long.toString(value.getTime()));
	}

	/**
	 * Writes a @c long value.
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final long value) {
		writeRaw("Integer", key, Long.toString(value));
	}

	/**
	 * Writes a @c String value.
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final String value) {
		writeRaw("String", key, value);
	}

	/**
	 * Writes a @c StringProperty value.
	 * @param key A configuration key
	 * @param value A value to write
	 */
	public void write(final String key, final StringProperty value) {
		write(key, value.get());
	}

	// private

	private boolean loadFromFile() {
		if (path == null)
			return false;

		properties = new Properties();

		BufferedInputStream input = null;
		try {
			if (resource)
				input = new BufferedInputStream(getClass().getClassLoader().getResourceAsStream(path.get()));
			else
				input = new BufferedInputStream(new FileInputStream(path.get()));
			properties.load(input);

			modified = false;

			return true;
		}
		catch (FileNotFoundException exception) {
			return false;
		}
		catch (IOException exception) {
			return false;
		}
		finally {
			FS.close(input);
		}
	}

	private String readRaw(final String prefix, final String key, final String defaultValue) {
		if ((properties == null) || (prefix == null) || (key == null))
			return defaultValue;

		String v;
		if (useI18N) {
			v = properties.getProperty(prefix + "." + key + LANG_KEY, null);

			if (v != null)
				return v;
		}
		v = properties.getProperty(prefix + "." + key, defaultValue);

		return (v == null) ? defaultValue : v;
	}

	private void removeRaw(final String key) {
		if (properties != null) {
			properties.remove(key);
			modified = true;
		}
	}

	private void writeRaw(final String prefix, final String key, final String value) {
		if (readOnly) {
			MLogger.warning("Config.writeRaw: config is read only (%s)", path);

			return;
		}

		if (key == null) {
			MLogger.error("Config.writeRaw: key == null (%s)", path);

			return;
		}

		if (prefix == null) {
			MLogger.error("Config.writeRaw: prefix == null (%s)", path);

			return;
		}

		try {
			String old = properties.getProperty(prefix + "." + key, "");
			if (!old.equals(value)) {
				properties.setProperty(prefix + "." + key, value);
				modified = true;
			}
		}
		catch (NullPointerException exception) {
			// "allow" null values
		}
	}

}
