// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.Image;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;
import javax.swing.Icon;
import javax.swing.ImageIcon;

/** A file system wrapper and icon loader. */
public final class FS {

	// public

	public static final int CREATE_DIR = 1;
	public static final int DEFAULT_ICON_SIZE = -1; /**< Default icon size (-1). */
	public static final int MAX_ICON_SIZE = 48; /**< Max. icon size. */
	public static final int MIN_ICON_SIZE = 16; /**< Min. icon size. */

	/**
	 * The current icon size.
	 */
	public static final IntegerProperty iconSize = new IntegerProperty(DEFAULT_ICON_SIZE);

	// private

	private static String configDir;
	private static String iconDir = "images";
	private static String iconType = ".png";
	private static String userDir;

	// public

	/**
	 * Silently closes a stream/reader/writer.
	 * @param closeable An object to close
	 * @return @c true if succsessfull; otherwise @c false
	 */
	public static boolean close(final Closeable closeable) {
		if (closeable == null)
			return false;

		try {
			closeable.close();

			return true;
		}
		catch (IOException exception) {
			return false;
		}
	}

	public static boolean copyFile(final String sourceFile, final String destinationFile) {
		BufferedInputStream input = null;
		BufferedOutputStream output = null;
		try {
			input = new BufferedInputStream(new FileInputStream(sourceFile));
			output = new BufferedOutputStream(new FileOutputStream(destinationFile));
			byte[] buf = new byte[4096];
			int bufLen;
			while ((bufLen = input.read(buf)) != -1)
				output.write(buf, 0, bufLen);

			return true;
		}
		catch (FileNotFoundException exception) {
			return false;
		}
		catch (IOException exception) {
			return false;
		}
		finally {
			close(input);
			close(output);
		}
	}

	/**
	 * Deletes a single @p path (file or directory, no recursion).
	 * @return @c true if succsessfull; otherwise @c false
	 */
	public static boolean delete(final String path) {
		return new File(path).delete();
	}

	/**
	 * Returns @c true if @p path already exists.
	 */
	public static boolean exists(final String path) {
		return new File(path).exists();
	}

	/**
	 * Creates and returns the user configuration directory.
	 * (e.g. "/home/kdt/.internalname")
	 * Internal name is a value returned by @ref AppInfo.internalName.
	 * @see getUserDir
	 */
	public static String getConfigDir() {
		if (configDir == null) {
			configDir = makePath(getUserDir(), "." + AppInfo.internalName);
			File file = new File(configDir);
			file.mkdirs();
			if (!file.exists()) {
				MMessage.fileError(
					null,
					// no i18n
					"Could not initialize configuration directory",
					configDir
				);
				TK.quit(1);
			}
			MLogger.info("Configuration directory: \"%s\"", configDir);
		}

		return configDir;
	}

	/**
	 * Loads icon from the specified resource.
	 * @b NOTE: Custom icon is not cached.
	 * @param baseClass A base class for resource loader
	 * @param path A full icon path relative to @p baseClass (e.g. "images/icon.png")
	 * @param size An icon size (@ref DEFAULT_ICON_SIZE - do not resize)
	 */
	public static Icon getCustomIcon(final Object baseClass, final String path, final int size) {
		URL url = baseClass.getClass().getResource(path);

		if (url == null)
			return null;

		if (size == DEFAULT_ICON_SIZE)
			return new ImageIcon(url);

		return new ImageIcon(
			new ImageIcon(url).getImage().getScaledInstance(
				size,
				size,
				Image.SCALE_SMOOTH
			)
		);
	}

	public static Icon getCustomIcon(final String path, final int size) {
		if (size == DEFAULT_ICON_SIZE)
			return new ImageIcon(path);

// !!!: common scale code [0.9.3]
		return new ImageIcon(
			new ImageIcon(path).getImage().getScaledInstance(
				size,
				size,
				Image.SCALE_SMOOTH
			)
		);
	}

	/**
	 * Returns an image icon with default size.
	 * @param name An icon name without extension
	 * @see getIcon
	 */
	public static ImageIcon getDefaultIcon(final String name) {
		return getIcon(name, DEFAULT_ICON_SIZE);
	}

	/**
	 * Reads an icon from the cache or file.
	 * The icon loaded from the file if is stored in the icon cache.
	 * @param name An icon name without extension (e.g. "exit")
	 * @param size An icon size (-1 = default icon size)
	 * @return An image icon if succsessfull; otherwise @c null
	 * @see getImage
	 */
	public static ImageIcon getIcon(final String name, final int size) {
// !!!: optimize image load [0.9.3]
		final IconCache iconCache = IconCache.getInstance();
		String path = iconDir + "/" + name + iconType;

		// fix icon size
		int s = ((size < MIN_ICON_SIZE) || (size > MAX_ICON_SIZE)) ? DEFAULT_ICON_SIZE : size;

		// read icon from the cache
		String cachePath = path + "-" + s;
		ImageIcon icon = iconCache.get(cachePath);
		if (icon == null) {
			// icon not found in the cache; read from the file
			URL url = FS.class.getClassLoader().getResource(path);
			if (url == null) {
// !!!: try to load icons from the user home directory
				MLogger.error("Could not find icon: \"%s\"", path);

				return null;
			}
			icon = new ImageIcon(url);
			int iconWidth = icon.getIconWidth();
			if (iconWidth < 1) {
				MLogger.error("Invalid icon: \"%s\"", path);

				return null;
			}
			if ((s != DEFAULT_ICON_SIZE) && (s != iconWidth)) {
				String description = icon.toString(); // save path
				icon = new ImageIcon(icon.getImage().getScaledInstance(s, s, Image.SCALE_SMOOTH));
				icon.setDescription(description); // restore path
			}
			// add icon to the cache
			iconCache.put(cachePath, icon);
			// MLogger.info("FS.getIcon: Caching icon: \"%s\" as \"%s\"", name, cachePath);
		}
		// else {
		//	MLogger.info("FS.getIcon: Cached icon: %s", name);
		// }

		return icon;
	}

	/**
	 * Reads an icon from the cache or file.
	 * The icon size is @ref iconSize.
	 * @param name An icon name without extension
	 * @return An image icon if succsessfull; otherwise @c null
	 * @see getImage
	 */
	public static ImageIcon getIcon(final String name) {
		return getIcon(name, iconSize.get());
	}

	/**
	 * Reads an image from the cache or file.
	 * The icon size is @ref iconSize.
	 * @param name An image name without extension
	 * @return An image if succsessfull; otherwise @c null
	 * @see getIcon
	 */
	public static Image getImage(final String name) {
		ImageIcon imageIcon = getDefaultIcon(name);

		if (imageIcon == null)
			return null;

		return imageIcon.getImage();
	}

	/**
	 * Returns a file type (extension) extracted from the @p path.
	 *
	 * EXAMPLE:
	 * @code
	 * ...
	 * FS.getType("image.png") == ".png"
	 * FS.getType("my.image.png") == ".png"
	 * FS.getType("image") == null
	 * FS.getType(".linux_hidden") == null
	 * ...
	 * @endcode
	 */
	public static String getType(final String path) {
		int i = path.lastIndexOf(".");

		if ((i == -1) || (i == 0) /* i == 0 - hidden file */)
			return null;

		return path.substring(i);
	}

	/**
	 * Returns the user home directory. (e.g. "/home/kdt")
	 * @see getConfigDir
	 */
	public static String getUserDir() {
		if (userDir == null) {
			userDir = System.getProperty("user.home");
			MLogger.info("User directory: \"%s\"", userDir);
		}

		return userDir;
	}

	/**
	 * Returns an UTF-8 reader.
	 * @param path A file to read
	 * @return A BufferedReader if succsessfull; otherwise @c null
	 */
	public static BufferedReader getUTF8Reader(final String path) throws FileNotFoundException {
		try {
			return new BufferedReader(
				new InputStreamReader(
					new FileInputStream(path),
					"UTF8"
				)
			);
		}
		catch (UnsupportedEncodingException exception) {
			MLogger.exception(exception);

			return null;
		}
	}

	/**
	 * Returns an UTF-8 writer.
	 * @param path A file to write
	 * @return A BufferedWriter if succsessfull; otherwise @c null
	 */
	public static BufferedWriter getUTF8Writer(final String path) throws FileNotFoundException {
		try {
			return new BufferedWriter(
				new OutputStreamWriter(
					new FileOutputStream(path),
					"UTF8"
				)
			);
		}
		catch (UnsupportedEncodingException exception) {
			MLogger.exception(exception);

			return null;
		}
	}

	/**
	 * Returns a list of files.
	 * @param dir A directory to list
	 */
	public static File[] list(final String dir) {
		return new File(dir).listFiles();
	}

	public static String makeConfigPath(final String part2) {
		return makeConfigPath(part2, 0);
	}

	public static String makeConfigPath(final String part2, final int flag) {
		String path = makePath(getConfigDir(), part2);
		if (flag == CREATE_DIR)
			mkdirs(path);

		return path;
	}

	/**
	 * Constructs and returns a new path created from two parts (part1 + File.separator + part2).
	 *
	 * EXAMPLE:
	 * @code
	 * ...
	 * FS.makePath("/usr/src", "linux"); // "/usr/src/linux"
	 * ...
	 * @endcode
	 */
	public static String makePath(final String part1, final String part2) {
		return part1 + File.separator + part2;
	}

	/**
	 * Creates a new directory with name @p path.
	 * @return @c true if succsessfull; otherwise @c false
	 */
	public static boolean mkdir(final String path) {
		return new File(path).mkdir();
	}

	/**
	 * Creates a new directory and its subdirectories.
	 * @param path A directory to create
	 * @return @c true if succsessfull; otherwise @c false
	 */
	public static boolean mkdirs(final String path) {
		return new File(path).mkdirs();
	}

	/**
	 * Renames file @p oldName to @p newName.
	 * @return @c true if succsessfull; otherwise @c false
	 */
	public static boolean rename(final String oldName, final String newName) {
		File oldFile = new File(oldName);
		File newFile = new File(newName);

		return oldFile.renameTo(newFile);
	}

	/**
	 * Converts @p path to URL.
	 *
	 * EXAMPLE:
	 * @code
	 * ...
	 * FS.toURL("/usr/src/linux"); // == "file:/usr/src/linux/"
	 * ...
	 * @endcode
	 */
	public static String toURL(final String path) {
		try {
			// since JDK6.0 "File(path).toURL()" is deprecated
			return new File(path).toURI().toURL().toString();
		}
		catch (MalformedURLException exception) {
			return null;
		}
	}

	// public classes

	/**
	 * Icon cache.
	 * @see FS.getIcon
	 */
	public static final class IconCache extends Hashtable<String, ImageIcon> {

		// private

		private static IconCache _instance;

		// public

		/**
		 * Returns the instance of the icon cache.
		 */
		public static IconCache getInstance() {
			if (_instance == null)
				_instance = new IconCache();

			return _instance;
		}

		// private

		private IconCache() { }

	}

}
