// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.io.File;
import java.util.Arrays;
import java.util.Comparator;

/**
 * A recursive file scanner.
 * The files are sorted by name (directories are first).
 * See @ref org.makagiga.tools.Backup and @ref org.makagiga.tree.TreeList for details.
 *
 * EXAMPLE:
 * @code
 * ...
 * new FileScanner<Object>("/home/kdt") {
 *   public Object processDir(final Object parentItem, final File dir) {
 *     ...
 *     return null;
 *   }
 *   public void processFile(final Object parentItem, final File file) {
 *     ...
 *   }
 * };
 * ...
 * @endcode
 */
public abstract class FileScanner<I> extends AbstractScanner {

	// protected

	/**
	 * A current file name.
	 */
	protected String fileName;

	/**
	 * A current file path (directory + name).
	 */
	protected String filePath;

	// public

	/**
	 * Constructs a new file scanner.
	 * @param parentItem A parent item. Can be @c null.
	 * @param root A start directory
	 */
	public FileScanner(final I parentItem, final String root) {
		if (!TK.isEmpty(root))
			recurse(parentItem, root);
	}

	/**
	 * Constructs a new file scanner.
	 * @param root A start directory
	 */
	public FileScanner(final String root) {
		this(null, root);
	}

	/**
	 * Called on each directory.
	 * @param parentItem A parent item. Can be @c null.
	 * @param dir A directory
	 */
	public abstract I processDir(final I parentItem, final File dir);

	/**
	 * Called on each regular file.
	 * @param parentItem A parent item. Can be @c null.
	 * @param file A regular file
	 */
	public abstract void processFile(final I parentItem, final File file);

	// private

	private void recurse(final I parentItem, final String dir) {
		if (isStopped())
			return;

		File[] files = FS.list(dir);

		if (TK.isEmpty(files))
			return;

		// sort files by name; directories first
// !!!: singleton Comparator [0.9.3]
		Comparator<File> filesComparator = new Comparator<File>() {
			public int compare(File f1, File f2) {
				boolean d1 = f1.isDirectory();
				boolean d2 = f2.isDirectory();
				String n1 = f1.getName();
				String n2 = f2.getName();

				if (d1 && !d2)
					return -1;

				if (d2 && !d1)
					return 1;

				return n1.compareTo(n2);
			}
		};
		Arrays.sort(files, filesComparator);

		for (File i : files) {
			fileName = i.getName();
			filePath = FS.makePath(dir, fileName);
			// folder
			if (i.isDirectory()) {
				I parent = processDir(parentItem, i);

				if (isStopped())
					return;

				recurse(parent, filePath);

				if (isStopped())
					return;
			}
			// file
			else if (i.isFile()) {
				processFile(parentItem, i);

				if (isStopped())
					return;
			}
		}
	}

}
