// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.Color;

/**
 * A HTML builder.
 * See @ref org.makagiga.editors.todo.TodoEditorCore.saveToFileAsHTML to see how to use this class.
 */
public class HTMLBuilder extends MarkupBuilder {

	// public

	public static final String HTML_TRANSITIONAL = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">"; /**< HTML 4.01 Transitional. */
	public static final String XHTML_STRICT = "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"; /**< XHTML 1.0 Strict. */

	public static final String UTF_8 = "UTF-8"; /**< UTF-8 encoding */

	// public

	/**
	 * Constructs a new HTML builder.
	 * @param doctype A document type (DOCTYPE) (e.g. @ref HTML_TRANSITIONAL)
	 */
	public HTMLBuilder(final String doctype) {
		super(doctype.equals(XHTML_STRICT));
		appendLine(doctype);
	}

	/**
	 * Adds an anchor.
	 * @code
	 * ...
	 * builder.addAnchor("top"); // <a name="top"></a>\n
	 * ...
	 * @endcode
	 * @param name An anchor name
	 */
	public void addAnchor(final String name) {
		appendLine("<a name=\"%s\"></a>", name);
	}

	/**
	 * Adds an CSS attribute.
	 * @code
	 * ...
	 * builder.addAttr("border", "1px solid black");
	 * // border: 1px solid black;\n
	 * ...
	 * @endcode
	 * @param name An attribute name
	 * @param value An attribute value
	 */
	public void addAttr(final String name, final String value) {
		appendLine("%s: %s;", name, value);
	}

	/**
	 * Adds an CSS color attribute.
	 * @code
	 * ...
	 * builder.addAttr("background-color", Color.RED);
	 * // background-color: #ff0000;\n
	 * ...
	 * @endcode
	 * @param name An attribute name
	 * @param value A color value
	*/
	public void addAttr(final String name, final Color value) {
		appendLine("%s: %s;", name, TK.toHTMLTriplet(value));
	}

	public void addFeed(final String href, final String title) {
		addLink(href, "alternate", "application/rss+xml", title);
	}

	public void addLink(final String href, final String rel, final String type, final String title) {
		appendLine(
			"<link href=\"%s\" rel=\"%s\" type=\"%s\"%s" + tagEnd,
			href,
			rel,
			type,
			(title == null) ? "" : (" title=\"" + title + "\"")
		);
	}

	/**
	 * Adds a meta information.
	 * @code
	 * ...
	 * builder.addMeta("Description", "\"Foo\" Home Page");
	 * // <meta name="Description" content="&quot;Foo&quot; Home Page">\n
	 * ...
	 * @endcode
	 * @param name A meta information name
	 * @param value A meta information value
	 */
	public void addMeta(final String name, final String value) {
		appendLine(
			String.format(
				"<meta name=\"%s\" content=\"%s\"" + tagEnd,
				name,
				TK.escapeXML(value)
			)
		);
	}

	public void addStyle(final String href) {
		addLink(href, "stylesheet", "text/css", null);
	}

	public void appendLine(final String code, final boolean nbspIfEmpty) {
		if (TK.isEmpty(code) && nbspIfEmpty)
			appendLine("&nbsp;");
		else
			appendLine(code);
	}

	/**
	 * Begins the HTML document (ends HEAD, begins BODY)
	 */
	public void beginDoc() {
		endTag("head");
		beginTag("body");
	}

	/**
	 * Begins the HTML (inserts HTML, HEAD, TITLE and META (Content-Type) tags.
	 * @param title A title
	 * @param encoding An encoding
	 */
	public void beginHTML(final String title, final String encoding) {
		if (xml) {
			beginTag(
				"html",
				"xmlns", "http://www.w3.org/1999/xhtml"
			);
		}
		else {
			beginTag("html");
		}
		decIndent();
		beginTag("head");
		appendChars("title", title);
		if (encoding != null)
			appendLine("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=%s\"" + tagEnd, encoding);
	}

	/**
	 * Begins the HTML (inserts HTML, HEAD and TITLE tags.
	 * @param title A title
	 */
	public void beginHTML(final String title) {
		beginHTML(title, null);
	}

	/**
	 * Begins an CSS rule.
	 * @code
	 * ...
	 * builder.beginRule("body"); // body {\n
	 * ...
	 * @endcode
	 */
	public void beginRule(final String name) {
		appendLine("%s {", name);
		indent++;
	}

	/**
	 * Begins an CSS style.
	 * @code
	 * ...
	 * // &lt;style type="text/css">\n
	 * // &lt;!--\n
	 * ...
	 * @endcode
	 */
	public void beginStyle() {
		appendLine("<style type=\"text/css\">");
		appendLine(xml ? "<![CDATA[" : "<!--");
	}

	public String createImage(final String className, final String src, final String alt, final String title) {
		return createImage(className, src, alt, title, -1, -1);
	}

	public String createImage(final String className, final String src, final String alt, final String title, final int width, final int height) {
		return String.format(
			"<img%s src=\"%s\"%s%s alt=\"%s\"%s" + tagEnd,
			(className == null) ? "" : (" class=\"" + className + "\""),
			src,
			(width == -1) ? "" : (" width=\"" + width + "\""),
			(height == -1) ? "" : (" height=\"" + height + "\""),
			alt,
			(title == null) ? "" : (" title=\"" + title + "\"")
		);
	}

	/**
	 * Returns a link tag.
	 * @code
	 * ...
	 * builder.createLink("http://sf.net", "SourceForge.net");
	 * // <a href="http://sf.net">SourceForge.net</a>
	 * ...
	 * @endcode
	 * @param url An URL
	 * @param name An URL name
	 */
	public static String createLink(final String url, final String name) {
		return String.format("<a href=\"%s\">%s</a>", url, name);
	}

	/**
	 * Returns a link tag.
	 * @code
	 * ...
	 * builder.createLink("http://sf.net");
	 * // <a href="http://sf.net">http://sf.net</a>
	 * ...
	 * @endcode
	 * @param url An URL
	 */
	public static String createLink(final String url) {
		return createLink(url, url);
	}

	/**
	 * Ends the HTML document (inserts BODY and HTML).
	 */
	public void endDoc() {
		endTag("body");
		endTag("html");
	}

	/**
	 * Ends an CSS rule (inserts "}").
	 */
	public void endRule() {
		decIndent();
		appendLine("}");
	}

	/**
	 * Ends an CSS style.
	 * @code
	 * ...
	 * builder.endStyle();
	 * // -->\n
	 * // </style>\n
	 * ...
	 * @endcode
	 */
	public void endStyle() {
		appendLine(xml ? "]]>" : "-->");
		appendLine("</style>");
	}

	/**
	 * Adds the BR tag.
	 */
	public void newLine() {
		append("<br" + tagEnd);
	}

	/**
	 * Prints a header.
	 * @code
	 * ...
	 * builder.printHeader(1, "Hello");
	 * // <h1>Hello</h1>\n
	 * ...
	 * @endcode
	 * @param level A header level
	 * @param text A text
	 */
	public void printHeader(final int level, final String text) {
		appendLine("<h%d>%s</h%d>", level, text, level);
	}

	/**
	 * Appends a default style for A and BODY tags.
	 * @code
	 * ...
	 * builder.beginStyle();
	 * builder.setDefaultStyle();
	 * ...
	 * @endcode
	 */
	public void setDefaultStyle() {
		beginRule("a");
			addAttr("color", "blue");
			addAttr("text-decoration", "none");
		endRule();
		beginRule("a:hover");
			addAttr("color", "red");
			addAttr("text-decoration", "underline");
		endRule();
		beginRule("body");
			addAttr("background-color", "white");
			addAttr("color", "black");
			addAttr("font-family", "helvetica, arial, sans-serif");
		endRule();
	}

}
