// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static java.awt.Color.*;

import static org.makagiga.common.UI._;

import java.awt.Color;
import javax.swing.BorderFactory;
import javax.swing.JColorChooser;
import javax.swing.border.Border;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/** A color chooser. */
public class MColorChooser extends MDialog {

	// private

	private ChangeListener changeListener;
	private Color color;
	private Color defaultColor;

	private static JColorChooser colorChooser;
	private static MLabel backgroundPreview;
	private static MLabel foregroundPreview;
	private static MPanel previewPanel;

	// public

	/**
	 * Constructs a new color chooser.
	 * @param parent A parent frame
	 * @param title A dialog title
	 * @param initialColor An initial color
	 * @param defaultColor A default color
	 * @see Use @ref showDialog instead
	 */
	public MColorChooser(final MFrame parent, final String title, final Color initialColor, final Color defaultColor) {
		super(parent, title, STANDARD_DIALOG | DEFAULT_BUTTON);
		this.defaultColor = defaultColor;

		// init preview
		if (previewPanel == null) {
			previewPanel = MPanel.createGridPanel(2);
			backgroundPreview = new MLabel(_("Preview"));
			backgroundPreview.setOpaque(true);
			previewPanel.add(backgroundPreview);
			foregroundPreview = new MLabel(_("Preview"));
			previewPanel.add(foregroundPreview);
		}
		addEast(previewPanel);

		// init color chooser
		if (colorChooser == null) {
			colorChooser = new JColorChooser();
			colorChooser.setPreviewPanel(previewPanel);
		}
		changeListener = new ChangeListener() {
			public void stateChanged(final ChangeEvent e) {
				setColor(colorChooser.getColor());
			}
		};
		colorChooser.getSelectionModel().addChangeListener(changeListener);

		addCenter(colorChooser);

		setColor(initialColor);

		pack();
		setResizable(false);
	}

	/**
	 * Displays a color chooser dialog.
	 * @param parent A parent frame
	 * @param title A dialog title
	 * @param initialColor An initial color
	 * @return The selected color, or @c null if dialog has been cancelled.
	 */
	public static Color colorBox(final MFrame parent, final String title, final Color initialColor) {
		return colorBox(parent, title, initialColor, initialColor);
	}

	/**
	 * Displays a color chooser dialog.
	 * @param parent A parent frame
	 * @param title A dialog title
	 * @param initialColor An initial color
	 * @param defaultColor A default color
	 * @return The selected color, or @c null if dialog has been cancelled.
	 */
	public static Color colorBox(final MFrame parent, final String title, final Color initialColor, final Color defaultColor) {
		MColorChooser cc = new MColorChooser(parent, title, initialColor, defaultColor);

		if (cc.isOK())
			return cc.getColor();

		return null;
	}

	/**
	 * Returns the selected color, or @c null if dialog has been cancelled.
	 */
	public Color getColor() { return color; }

	public void setColor(final Color value) {
		color = value;
		Border emptyBorder = BorderFactory.createEmptyBorder(10, 10, 10, 10);
		Border lineBorder;
		if (color == null) {
			lineBorder = BorderFactory.createLineBorder(getBackground(), 2);
			backgroundPreview.setEnabled(false);
			foregroundPreview.setEnabled(false);
		}
		else {
			lineBorder = BorderFactory.createLineBorder(color, 2);
			backgroundPreview.setEnabled(true);
			foregroundPreview.setEnabled(true);
			// update chooser
			colorChooser.setColor(color);
		}
		backgroundPreview.setBackground(color);
		foregroundPreview.setForeground(color);
		Border border = BorderFactory.createCompoundBorder(lineBorder, emptyBorder);
		backgroundPreview.setBorder(border);
		foregroundPreview.setBorder(border);
	}

	// protected

	@Override
	protected void done(final MDialogAction.Code value) {
		if ((changeListener != null) && (colorChooser != null))
			colorChooser.getSelectionModel().removeChangeListener(changeListener);
		super.done(value);
	}

	@Override
	protected void onDefault() {
		setColor(defaultColor);
	}

}
