// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static java.awt.event.KeyEvent.*;

import static org.makagiga.common.UI._;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JComponent;
import javax.swing.JDialog;

/** A dialog base. */
public class MDialog extends JDialog implements MBorderLayout {

	// public

	public static final int CANCEL_BUTTON = 1 << 0;
	public static final int OK_BUTTON = 1 << 1;
	public static final int MODAL = 1 << 2;
	public static final int DEFAULT_BUTTON = 1 << 3;
	public static final int OPTIONS_BUTTON = 1 << 4;

	/**
	 * A modal dialog with @b OK and @b Cancel buttons.
	 */
	public static final int STANDARD_DIALOG = OK_BUTTON | CANCEL_BUTTON | MODAL;

	// private

	private int flags;
	private MDialogAction.Code code = MDialogAction.Code.REJECTED;
	private MButton cancelButton;
	private MButton okButton;
	private MButton defaultButton;
	private MURLButton optionsButton;

	// public

// MUSTANG: Dialog(Window parent)
	/**
	 * Constructs a new dialog with border layout.
	 * @param parent A parent dialog
	 * @param title A dialog title
	 * @param flags A dialog flags
	 */
	public MDialog(final MDialog parent, final String title, final int flags) {
		super(parent);
		init(title, flags);
	}

	/**
	 * Constructs a new dialog with border layout.
	 * @param parent A parent frame
	 * @param title A dialog title
	 * @param flags A dialog flags
	 */
	public MDialog(final MFrame parent, final String title, final int flags) {
		super(parent);
		init(title, flags);
	}

	/**
	 * Executes the modal or non-modal dialog.
	 * The window is centered on the screen.
	 * @see isOK
	 */
	public MDialogAction.Code exec() {
		Container parent = getParent();
		if (parent == null)
			UI.centerOnScreen(this);
		else
			setLocationRelativeTo(parent);

		if (isModal() && (parent instanceof MDialog)) {
			MDialog parentDialog = MDialog.class.cast(parent);
			BooleanProperty isCancelEnabled =
				(parentDialog.getCancelButton() == null)
				? null
				: new BooleanProperty(parentDialog.getCancelButton().isEnabled());
			BooleanProperty isOKEnabled =
				(parentDialog.getOKButton() == null)
				? null
				: new BooleanProperty(parentDialog.getOKButton().isEnabled());
			// disable OK/Cancel buttons
			if ((isCancelEnabled != null) && isCancelEnabled.get())
				parentDialog.getCancelButton().setEnabled(false);
			if ((isOKEnabled != null) && isOKEnabled.get())
				parentDialog.getOKButton().setEnabled(false);
			// show dialog
			setVisible(true);
			// enable OK/Cancel buttons
			if ((isCancelEnabled != null) && isCancelEnabled.get())
				parentDialog.getCancelButton().setEnabled(true);
			if ((isOKEnabled != null) && isOKEnabled.get())
				parentDialog.getOKButton().setEnabled(true);
		}
		else {
			setVisible(true);
		}

		return code;
	}

	/**
	 * Returns the @b Cancel button.
	 */
	public MButton getCancelButton() { return cancelButton; }

	/**
	 * Returns the @b OK button.
	 */
	public MButton getOKButton() { return okButton; }

	public MURLButton getOptionsButton() { return optionsButton; }

	/**
	 * Executes the dialog.
	 * @return @c true if @b OK button has been clicked; otherwise @c false
	 */
	public boolean isOK() {
		return exec() == MDialogAction.Code.ACCEPTED;
	}

	// MBorderLayout

	public void addCenter(final JComponent component) {
		UI.addCenter(this, component);
	}

	public void addEast(final JComponent component) {
		add(component, BorderLayout.EAST);
	}

	public void addNorth(final JComponent component) {
		add(component, BorderLayout.NORTH);
	}

	public void addSouth(final JComponent component) {
		add(component, BorderLayout.SOUTH);
	}

	public void addWest(final JComponent component) {
		add(component, BorderLayout.WEST);
	}

	// protected

	/**
	 * Sets return code to @p value,
	 * and disposes the dialog.
	 */
	protected void done(final MDialogAction.Code value) {
		code = value;
		switch (code) {
			case ACCEPTED:
				if (this instanceof MDialogAction)
					MDialogAction.class.cast(this).accept();
				break;
			case REJECTED:
				if (this instanceof MDialogAction)
					MDialogAction.class.cast(this).reject();
				break;
		}
		dispose();
	}

	protected boolean isButton(final int button) {
		return (flags & button) != 0;
	}

	protected void onDefault() { }

	protected void onOptions() { }

	// private

	private void doCancel() {
		if ((cancelButton != null) && cancelButton.isEnabled() && cancelButton.isVisible()) {
			cancelButton.doClick();
		}
		else {
			done(MDialogAction.Code.REJECTED);
		}
	}

	private boolean doClick(final MButton b) {
		if ((b != null) && b.hasFocus() && b.isEnabled() && b.isVisible()) {
			b.doClick();

			return true;
		}

		return false;
	}

	private void init(final String title, final int flags) {
		this.flags = flags;
		setModal((this.flags & MODAL) != 0);
		setTitle(title);

		// set custom content pane
		MPanel contentPane = MPanel.createBorderPanel(5);
		contentPane.setMargin(10);
		setContentPane(contentPane);

		MPanel buttonsPanel = null;
		// init buttons
		if (isButton(OK_BUTTON) || isButton(CANCEL_BUTTON)) {
			buttonsPanel = MPanel.createHBoxPanel();

			// add "Default" button
			if (isButton(DEFAULT_BUTTON)) {
				defaultButton = new MButton(_("&Default"), FS.getIcon("ui/undo")) {
					@Override
					protected void onClick() {
						onDefault();
					}
				};
				buttonsPanel.add(defaultButton);
			}

			// add "Options..." button
			if (isButton(OPTIONS_BUTTON)) {
				optionsButton = new MURLButton(_("Options...")) {
					@Override
					protected void onClick() {
						onOptions();
					}
				};
				if (isButton(DEFAULT_BUTTON))
					buttonsPanel.addGap();
				buttonsPanel.add(optionsButton);
			}

			// add stretch and gap
			buttonsPanel.addStretch();
			buttonsPanel.addGap();

// !!!: option to swap OK/Cancel buttons (Mac/Win style)  [0.9.3]
			// add "OK" button
			if (isButton(OK_BUTTON)) {
				okButton = new MButton(_("&OK"), FS.getIcon("ui/ok")) {
					@Override
					protected void onClick() {
						done(MDialogAction.Code.ACCEPTED);
					}
				};
				buttonsPanel.add(okButton);

				// set default button
				if (!isButton(CANCEL_BUTTON)) {
					okButton.setDefaultCapable(true);
					getRootPane().setDefaultButton(okButton);
				}
			}

			// add gap between buttons
			if (isButton(OK_BUTTON) && isButton(CANCEL_BUTTON))
				buttonsPanel.addGap();

			// add "Cancel" button
			if (isButton(CANCEL_BUTTON)) {
				cancelButton = new MButton(_("&Cancel"), FS.getIcon("ui/cancel")) {
					@Override
					protected void onClick() {
						done(MDialogAction.Code.REJECTED);
					}
				};
				buttonsPanel.add(cancelButton);
			}

			if (buttonsPanel != null) {
				buttonsPanel.setMargin(5, 0, 0, 0);
				addSouth(buttonsPanel);
			}
		}
		UI.connect(
			contentPane, UI.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, VK_ENTER,
			new SimpleAction("MDialog.accept") {
				public void onAction() {
					if (doClick(cancelButton))
						return;

					if (doClick(okButton))
						return;

					if (doClick(defaultButton))
						return;

					if (doClick(optionsButton))
						return;

					if (okButton == null)
						done(MDialogAction.Code.ACCEPTED);
					else
						okButton.doClick();
				}
			}
		);
		UI.connect(
			contentPane, UI.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, VK_ESCAPE,
			new SimpleAction("MDialog.reject") {
				public void onAction() {
					doCancel();
				}
			}
		);

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(final WindowEvent e) {
				doCancel();
			}
		} );
	}

}
