// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import javax.swing.DefaultListModel;
import javax.swing.JList;
import javax.swing.ListSelectionModel;

/**
 * A list.
 *
 * EXAMPLE:
 * @code
 * ...
 * makagiga.common.MList<String> list = makagiga.common.new MList<String>();
 * list.addItem("Item 1");
 * list.addItem("Item 2");
 * // sort items by name
 * list.sort(new makagiga.common.StringComparator());
 * ...
 * @endcode
 */
public class MList<I> extends JList implements Iterable<I> {

	// private

	private DefaultListModel model;

	// public

	/**
	 * Constructs a new list with @c DefaultListModel.
	 */
	public MList() {
		model = new DefaultListModel();
		setModel(model);
	}

	/**
	 * Adds an @p item to the list.
	 */
	public void addItem(final I item) {
		model.addElement(item);
	}

	/**
	 * Adds an @p item to the list at @p index.
	 */
	public void addItem(final I item, final int index) {
		model.add(index, item);
	}

	/**
	 * Removes all items.
	 */
	public void clear() {
		model.clear();
	}

	@SuppressWarnings("unchecked")
	public I getItemAt(final int index) {
		return (I)model.getElementAt(index);
	}

	public I getItemAt(final Point p) {
		int index = locationToIndex(p);

		if (index == -1)
			return null;

		Rectangle r = getCellBounds(index, index);

		if (r == null)
			return null;

		if (!r.contains(p))
			return null;

		return getItemAt(index);
	}

	/**
	 * Returns the number of items.
	 */
	public int getItemCount() {
		return model.getSize();
	}

	/**
	 * Returns the first selected item, or @c null if no selection.
	 */
	@SuppressWarnings("unchecked")
	public I getSelectedItem() {
		return (I)getSelectedValue();
	}

	public void setSelectedItem(final I value) {
		setSelectedValue(value, false);
	}

	/**
	 * Returns @c true if list is empty.
	 */
	public boolean isEmpty() {
		return model.isEmpty();
	}

	public Iterator<I> iterator() {
		return new MListIterator();
	}

	public MListIterator listIterator() {
		return (MListIterator)iterator();
	}

	public boolean removeItem(final I element) {
		return model.removeElement(element);
	}

	public boolean removeRange(final int fromIndex) {
		int count = getItemCount();

		if (count == 0)
			return false;

		if (fromIndex > count - 1)
			return false;

		try {
			model.removeRange(fromIndex, count - 1);

			return true;
		}
		catch (ArrayIndexOutOfBoundsException exception) {
			MLogger.exception(exception);

			return false;
		}
		catch (IllegalArgumentException exception) {
			MLogger.exception(exception);

			return false;
		}
	}

	public void replaceItem(final I item, final int index) {
		model.set(index, item);
	}

	/**
	 * Sets "single selection mode".
	 */
	public void setSingleSelectionMode() {
		setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	}

	/**
	 * Sorts list using @p comparator.
	 */
	@SuppressWarnings("unchecked")
	public void sort(final Comparator<I> comparator) {
		int count = getItemCount();

		if (count == 0)
			return;

// !!!: implement Collator in text compares
		ArrayList<I> arrayList = new ArrayList<I>(count);
		I[] array = (I[])model.toArray();
		for (I i : array)
			arrayList.add(i);
		Collections.sort(arrayList, comparator);
		for (int i = 0; i < count; ++i)
			model.set(i, arrayList.get(i));
	}

	// public classes

	public final class MListIterator extends AbstractIterator<I> {

		// public

		public I getObjectAt(final int index) {
			return getItemAt(index);
		}

		public int getObjectCount() {
			return getItemCount();
		}

	}

}
