// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.Component;
import java.awt.Point;
import javax.swing.AbstractButton;
import javax.swing.Icon;
import javax.swing.JMenu;

/** A menu. */
public class MMenu extends JMenu {

	// public

	/**
	 * Constructs a new menu.
	 */
	public MMenu() { }

	/**
	 * Constructs a new menu.
	 * @param text A menu text
	 */
	public MMenu(final String text) {
		this(text, (Icon)null);
	}

	/**
	 * Constructs a new menu.
	 * @param text A menu text
	 * @param icon An icon
	 */
	public MMenu(final String text, final Icon icon) {
		if (icon != null)
			setIcon(icon);
		setText(text);
	}

	/**
	 * Constructs a new menu.
	 * @param text A menu text
	 * @param iconName An icon name
	 */
	public MMenu(final String text, final String iconName) {
		if (!TK.isEmpty(iconName))
			setIcon(FS.getIcon(iconName));
		setText(text);
	}

	/**
	 * Adds an @p action to the menu.
	 * @return A new menu item
	 */
	public MMenuItem add(final AppAction action) {
		return (MMenuItem)super.add(new MMenuItem(action));
	}

	/**
	 * Adds a new item to the menu.
	 * @param text A menu item text
	 * @return A new menu item
	 */
	public MMenuItem add(final String text) {
		return (MMenuItem)super.add(new MMenuItem(text));
	}

	/**
	 * Adds a new check box to the menu.
	 * @param action A "toggle" action
	 */
	public void addCheckBox(final AppToggleAction action) {
		MMenuItem item = add(action);
		item.setCheckBox();
	}

	public void addItemsFromMenu(final MMenu menu) {
		if (menu == null)
			return;

		Component[] components = menu.getMenuComponents();
		for (Component c : components)
			add(c);
	}

	public Icon getComponentIcon() {
		Component c = getComponent();

		if (c instanceof AbstractButton)
			return AbstractButton.class.cast(c).getIcon();

		return null;
	}

	/**
	 * Returns @c true if menu has no items.
	 */
	public boolean isEmpty() {
		return getItemCount() == 0;
	}

	/**
	 * Removes all items from @p index.
	 */
	public void removeFrom(final int index) {
		int count = getItemCount();

		if (count == 0)
			return;

		for (int i = count - 1; i >= index; --i)
			remove(i);
	}

	public void setHTMLHelp(final String value) {
		setToolTipText(UI.makeHTML(value));
	}

	/**
	 * Overriden to handle mnemonics.
	 * @param text A new text (& - mnemonic)
	 */
	@Override
	public void setText(final String text) {
		super.setText(UI.setTextWithMnemonic(this, text));
	}

	public void showPopup(final Component parent) {
		Point mouseLocation = parent.getMousePosition();
		getPopupMenu().show(parent, mouseLocation.x, mouseLocation.y);
	}

	public void showPopup(final Component parent, final int x, final int y) {
		getPopupMenu().show(parent, x, y);
	}

}
