// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static org.makagiga.common.UI._;

import java.awt.Component;
import java.awt.Container;
import java.awt.Window;
import java.text.MessageFormat;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.UIManager;

/** Misc message and input dialogs. */
public final class MMessage {

	// private

	private static boolean initOptionPaneDone;

	// public

	/**
	 * Displays a confirmation box.
	 * @param parent A parent window
	 * @param text A message text
	 * @return @c true if @b OK pressed; otherwise @c false
	 */
	public static boolean confirm(final Window parent, final String text) {
		return confirmWithIcon(parent, "ui/question", text);
	}

	/**
	 * Displays a confirmation box.
	 * @param parent A parent window
	 * @param text A message text
	 * @param args Arguments used to format message
	 * @return @c true if @b OK pressed; otherwise @c false
	 */
	public static boolean confirm(final Window parent, final String text, final Object... args) {
		return confirmWithIcon(parent, "ui/question", MessageFormat.format(text, args));
	}

	public static boolean confirmFileOverwrite(final Window parent, final String path) {
		return confirm(
			parent,
			_("File already exists. Overwrite?\n\n\"{0}\"", path)
		);
	}

	/**
	 * Displays a confirmation box.
	 * @param parent A parent window
	 * @param iconName An icon name
	 * @param text A message text
	 * @return @c true if @b OK pressed; otherwise @c false
	 */
	public static boolean confirmWithIcon(final Window parent, final String iconName, final String text) {
		MSplashScreen.stop(); // hide splash screen

		JOptionPane view = createView(iconName, text);
		Base base = Base.create(
			parent,
			null,
			_("Confirm"),
			null,
			view,
			MDialog.STANDARD_DIALOG
		);
		base.setResizable(false);

		if (base.isOK())
			return true;

		return false;
	}

	public static boolean confirmWithIcon(final Window parent, final String iconName, final String text, final Object... args) {
		return confirmWithIcon(parent, iconName, MessageFormat.format(text, args));
	}

	/**
	 * Displays an error message box.
	 * @param parent A parent window
	 * @param text A message text
	 */
	public static void error(final Window parent, final String text) {
		MSplashScreen.stop(); // hide splash screen

		JOptionPane view = createView("ui/error", text);
		Base base = Base.create(
			parent,
			null,
			_("Error"),
			null,
			view,
			MDialog.OK_BUTTON | MDialog.MODAL
		);
		base.setResizable(false);
		base.exec();
	}

	/**
	 * Displays an error message box.
	 * @param parent A parent window
	 * @param text A message text
	 * @param args Arguments used to format message
	 */
	public static void error(final Window parent, final String text, final Object... args) {
		error(parent, MessageFormat.format(text, args));
	}

	/**
	 * Displays a file error message box.
	 * @param parent A parent window
	 * @param text A message text
	 * @param info A file path or additional info
	 */
	public static void fileError(final Window parent, final String text, final Object info) {
		error(parent, text + "\n\n\"{0}\"", info);
	}

	/**
	 * Displays an information message box.
	 * @param parent A parent window
	 * @param text A message text
	 */
	public static void info(final Window parent, final String text) {
		infoWithIcon(parent, "ui/info", text);
	}

	/**
	 * Displays an information message box.
	 * @param parent A parent window
	 * @param text A message text
	 * @param args Arguments used to format message
	 */
	public static void info(final Window parent, final String text, final Object... args) {
		info(parent, MessageFormat.format(text, args));
	}

	/**
	 * Displays an information message box.
	 * @param parent A parent window
	 * @param iconName An icon name
	 * @param text A message text
	 */
	public static void infoWithIcon(final Window parent, final String iconName, final String text) {
		MSplashScreen.stop(); // hide splash screen

		JOptionPane view = createView(iconName, text);
		Base base = Base.create(
			parent,
			null,
			_("Information"),
			null,
			view,
			MDialog.OK_BUTTON | MDialog.MODAL
		);
		base.setResizable(false);
		base.exec();
	}

	/**
	 * Displays an input box (prompt).
	 * @param parent A parent window
	 * @param initialValue An initial value
	 * @param label A label
	 * @param title A dialog title
	 * @param iconName An icon name
	 * @return Entered text or @c null if cancelled
	 */
	public static String input(final Window parent, final String initialValue, final String label, final String title, final String iconName) {
		MTextField view = new MTextField(initialValue);
		Base base = Base.create(
			parent,
			label,
			title,
			iconName,
			view,
			MDialog.STANDARD_DIALOG
		);
		base.setResizable(false);
		view.setDefault();

		if (base.isOK())
			return view.getText();

		return null;
	}

	/**
	 * Displays an input box (prompt).
	 * @param parent A parent window
	 * @param initialValue An initial value
	 * @param label A label
	 * @param title A dialog title
	 * @return Entered text or @c null if cancelled
	 */
	public static String input(final Window parent, final String initialValue, final String label, final String title) {
		return input(parent, initialValue, label, title, null);
	}

	/**
	 * Displays a text input box (prompt).
	 * @param parent A parent window
	 * @param initialValue An initial value
	 * @param label A label
	 * @param title A dialog title
	 * @param iconName An icon name
	 * @return Entered text or @c null if cancelled
	 */
	public static String text(final Window parent, final String initialValue, final String label, final String title, final String iconName) {
		MTextArea view = new MTextArea(initialValue);
		view.setWrapStyleWord(true);
// TODO: 2.0: wrap lines
// TODO: 2.0: problem with tab keys
		Base base = Base.create(
			parent,
			label,
			title,
			iconName,
			new MScrollPane(view),
			MDialog.STANDARD_DIALOG
		);
		base.setSize(320, 200);

		if (base.isOK())
			return view.getText();

		return null;
	}

	// private

	private static JOptionPane createView(final String iconName, final String text) {
		// init custom icons
		if (!initOptionPaneDone) {
			if (UI.customIcons.get()) {
				UIManager.put("OptionPane.errorIcon", FS.getIcon("ui/error"));
				UIManager.put("OptionPane.informationIcon", FS.getIcon("ui/info"));
				UIManager.put("OptionPane.questionIcon", FS.getIcon("ui/question"));
			}
			initOptionPaneDone = true;
		}

		JOptionPane pane = new JOptionPane(text, JOptionPane.PLAIN_MESSAGE, JOptionPane.DEFAULT_OPTION, FS.getIcon(iconName));
		hideButtons(pane);

		return pane;
	}

	private static void hideButtons(final Container container) {
		Component[] components = container.getComponents();

		if (TK.isEmpty(components))
			return;

		for (Component i : components) {
			if (i instanceof JButton)
				container.remove(i);
			else if (i instanceof Container)
				hideButtons((Container)i);
		}
	}

	// public classes

	public static final class Base extends MDialog {

		// private

		private MLabel label;

		// public

		public static Base create(final Window parent, final String label, final String title, final String iconName, final JComponent view, final int flags) {
			if (parent instanceof MDialog)
				return new Base((MDialog)parent, label, title, iconName, view, flags);

			if (parent instanceof MFrame)
				return new Base((MFrame)parent, label, title, iconName, view, flags);

			return new Base(MFrame.getGlobal(), label, title, iconName, view, flags);
		}

		// private

		private Base(final MDialog parent, final String label, final String title, final String iconName, final JComponent view, final int flags) {
			super(parent, title, flags);
			init(label, iconName, view);
		}

		private Base(final MFrame parent, final String label, final String title, final String iconName, final JComponent view, final int flags) {
			super(parent, title, flags);
			init(label, iconName, view);
		}

		private void init(final String label, final String iconName, final JComponent view) {
			MPanel mainPanel = MPanel.createVBoxPanel();

			// init label
			if (label != null) {
				this.label = new MLabel(label);
				mainPanel.add(this.label);
				mainPanel.addGap();
			}

			// init icon
			if (iconName != null)
				addWest(new MLabel(FS.getIcon(iconName)));

			// init view
			if (view != null) {
				mainPanel.add(view);
				if (this.label != null)
					this.label.setLabelFor(view);
			}

			addCenter(mainPanel);
			pack();
		}

	}

}
