// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Hashtable;
import javax.swing.BorderFactory;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * A slider.
 *
 * EXAMPLE:
 * @code
 * ...
 * MSlider slider = new MSlider();
 * slider.setTitle(_("Brush Size"));
 * slider.addLabel(Brush.MIN_SIZE, _("Small"));
 * slider.addLabel(Brush.MEDIUM_SIZE, _("Medium"));
 * slider.addLabel(Brush.MAX_SIZE, _("Large"));
 * slider.setPaintLabels(true);
 * slider.setMinimum(Brush.MIN_SIZE);
 * slider.setMaximum(Brush.MAX_SIZE);
 * slider.setValue(Brush.DEFAULT_SIZE);
 * ...
 * @endcode
 */
public class MSlider extends JSlider implements MouseWheelListener {

	// private

	private Hashtable<Integer, MLabel> labelTable;

	// public

	/**
	 * Constructs a new slider.
	 */
	public MSlider() {
		addChangeListener(new ChangeListener() {
			public void stateChanged(final ChangeEvent e) {
				onChange();
			}
		} );
		addMouseWheelListener(this);
	}

	/**
	 * Adds a new label with @p text for @p value.
	 */
	public void addLabel(final int value, final String text) {
		if (labelTable == null)
			labelTable = new Hashtable<Integer, MLabel>();
		labelTable.put(value, new MLabel(text));
		setLabelTable(labelTable);
	}

	/**
	 * Mouse wheel event handler.
	 * Increments/decrements slider value.
	 * @param e A mouse wheel event
	 */
	public void mouseWheelMoved(final MouseWheelEvent e) {
		UI.fire(
			(e.getWheelRotation() > 0)
				? "positiveUnitIncrement"
				: "negativeUnitIncrement",
			this
		);
	}

	/**
	 * Sets border title to @p value.
	 */
	public void setTitle(final String value) {
		setBorder(BorderFactory.createTitledBorder(value));
	}

	// protected

	/**
	 * Invoked when value is changed.
	 */
	protected void onChange() { }

}
