// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static java.awt.Color.*;

import static org.makagiga.common.UI._;

import java.awt.Color;

/**
 * A status bar.
 *
 * EXAMPLE:
 * @code
 * ...
 * MStatusBar statusBar = new MStatusBar();
 * statusBar.makeGlobal(); // main status bar
 * MLabel anotherStatusText = new MLabel();
 * anotherStatusText.setBorder(BorderFactory.createLoweredBevelBorder());
 * anotherStatusText.setMnemonicEnabled(false);
 * statusBar.add(anotherStatusText);
 * add(statusBar, BorderLayout.SOUTH);
 * ...
 * MStatusBar.setText("Loading...");
 * ...
 * @endcode
 */
public class MStatusBar extends MPanel {

	// public

	public static final int AUTOHIDE = 1 << 0; /**< Autohide message after the specified time. */

	// private

	private MLabel text;
	private static MStatusBar global;
	private MTimer textTimer;

	// public

	/**
	 * Constructs a new status bar.
	 */
	public MStatusBar() {
		super(NO_LAYOUT);
		setHBoxLayout();

		text = new MLabel();
		text.setMnemonicEnabled(false);
		add(text);
		text.setOpaque(true);

		// -
		addStretch();

// TODO: 2.0: configurable timeout
		textTimer = new MTimer(MTimer.Format.SECONDS, 10) {
			public void onTimeout() {
				ready();
				stop();
			}
		};
	}

	/**
	 * Returns the global status bar.
	 */
	public static MStatusBar getGlobal() { return global; }

	/**
	 * Makes @c this status bar the global status bar.
	 */
	public void makeGlobal() { global = this; }

	/**
	 * Sets text to "".
	 */
	public synchronized static void ready() {
		setText("");
	}

	/**
	 * Displays an error (red) message text.
	 * @param value A message text
	 */
	public synchronized static void setError(final String value) {
		global.setTextData(value, AUTOHIDE, RED);
	}

	/**
	 * Displays a message.
	 * @param value A message text
	 */
	public synchronized static void setText(final String value) {
		global.setTextData(value, 0, null);
	}

	/**
	 * Displays a message.
	 * @param value A message text
	 * @param flags Flags (e.g. @ref AUTOHIDE)
	 */
	public synchronized static void setText(final String value, final int flags) {
		global.setTextData(value, flags, null);
	}

	/**
	 * Displays a warning (yellow) message text.
	 * @param value A message text
	 */
	public synchronized static void setWarning(final String value) {
		global.setTextData(value, AUTOHIDE, YELLOW);
	}

	// private

	private void setTextData(final String value, final int flags, final Color color) {
		if (color == null) {
			text.setBackground(UI.getUIColor("Label.background"));
			text.setForeground(UI.getUIColor("Label.foreground"));
		}
		else {
			text.setBackground(color);
			text.setForeground(BLACK);
		}
		// " " is a workaround for component height problem
		text.setText(TK.isEmpty(value) ? " " : value);
		text.setToolTipText(UI.makeHTML(value, UI.ESCAPE));
		if ((flags & AUTOHIDE) != 0)
			textTimer.restart();
		else
			textTimer.stop();
	}

}
