// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static java.awt.event.KeyEvent.*;

import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Iterator;
import javax.swing.JComponent;
import javax.swing.JTabbedPane;

/** A tabbed pane. */
// TODO: 2.0: add built-in GUI option for tabs placement
public class MTabbedPane<T extends JComponent> extends JTabbedPane
	implements
		Iterable<T>,
		MouseWheelListener
{

	// public

	/**
	 * Constructs a new tabbed pane.
	 */
	public MTabbedPane() {
// FIXME: 2.0: "Ctrl+Tab" and focus manager
		addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(final KeyEvent e) {
				// Ctrl+Shift+Tab - previous tab
				if (e.isControlDown() && e.isShiftDown() && (e.getKeyCode() == VK_TAB)) {
					e.consume();
					selectPreviousTab();
				}
				// Ctrl+Tab - next tab
				else if (e.isControlDown() && (e.getKeyCode() == VK_TAB)) {
					e.consume();
					selectNextTab();
				}
			}
		} );
		addMouseWheelListener(this);
	}

	/**
	 * Adds a new tab.
	 * @param title A title
	 * @param iconName An icon name
	 * @param tab A component
	 */
	public void addTab(final String title, final String iconName, final T tab) {
		addTab(title, FS.getIcon(iconName), tab);
	}

	/**
	 * Adds a @c null tab.
	 * @param title A title
	 * @param iconName An icon name
	 */
	public void addTab(final String title, final String iconName) {
		addTab(title, FS.getIcon(iconName), null);
	}

	/**
	 * Returns the selected tab, or @c null if no selection.
	 */
	@SuppressWarnings("unchecked")
	public T getSelectedTab() {
		T component = (T)getSelectedComponent();

		if (component == null)
			return null;

		return (T)component;
	}

	@SuppressWarnings("unchecked")
	public T getTabAt(final int index) {
		T component;
		try {
			component = (T)getComponentAt(index);

			if (component == null)
				return null;
		}
		catch (IndexOutOfBoundsException exception) {
			return null;
		}

		return (T)component;
	}

	/**
	 * Returns @c true if no tabs.
	 */
	public boolean isEmpty() {
		return getTabCount() == 0;
	}

	public Iterator<T> iterator() {
		return new MTabIterator();
	}

	/**
	 * A mouse wheel event handler.
	 * It selects the next/previous tab.
	 * @param e A mouse wheel event
	 */
	public void mouseWheelMoved(final MouseWheelEvent e) {
		JComponent c = (JComponent)getComponentAt(e.getX(), e.getY());
		if ((c == null) || !(c instanceof MTabbedPane))
			return;

		if (e.getWheelRotation() < 0)
			selectPreviousTab();
		else
			selectNextTab();
	}

	// private

	private void selectNextTab() {
		int count = getTabCount();

		if (count < 2)
			return;

		int current = getSelectedIndex();
		if (current == -1)
			current = 0;

		int index = current + 1;
		while (true) {
			if (index > count - 1)
				index = 0;

			if (index == current)
				return;

			if (isEnabledAt(index))
				break; // while

			index++;
		}
		setSelectedIndex(index);
	}

	private void selectPreviousTab() {
		int count = getTabCount();

		if (count < 2)
			return;

		int current = getSelectedIndex();
		if (current == -1)
			current = 0;

		int index = current - 1;
		while (true) {
			if (index < 0)
				index = count - 1;

			if (index == current)
				return;

			if (isEnabledAt(index))
				break; // while

			index--;
		}
		setSelectedIndex(index);
	}

	public MTabIterator tabIterator() {
		return (MTabIterator)iterator();
	}

	// public classes

	public final class MTabIterator extends AbstractIterator<T> {

		// public

		public T getObjectAt(final int index) {
			return getTabAt(index);
		}

		public int getObjectCount() {
			return getTabCount();
		}

	}

}
