// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.DefaultTableModel;

// TODO: 2.0: double click on the header should autoresize column (like in Qt or OO.org Calc)
/** A table. */
public class MTable extends JTable implements MouseWheelListener {

	// protected

	/**
	 * The default table model.
	 */
	protected DefaultTableModel model;

	// private

	private static Border emptyBorder;
	private static Border focusBorder;

	// public

	/**
	 * Constructs a new table.
	 * Defaults:
	 * - Resize mode: AUTO_RESIZE_LAST_COLUMN
	 * - Inter cell spacing: 1, 1
	 * - Row height: 24
	 */
	public MTable() {
		setAutoResizeMode(AUTO_RESIZE_LAST_COLUMN);
		setIntercellSpacing(new Dimension(1, 1));
		setRowHeight(24);
		model = (DefaultTableModel)getModel();
		addMouseWheelListener(this);
// MUSTANG: setRowSorter(new javax.swing.table.TableRowSorter(model));
	}

	/**
	 * Deletes all items.
	 */
	public void deleteAll() {
		int count = getRowCount();

		if (count == 0)
			return;

		doneEdit();
		try {
			for (int row = count - 1; row >= 0; --row)
				model.removeRow(row);
		}
		catch (ArrayIndexOutOfBoundsException exception) {
			MLogger.exception(exception);
		}
	}

	/**
	 * Deletes all the selected items.
	 */
	public void deleteSelected() {
		doneEdit();
		int selectedRow = getSelectedRow();
		try {
			for (int row = getRowCount() - 1; row >= 0; --row) {
				if (isRowSelected(row))
					model.removeRow(row);
			}
		}
		catch (ArrayIndexOutOfBoundsException exception) {
			MLogger.exception(exception);
		}
		selectRow(selectedRow);
	}

	/**
	 * Stops (or cancels if stop failed) the current cell edition.
	 */
	public void doneEdit() {
		if (isEditing()) {
			DefaultCellEditor dce = (DefaultCellEditor)getCellEditor();
			if (dce != null) {
				if (!dce.stopCellEditing())
					dce.cancelCellEditing();
			}
		}
	}

	/**
	 * Returns @c true if table has no rows.
	 */
	public boolean isEmpty() {
		return getRowCount() == 0;
	}

	/**
	 * Mouse wheel event handler.
	 * Scrolls table vertically or horizontally.
	 * @param e A mouse wheel event
	 * @see org.makagiga.common.MScrollPane.mouseWheelScroll
	 */
	public void mouseWheelMoved(final MouseWheelEvent e) {
		MScrollPane.mouseWheelScroll(this, e);
	}

	/**
	 * Selects row at @p index.
	 */
	public void selectRow(final int index) {
		if (isEmpty())
			return;

		try {
			setRowSelectionInterval(index, index);
			Rectangle r = getCellRect(index, 0, true);
			if (r != null)
				scrollRectToVisible(r);
// !!!: scroll to the last row
		}
		catch (IllegalArgumentException exception) { }
	}

	/**
	 * Sets border of the specified component.
	 * @param renderer A component to setup
	 * @param hasFocus If @c true, the @p component should have the <b>highlighted border</b>
	 */
	public static void setupCellBorder(final JComponent renderer, final boolean hasFocus) {
		if (hasFocus) {
			if (focusBorder == null)
				focusBorder = UIManager.getBorder("Table.focusCellHighlightBorder");
			renderer.setBorder(focusBorder);
		}
		else {
			if (emptyBorder == null)
				emptyBorder = BorderFactory.createEmptyBorder(1, 1, 1, 1);
			renderer.setBorder(emptyBorder);
		}
	}

}
