// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import java.util.Enumeration;

/**
 * A tree scanner.
 *
 * EXAMPLE:
 * @code
 * ...
 * new MTreeScanner(TreeList.getInstance().getDataRoot()) {
 *   public void processItem(final MTreeItem<MetaInfo> item) {
 *     ...
 *   }
 * };
 * ...
 * @endcode
 */
public abstract class MTreeScanner<V extends Object> extends AbstractScanner {

	// public

	public static final int PROCESS_ROOT = 1 << 0; /**< Process the root item. */

	// public

	/**
	 * Constructs a new tree scanner.
	 * The scan will begin automatically from the @p root item.
	 * The @p root item is not processed.
	 * @param root A tree root item
	 */
	public MTreeScanner(final MTreeItem<V> root) {
		this(root, 0);
	}

	/**
	 * Constructs a new tree scanner.
	 * The scan will begin automatically from the @p root item.
	 * @param root A tree root item
	 * @param flags Flags (e.g. @ref PROCESS_ROOT)
	 */
	public MTreeScanner(final MTreeItem<V> root, final int flags) {
		if (root != null) {
			if ((flags & PROCESS_ROOT) != 0)
				processItem(root);
			recurse(root);
		}
	}

	/**
	 * Called for each tree list item.
	 * @param item A tree list item
	 */
	public abstract void processItem(final MTreeItem<V> item);

	// private

	@SuppressWarnings("unchecked")
	private void recurse(final MTreeItem item) {
		if (isStopped() || (item.getChildCount() == 0))
			return;

		MTreeItem i;
		for (Enumeration<MTreeItem> e = item.children(); e.hasMoreElements();) {
			i = e.nextElement();
			if (i != null) {
				processItem(i);

				if (isStopped())
					return;

				recurse(i);

				if (isStopped())
					return;
			}
		}
	}

}
