// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static org.makagiga.common.UI._;

import java.io.IOException;
import java.net.URL;

/** Wrapper for OS dependent functions. */
public final class OS {

	// public

	/**
	 * The custom browser, or @c null if no custom browser.
	 */
	public static final StringProperty customBrowser = new StringProperty();

	// private

	// apps
	private static float javaVersionNumber;
	private static String internalName;
	private static String name;
	private static String version;

	// os
	private static boolean linux;
	private static boolean mac;
	private static boolean windows;
	private static boolean windowsSupported;

	// session
	private static boolean kde;

	// misc
	private static Thread shutDownHook;

	// public

	/**
	 * Returns the OS internal name. (e.g. "linux")
	 * This name is in lowercase.
	 */
	public static String getInternalName() { return internalName; }

	/**
	 * Returns the OS name. (e.g. "Linux")
	 */
	public static String getName() { return name; }

	/**
	 * Returns the OS version. (e.g. "2.6.10")
	 */
	public static String getVersion() { return version; }

	/**
	 * Initializes the OS class.
	 * You MUST call this function at application startup.
	 *
	 * EXAMPLE:
	 * @code
	 * ...
	 * public static final void main(String[] args) {
	 *   OS.init();
	 * ...
	 * @endcode
	 */
	public static void init() {
		String javaVersion = System.getProperty("java.version");
		javaVersionNumber = Float.valueOf(javaVersion.substring(0, javaVersion.lastIndexOf(".")));
		if (!isJavaTiger()) {
			MLogger.warning("Your Java is too old (version %s).\nThis application requires Java Runtime Environment compatible with version (1.)5.0.\nWARNING: The application will crash.", javaVersion);
		}

		name = System.getProperty("os.name");
		version = System.getProperty("os.version");
		// detect os
		String s = name.toUpperCase();
		linux = s.startsWith("LINUX");
		mac = s.startsWith("MAC");
		windows = s.startsWith("WINDOWS");

		if (linux) {
			internalName = "linux";
			// detect session type
			s = System.getenv("DESKTOP_SESSION");
			kde = (s != null) && s.equalsIgnoreCase("KDE");
		}
		else if (mac) {
			internalName = "mac";
		}
		else if (windows) {
			internalName = "windows";
			float windowsVersionNumber = Float.valueOf(version);
			windowsSupported = windowsVersionNumber >= 5.0;
		}
		else {
			internalName = "unknown";
		}
	}

	/**
	 * Returns @c true if Java version is 6.0 or later.
	 */
	public static boolean isJavaMustang() { return javaVersionNumber >= 1.6; }

	/**
	 * Returns @c true if Java version is 5.0 or later.
	 */
	public static boolean isJavaTiger() { return javaVersionNumber >= 1.5; }

// !!!: OS.isGNOME
// MUSTANG: use "GTK+" LAF by default on GNOME
	/**
	 * Returns @c true if KDE is a current session.
	 */
	public static boolean isKDE() { return kde; }

	/**
	 * Returns @c true if application is running under Mac OS.
	 */
	public static boolean isMac() { return mac; }

	/**
	 * Returns @c true if application is running under Windows OS.
	 */
	public static boolean isWindows() { return windows; }

	/**
	 * Returns @c true if application is running under supported Windows OS
	 * (2000, XP, 2003 or newer).
	 */
	public static boolean isWindowsSupported() { return windowsSupported; }

// TODO: 2.0: test & improve: mac, solaris
	/**
	 * Launches the WWW browser with the specified URL.
	 *
	 * EXAMPLE:
	 * @code
	 * ...
	 * OS.launchBrowser("http://www.google.com");
	 * ...
	 * @endcode
	 * @param url An URL to open
	 */
	public static boolean launchBrowser(final String url) {
		// try custom browser
		if (!customBrowser.isEmpty()) {
			if (windows) {
				// windows xp
				if (windowsSupported) {
					if (tryLaunchBrowser(
						"cmd.exe",
						"/C",
						"start",
						"\"\"", // no title :/
						"\"" + customBrowser + "\"",
						"\"" + url + "\""
					))
						return true;
				}
				// windows 9x
				else {
// TODO: 2.0: test "start.exe"
					if (tryLaunchBrowser(
						"start.exe",
						"\"" + customBrowser + "\"",
						"\"" + url + "\""
					))
						return true;
				}
			}
			else {
		 		if (tryLaunchBrowser(customBrowser.get(), url))
					return true;
			}
		}

		// linux
		if (linux) {
			if (kde && tryLaunchBrowser("konqueror", url))
				return true;
		}
		// windows
		else if (windows) {
			// windows xp
			if (windowsSupported) {
				if (tryLaunchBrowser(
					"cmd.exe",
					"/C",
					"start",
					"\"\"", // no title :/
					"\"" + url + "\""
				))
					return true;
			}
			// windows 9x
			else {
				if (tryLaunchBrowser("start.exe", "\"" + url + "\""))
					return true;
			}
		}

		// this should work on most systems

		if (tryLaunchBrowser("firefox", url))
			return true;

		if (tryLaunchBrowser("mozilla", url))
			return true;

		if (tryLaunchBrowser("netscape", url))
			return true;

		if (tryLaunchBrowser("opera", url))
			return true;

		// error ..
		MMessage.fileError(
			null,
			_("Could not launch browser"),
			url
		);

		return false;
	}

	/**
	 * Launches the WWW browser with the specified URL.
	 * @param url An URL to open
	 */
	public static boolean launchBrowser(final URL url) {
		if (url == null)
			return false;

		return launchBrowser(url.toString());
	}

	/**
	 * Removes the shut down hook if any.
	 * Call this function before application exit.
	 * @see setShutDownHook
	 */
	public static void removeShutDownHook() {
		Runtime.getRuntime().removeShutdownHook(shutDownHook);
		shutDownHook = null;
	}

	/**
	 * Sets the shut down hook to @p thread.
	 * Call this function at application startup.
	 * @param thread A thread to execute when application is being terminated
	 * @see removeShutDownHook
	 * @see java.lang.Runtime.addShutdownHook
	 */
	public static void setShutDownHook(final Thread thread) {
		if (shutDownHook != null)
			removeShutDownHook();
		shutDownHook = thread;
		Runtime.getRuntime().addShutdownHook(shutDownHook);
	}

	// private

	private static boolean tryLaunchBrowser(final String... args) {
		MLogger.info("Launching: \"%s\"", args[0]);
		try {
			ProcessBuilder pb = new ProcessBuilder(args);
			Process p = pb.start();

			if (p == null)
				return false;

			// full i/o buffer blocks launched app
			FS.close(p.getErrorStream());
			FS.close(p.getInputStream());
			FS.close(p.getOutputStream());

			return true;
		}
		catch (IOException exception) {
			MLogger.exception(exception);

			return false;
		}
	}

}
