// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import static java.awt.Color.*;
import static java.awt.Cursor.*;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.text.MessageFormat;
import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.plaf.metal.MetalLookAndFeel;

/** User interface wrapper. */
public final class UI {

	// public

	public static final BooleanProperty customIcons = new BooleanProperty(true);
	public static final BooleanProperty toolTipsDetails = new BooleanProperty();

	public static final char TM = '\u2122';

	public static final Color TOOL_TIP_BACKGROUND_COLOR = new Color(0xffffdc); /**< The tool tip text background color. */
	public static final Color TOOL_TIP_FOREGROUND_COLOR = BLACK; /**< The tool tip text foreground color. */

	// makeHTML flags
	public static final int ESCAPE = 1 << 0; /**< Escape text. */

	public static final int WHEN_ANCESTOR_OF_FOCUSED_COMPONENT = JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT; /**< See @c JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT. */
	public static final int WHEN_FOCUSED = JComponent.WHEN_FOCUSED; /**< See @c JComponent.WHEN_FOCUSED. */
	public static final int WHEN_IN_FOCUSED_WINDOW = JComponent.WHEN_IN_FOCUSED_WINDOW; /**< See @c JComponent.WHEN_IN_FOCUSED_WINDOW. */

	/**
	 * HTML tags used internally.
	 */
	public static final String HTML_BEGIN = "<html><body style=\"margin: 4px\">";

	/**
	 * HTML tags used internally.
	 */
	public static final String HTML_END = "</body></html>";

	// private

	private static int metal = -1; // -1 = not initialized, 1 = metal theme, 0 = not metal theme

	// public

	public static String _(final String id) {
		return Gettext.translate(id);
	}

	public static String _(final String id, final Object... args) {
		return MessageFormat.format(Gettext.translate(id), args);
	}

	public static void addCenter(final Container parent, final JComponent component) {
		if (
			(component instanceof Scrollable) &&
			!(component instanceof MScrollPane) &&
			!(component instanceof MTextField)
		)
			parent.add(new MScrollPane(component), BorderLayout.CENTER);
		else
			parent.add(component, BorderLayout.CENTER);
	}

	/**
	 * Centers a window on the screen.
	 * @param w A window to center.
	 */
	public static void centerOnScreen(final Window w) {
		if (w == null)
			return;

		w.setLocationByPlatform(true);
		w.setLocationRelativeTo(null); // center
	}

	/**
	 * Connects component with the specified action.
	 * @param component A component to connect with action
	 * @param condition A condition which must be complete to perform the @p action (e.g. @c UI.WHEN_FOCUSED)
	 * @param keyCode A shortcut associated with component-action
	 * @param action An action to connect with component
	 */
	public static void connect(
		final JComponent component,
		final int condition,
		final int keyCode,
		final Action action
	) {
		connect(component, condition, keyCode, 0, action);
	}

	/**
	 * Connects component with the specified action.
	 * @param component A component to connect with action
	 * @param condition A condition which must be complete to perform the @p action (e.g. @c UI.WHEN_FOCUSED)
	 * @param keyCode A shortcut associated with component-action
	 * @param modifiers A shortcut modifiers associated with component-action
	 * @param action An action to connect with component
	 */
	public static void connect(
		final JComponent component,
		final int condition,
		final int keyCode,
		final int modifiers,
		final Action action
	) {
		Object name = action.getValue(Action.NAME);
		String actionName = (name == null) ? null : name.toString();
		component.getActionMap().put(actionName, action);
		component.getInputMap(condition).put(KeyStroke.getKeyStroke(keyCode, modifiers), actionName);
	}

	/**
	 * Copies content of the @p source image to the @p target image.
	 */
	public static void copyImage(final Image source, final BufferedImage target) {
		Graphics g = target.createGraphics();
		g.drawImage(source, 0, 0, null, null);
		g.dispose();
	}

	/**
	 * Creates screenshot of the component.
	 * @param component A component
	 * @param format An image format (e.g. "png")
	 * @param path A file name where to save the screenshot
	 */
	public static void createScreenshot(final Component component, final String format, final String path) throws IOException {
		BufferedImage image = new BufferedImage(
			component.getWidth(),
			component.getHeight(),
			BufferedImage.TYPE_INT_ARGB
		);
		component.paint(image.getGraphics());
		ImageIO.write(image, format, new File(path));
	}

	/**
	 * Fires an action.
	 * @param action An action to fire
	 * @param source A source component
	 */
	public static void fire(final Action action, final JComponent source) {
		if (action == null)
			return;

		if (source == null)
			return;

		if (!action.isEnabled())
			return;

		ActionEvent e = new ActionEvent(source, ActionEvent.ACTION_PERFORMED, "");
		action.actionPerformed(e);
	}

	/**
	 * Fires an action.
	 * @param actionName An action name to fire
	 * @param source A source component
	 */
	public static void fire(final String actionName, final JComponent source) {
		if (TK.isEmpty(actionName))
			return;

		if (source == null)
			return;

		fire(source.getActionMap().get(actionName), source);
	}

	/**
	 * Returns mnemonic character extracted from the text, or @c 0 if no mnemonic.
	 * @param text A text with mnemonic
	 * @see removeMnemonic
	 */
	public static char getMnemonic(final String text) {
		if (text == null)
			return 0;

		int len = text.length();

		if (len == 0)
			return 0;

// !!!: double "&" is partially implemented; see setTextWithMnemonic and removeMnemonic
		boolean wasDoubleAmp = false;
		char[] chars = text.toCharArray();
		char c;
		int ampIndex = -2;
		for (int i = 0; i < len; i++) {
			if (chars[i] == '&') {
				if (ampIndex == i - 1) {
					wasDoubleAmp = true;
					ampIndex = -2;
				}
				else {
					ampIndex = i;
				}
			}
			else {
				if (!wasDoubleAmp && (ampIndex != -2))
					return chars[i];
			}
		}

		return (ampIndex == -2) ? 0 : chars[ampIndex];
	}

	public static Color getUIColor(final String key) {
		return (Color)UIManager.get(key);
	}

	/**
	 * Returns a "xor-ed" color value (rgb ^ 0xffffff).
	 * @param color A color value to "xor"
	 */
	public static Color getXorColor(final Color color) {
		if (color == null)
			return null;

		return new Color(color.getRGB() ^ 0xffffff);
	}

	public static boolean isMetal() {
		if (metal == -1) {
			if (UIManager.getLookAndFeel() instanceof MetalLookAndFeel)
				metal = 1;
			else
				metal = 0;
		}

		return (metal == 1);
	}

	/**
	 * Returns @ref HTML_BEGIN + @p text + @ref HTML_END.
	 * @param text A text to convert to HTML
	 * @param flags Flags (e.g. @ref ESCAPE)
	 */
	public static String makeHTML(final String text, final int flags) {
		String t =
			((flags & ESCAPE) != 0)
			? TK.escapeXML(text)
			: text;
		return HTML_BEGIN + t + HTML_END;
	}

	/**
	 * Returns @ref HTML_BEGIN + @p text + @ref HTML_END.
	 */
	public static String makeHTML(final String text) {
		return makeHTML(text, 0);
	}

	/**
	 * Returns the text without the first <code>&amp;</code> character.
	 * @param text A text with mnemonic
	 * @see getMnemonic
	 */
	public static String removeMnemonic(final String text) {
		int i = text.indexOf("&");

		if (i == -1)
			return text;

		return text.substring(0, i) + text.substring(i + 1);
	}

	public static void setHandCursor(final Component component, final boolean yes) {
		component.setCursor(getPredefinedCursor(yes ? HAND_CURSOR : DEFAULT_CURSOR));
	}

	public static String setTextWithMnemonic(final AbstractButton button, final String text) {
		char mnemonic = getMnemonic(text);
		if (mnemonic != 0) {
			button.setMnemonic(mnemonic);

			return removeMnemonic(text);//!!!!
		}

		return text;
	}

	public static String setTextWithMnemonic(final MLabel label, final String text) {//!!FB circular ref. to MLabel
		char mnemonic = getMnemonic(text);
		if (mnemonic != 0) {
			label.setDisplayedMnemonic(mnemonic);

			return removeMnemonic(text);
		}

		return text;
	}

	public static void setWaitCursor(final Component component, final boolean yes) {
		component.setCursor(getPredefinedCursor(yes ? WAIT_CURSOR : DEFAULT_CURSOR));
	}

}
