// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.common;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/** An SAX2 event handler. */
public class XMLHandler extends DefaultHandler {

	// private

	private Attributes attrs; // null by default
	private StringBuilder chars;

	// public

	/**
	 * Receive notification of character data inside an element.
	 * @see getChars
	 */
	@Override
	public void characters(final char[] chars, final int start, final int length) throws SAXException {
		if (length > 0) {
			if (this.chars == null)
				this.chars = new StringBuilder(1024);
			this.chars.append(chars, start, length);
		}
	}

	/**
	 * Receive notification of the end of an element.
	 * @param uri A namespace URI, or the empty string
	 * @param localName A local name (without prefix)
	 * @param qName A qualified name (with prefix)
	 * @see onEnd
	 */
	@Override
	public void endElement(final String uri, final String localName, final String qName) throws SAXException {
		onEnd(localName);
	}

	/**
	 * Returns a string buffered by the @ref characters function.
	 * The string may be empty.
	 */
	public String getChars() {
		return (chars == null) ? "" : chars.toString();
	}

	/**
	 * Receive notification of the end of an element.
	 * @param name A local name (without prefix)
	 */
	public void onEnd(final String name) throws SAXException { }

	/**
	 * Receive notification of the start of an element.
	 * @param name A local name (without prefix)
	 */
	public void onStart(final String name) throws SAXException { }

	/**
	 * Receive notification of the start of an element.
	 * @param uri A namespace URI, or the empty string
	 * @param localName A local name (without prefix)
	 * @param qName A qualified name (with prefix)
	 * @param attrs Attributes attached to the element
	 * @see onStart
	 */
	@Override
	public void startElement(final String uri, final String localName, final String qName, final Attributes attrs) throws SAXException {
		this.attrs = attrs;
		try {
			onStart(localName);
		}
		finally {
			this.attrs = null;
			chars = null;
		}
	}

	// protected

	/**
	 * Returns an @c int attribute value.
	 * @param name An attribute name
	 * @param defaultValue A default value
	 */
	protected int getIntegerValue(final String name, final int defaultValue) {
		try {
			return Integer.parseInt(attrs.getValue(name));
		}
		catch (NumberFormatException exception) {
			return defaultValue;
		}
	}

	/**
	 * Returns a @c long attribute value.
	 * @param name An attribute name
	 * @param defaultValue A default value
	 */
	protected long getLongValue(final String name, final long defaultValue) {
		try {
			return Long.parseLong(attrs.getValue(name));
		}
		catch (NumberFormatException exception) {
			return defaultValue;
		}
	}

	/**
	 * Returns a @c String attribute value.
	 * The default value is @c null.
	 * @param name An attribute name
	 */
	protected String getStringValue(final String name) {
		return attrs.getValue(name);
	}

}
