// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.editors;

import javax.swing.JComponent;

import org.makagiga.common.*;

/**
 * An editor action.
 * All editor actions are disabled by default.
 * Use @ref Editor.updateActions to enable/disable actions.
 *
 * EXAMPLE:
 * @code
 * ...
 * private static final class CutAction extends EditorAction&lt;TextEditor,TextEditorCore&gt; {
 *   CutAction() {
 *     super(_("Cu&amp;t"), "ui/cut", KeyEvent.VK_X, CTRL_MASK);
 *     setHTMLHelp(_("Cuts the selected text to the clipboard."));
 *   }
 *
 *   public void onAction(final TextEditor editor, final TextEditorCore core) {
 *     core.cut();
 *   }
 * }
 * ...
 * @endcode
 */
public abstract class EditorAction<E extends Editor, C extends JComponent> extends AppAction {

	// public

	/**
	 * @see org.makagiga.common.AppAction::AppAction(final String, final String, final int, final int)
	 */
	public EditorAction(final String text, final String iconName, final int keyCode, final int modifiers) {
		super(text, iconName, keyCode, modifiers);
		setEnabled(false);
	}

// !!!: Doxygen: @copydoc
	/**
	 * @see org.makagiga.common.AppAction::AppAction(final String, final String, final int)
	 */
	public EditorAction(final String text, final String iconName, final int keyCode) {
		this(text, iconName, keyCode, 0);
	}

	/**
	 * Invoked when an action occurs.
	 * This function calls @ref onAction.
	 */
	@SuppressWarnings("unchecked")
	public void onAction() {
		if (!isEnabled())
			return;

		Editor editor = Editor.currentEditor.get();
		if (editor != null)
			onAction((E)editor, (C)editor.getCore());
	}

	/**
	 * Invoked when an editor action occurs.
	 * This function is called from @ref onAction.
	 * @param editor The active editor (&lt;&gt; @c null)
	 * @param core The active editor core (&lt;&gt; @c null)
	 */
	public abstract void onAction(final E editor, final C core);

}
