// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.editors.image;

import static java.awt.BasicStroke.*;
import static java.awt.Color.*;
import static java.awt.RenderingHints.*;

import java.awt.BasicStroke;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.BufferedImage;

import org.makagiga.common.*;

/** The image editor brush. */
public final class Brush {

	// public

	// sizes
	public static final int DEFAULT_SIZE = 3; /**< Default brush size. */
	public static final int MAX_SIZE = 50; /**< Max. brush size. */
	public static final int MIN_SIZE = 1; /**< Min. brush size. */
	public static final int MEDIUM_SIZE = (MAX_SIZE - MIN_SIZE) / 2; /**< The medium brush size. */

	/**
	 * The brush color.
	 */
	public final ColorProperty color = new ColorProperty(RED);

	// shapes
	private enum Shapes { CIRCLE, SQUARE };

	// private

	private int size = DEFAULT_SIZE;
	private Shapes type = Shapes.CIRCLE;

	// public

	/**
	 * Draws a line with current brush size and color.
	 * @param g A graphics (2D)
	 * @param x1 Left
	 * @param y1 Top
	 * @param x2 Right
	 * @param y2 Bottom
	 */
	public void drawLine(
		final Graphics2D g,
		final int x1,
		final int y1,
		final int x2,
		final int y2
	) {
		setupGraphics(g);
		g.setColor(color.get());
		if (type == Shapes.CIRCLE)
			g.setStroke(new BasicStroke(size, CAP_ROUND, JOIN_ROUND));
		else
			g.setStroke(new BasicStroke(size, CAP_SQUARE, JOIN_ROUND));
		g.drawLine(x1, y1, x2, y2);
	}

	/**
	 * Draws a single point with the specified size.
	 * @param g A graphics (2D)
	 * @param x An @b X position
	 * @param y An @b Y position
	 * @param size A point size
	 */
	public void drawPoint(final Graphics2D g, int x, int y, final int size) {
		setupGraphics(g);
		int offset = size / 2;
		x -= offset;
		y -= offset;
		g.setColor(color.get());
		switch (type) {
			case CIRCLE:
				g.fillOval(x, y, size, size);
				break;
			case SQUARE:
				g.fillRect(x, y, size, size);
				break;
		}
	}

	/**
	 * Draws a single point with current brush size and color.
	 * @param g A graphics (2D)
	 * @param x An @b X position
	 * @param y An @b Y position
	 */
	public void drawPoint(final Graphics2D g, final int x, final int y) {
		drawPoint(g, x, y, size);
	}

	/**
	 * Returns the brush preview image.
	 */
	public Image getPreview(final int maxSize, final boolean drawCrosshair) {
		BufferedImage image = new BufferedImage(
			maxSize,
			maxSize,
			BufferedImage.TYPE_INT_ARGB
		);
		Graphics2D g = image.createGraphics();
		int hotSpot = maxSize / 2;
		drawPoint(g, hotSpot, hotSpot, maxSize);

		if (drawCrosshair) {
			g.setColor(UI.getXorColor(color.get()));
			g.drawLine(hotSpot, 0, hotSpot, maxSize - 1); // vertical
			g.drawLine(0, hotSpot, maxSize - 1, hotSpot); // horizontal
		}

		g.dispose();

		return image;
	}

	/**
	 * Returns the brush size.
	 */
	public int getSize() { return size; }

	/**
	 * Sets brush size to @p value.
	 */
	public void setSize(final int value) {
		if (value < MIN_SIZE)
			size = MIN_SIZE;
		else if (value > MAX_SIZE)
			size = MAX_SIZE;
		else
			size = value;
	}

	// private

	private void setupGraphics(final Graphics2D g) {
		g.setRenderingHint(KEY_ANTIALIASING, VALUE_ANTIALIAS_ON);
		g.setRenderingHint(KEY_RENDERING, VALUE_RENDER_QUALITY);
	}

}
