// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.editors.image;

import static java.awt.Color.*;
import static java.awt.Cursor.*;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.util.EmptyStackException;
import java.util.Stack;
import javax.imageio.ImageIO;
import javax.swing.JComponent;

import org.makagiga.common.*;
import org.makagiga.editors.Editor;

/** An image editor core. */
public final class ImageEditorCore extends JComponent
	implements
		MouseListener,
		MouseMotionListener,
		MouseWheelListener
{

	// public

	public final BooleanProperty readOnly = new BooleanProperty();

	// private

	private boolean exportMode;
	private boolean lineDrawing;
	private Brush brush = new Brush();
	private BufferedImage canvas;
	private Editor editor;
	private int oldX = -1;
	private int oldY = -1;
	private static final int TYPE = BufferedImage.TYPE_INT_ARGB;
	private static final int UNDO_LIMIT = 5;
	private Stack<UndoItem> undoStack = new Stack<UndoItem>();

	// public

	/**
	 * Constructs a new image editor core.
	 * @param editor An editor
	 */
	public ImageEditorCore(final Editor editor) {
		this.editor = editor;
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
	}

	/**
	 * Returns @c true if undo action is available.
	 */
	public boolean canUndo() {
		return !editor.isReadOnly() && !undoStack.isEmpty();
	}

	/**
	 * Clears the canvas.
	 */
	public void clear() {
		beforeModification();

		canvas = new BufferedImage(canvas.getWidth(), canvas.getHeight(), TYPE);

		afterModification();
	}

	/**
	 * Returns the current brush.
	 */
	public Brush getBrush() { return brush; }

	/**
	 * Returns the canvas.
	 */
	public BufferedImage getCanvas() { return canvas; }

	/**
	 * Loads image from @p path.
	 * @return @c true if successful; otherwise @c false
	 */
	public void loadFromFile(final String path) throws Exception {
		BufferedImage bi = ImageIO.read(new File(path));
		int w, h;
		// create an empty canvas
		if (bi == null) {
			w = 800;
			h = 600;
			canvas = new BufferedImage(w, h, TYPE);
		}
		// copy read image to the canvas
		else {
			w = bi.getWidth();
			h = bi.getHeight();
			canvas = new BufferedImage(w, h, TYPE);
			UI.copyImage(bi, canvas);
		}
		setPreferredSize(new Dimension(w, h));
	}

	/**
	 * Saves image to @p path.
	 * Image is saved in the @b PNG format.
	 */
	public void saveToFile(final String path) throws Exception {
		saveToFileAs("png", path);
	}

	/**
	 * Saves image.
	 * @param format A target image format (e.g. "png")
	 * @param path A destination file name
	 */
	public void saveToFileAs(final String format, final String path) throws Exception {
		if (!exportMode) {
			ImageIO.write(canvas, format, new File(path));
		}
		else {
			int type = TYPE; // png
			if (format.equals("bmp"))
				type = BufferedImage.TYPE_3BYTE_BGR;
			else if (format.equals("jpg"))
				type = BufferedImage.TYPE_INT_RGB;
			BufferedImage bi = new BufferedImage(canvas.getWidth(), canvas.getHeight(), type);
			Graphics g = bi.createGraphics();
			// draw background
			g.setColor(getBackground());
			g.fillRect(0, 0, bi.getWidth(), bi.getHeight());
			// draw image
			g.drawImage(canvas, 0, 0, null, null);
			g.dispose();
			ImageIO.write(bi, format, new File(path));
		}
	}

	/**
	 * Paints image.
	 * @param g A graphics
	 */
	@Override
	public void paint(final Graphics g) {
		int w = getWidth();
		int h = getHeight();

		// draw background
		g.setColor(getBackground());
		g.fillRect(0, 0, w, h);

		if (!exportMode) {
			w = canvas.getWidth();
			h = canvas.getHeight();
			// draw h lines (grid)
			if (ImageEditor.getShowGrid()) {
				g.setColor(LIGHT_GRAY);
				for (int y = 0; y < h; y++) {
					if (y % 15 == 0) {
						g.drawLine(0, y, w, y);
					}
				}
			}

/*
			// draw margin
			g.setColor(BLUE);
			g.drawLine(0, h, w, canvas.getHeight());
			g.drawLine(w, 0, w, h);
*/
		}

		// draw canvas
		g.drawImage(canvas, 0, 0, null);
	}

	/**
	 * Sets "export mode" to @p value.
	 */
	public void setExportMode(final boolean value) {
		exportMode = value;
	}

	/**
	 * Undoes the last operation.
	 */
	public void undo() {
		try {
			UndoItem undoItem = undoStack.pop();
			canvas.setData(undoItem.raster);
			afterModification();
		}
		catch (EmptyStackException exception) {
			MLogger.exception(exception);
		}
	}

	/**
	 * Updates the mouse cursor.
	 */
	public void updateCursor() {
		if (readOnly.get()) {
			setCursor(getPredefinedCursor(DEFAULT_CURSOR));

			return;
		}

		int size = brush.getSize();
// FIXME: 2.0: win32 brush cursor
		if ((size < 5) || OS.isWindows()) {
			setCursor(getPredefinedCursor(CROSSHAIR_CURSOR));

			return;
		}

		Toolkit tk = Toolkit.getDefaultToolkit();
		Dimension d = tk.getBestCursorSize(size, size);

		if ((d.width == 0) || (d.height == 0))
			return;

		Cursor cursor = tk.createCustomCursor(
			brush.getPreview(d.width, true),
			new Point(d.width / 2, d.width / 2),
			"Brush Cursor"
		);
		setCursor(cursor);
	}

	// MouseX methods

	public void mouseClicked(final MouseEvent e) {
		editor.updateStatusText();
	}

	public void mouseDragged(final MouseEvent e) {
		if (!MMouseAdapter.isLeft(e))
			return;

		if (readOnly.get())
			return;

		if (!lineDrawing) {
			setLineDrawing(true, e.getX(), e.getY());
			beforeModification();
		}
		drawLine(e.getX(), e.getY());
		editor.updateStatusText();

		// auto scroll
		MScrollPane.autoScroll(this, e.getX(), e.getY());
	}

	public void mouseEntered(final MouseEvent e) { }

	public void mouseExited(final MouseEvent e) { }

	public void mouseMoved(final MouseEvent e) {
		editor.updateStatusText();
	}

	public void mousePressed(final MouseEvent e) {
		editor.showLockInfo();
	}

	public void mouseReleased(final MouseEvent e) {
		if (!MMouseAdapter.isLeft(e))
			return;

		if (readOnly.get())
			return;

		if (lineDrawing) {
			setLineDrawing(false, -1, -1);
			afterModification();
		}
		else {
			drawPoint(e.getX(), e.getY());
		}
	}

	public void mouseWheelMoved(final MouseWheelEvent e) {
		MScrollPane.mouseWheelScroll(this, e);
	}

	// private

	private void afterModification() {
		repaint();
		editor.modified.yes();
		editor.updateActions();
	}

	private void beforeModification() {
		// save undo image
		UndoItem undoItem = new UndoItem();
// !!!: slow "copyData"
		undoItem.raster = canvas.copyData(null);
		if (undoStack.size() == UNDO_LIMIT)
			undoStack.removeElementAt(0);
		undoStack.push(undoItem);
	}

	private void drawLine(final int x, final int y) {
		if (readOnly.get())
			return;

		brush.drawLine((Graphics2D)canvas.getGraphics(), oldX, oldY, x, y);
		oldX = x;
		oldY = y;

		afterModification();
	}

	private void drawPoint(final int x, final int y) {
		if (readOnly.get())
			return;

		beforeModification();

		brush.drawPoint((Graphics2D)canvas.getGraphics(), x, y);

		afterModification();
	}

	private void setLineDrawing(final boolean yes, final int x, final int y) {
		if (lineDrawing == yes)
			return;

		lineDrawing = yes;
		oldX = x;
		oldY = y;
	}

	// private classes

	private static final class UndoItem {//1!!

		// private

		private WritableRaster raster;

	}

}
