// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.editors.image;

import static org.makagiga.common.UI._;

import java.awt.Color;
import java.awt.FlowLayout;
import javax.swing.ImageIcon;

import org.makagiga.common.*;
import org.makagiga.editors.Designer;
import org.makagiga.editors.Editor;

/** An image editor designer. */
public final class ImageEditorDesigner extends Designer {

	// private

	private static ImageEditorDesigner _instance;
	private MColorButton brushColor;
	private MColorPicker colorPicker;
	private MSlider brushSize;

	// public

	/**
	 * Constructs and returns the instance of the designer.
	 */
	public static ImageEditorDesigner getInstance() {
		if (_instance == null)
			_instance = new ImageEditorDesigner();

		return _instance;
	}

	/**
	 * Sets brush values.
	 * The @p brush is the active editor brush.
	 * @param core An image editor core
	 * @param brush A brush to setup
	 */
	public void setBrush(final ImageEditorCore core, final Brush brush) {
		MLogger.XXX("ImageEditorCore.setBrush()");
		brushColor.setColor(brush.color.get());
		brushSize.setValue(brush.getSize());
		updateCursor(core, true);
	}

	/**
	 * If @p value is @c true, the designer is switched into read only mode.
	 */
	@Override
	public void setReadOnly(final boolean value) {
		super.setReadOnly(value);
		brushColor.setEnabled(!value);
		brushSize.setEnabled(!value);
		colorPicker.setEnabled(!value);
	}

	// private

	private ImageEditorDesigner() {
		setLayout(new FlowLayout());

		// brush size
		brushSize = new MSlider() {
			@Override
			protected void onChange() {
				if (!(Editor.getCurrentCore() instanceof ImageEditorCore))
					return;

				ImageEditorCore core = (ImageEditorCore)Editor.getCurrentCore();
				if (core != null) {
					core.getBrush().setSize(brushSize.getValue());
					updateCursor(core, !brushSize.getValueIsAdjusting());
				}
			}
		};
		brushSize.setTitle(_("Brush Size"));
		brushSize.addLabel(Brush.MIN_SIZE, _("Small"));
		brushSize.addLabel(Brush.MEDIUM_SIZE, _("Medium"));
		brushSize.addLabel(Brush.MAX_SIZE, _("Large"));
		brushSize.setPaintLabels(true);
		brushSize.setMinimum(Brush.MIN_SIZE);
		brushSize.setMaximum(Brush.MAX_SIZE);
		brushSize.setValue(Brush.DEFAULT_SIZE);
		add(brushSize);

		// brush color
		MPanel brushColorPanel = MPanel.createGridPanel(2);
		brushColorPanel.setTitle(_("Brush Color"));
		// color button
		brushColor = new MColorButton() {
			@Override
			protected void colorChanged() {
				ImageEditorCore core = (ImageEditorCore)Editor.getCurrentCore();
				if (core != null) {
					core.getBrush().color.set(getColor());
					updateCursor(core, true);
				}
			}
		};
		brushColor.updateBackground.no();
		brushColorPanel.add(brushColor);
		// color picker
		colorPicker = new MColorPicker() {
			@Override
			protected void colorPicked(final Color color) {
				ImageEditorCore core = (ImageEditorCore)Editor.getCurrentCore();
				if (core != null) {
					brushColor.setColor(color);
					core.getBrush().color.set(color);
					updateCursor(core, true);
				}
			}
		};
		brushColorPanel.add(colorPicker);
		add(brushColorPanel);
	}

	private void updateCursor(final ImageEditorCore core, final boolean updateCore) {
		if (updateCore)
			core.updateCursor();
		Brush brush = core.getBrush();
		brushColor.setIcon(new ImageIcon(brush.getPreview(brush.getSize(), false)));
	}

}
