// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.plugins;

import javax.swing.Icon;

import org.makagiga.common.*;

/**
 * A plugin info.
 * Each registered plugin has associated info.
 */
public final class PluginInfo implements Comparable<PluginInfo> {

	// public

	/**
	 * Is plugin enabled?
	 * Disabled plugins are not available for the user.
	 */
	public final BooleanProperty enabled = new BooleanProperty(true);

	/**
	 * A plugin associated with this info.
	 */
	public final Property<Pluggable> plugin = new Property<Pluggable>();

	/**
	 * A name of the Jar file that contains the plugin.
	 * The name is relative to the @ref path.
	 * @b NOTE: This property is not used in internal plugins.
	 */
	public final StringProperty classJar = new StringProperty();

	/**
	 * A plugin copyright text.
	 */
	public final StringProperty copyright = new StringProperty();

	/**
	 * A plugin single-line description.
	 */
	public final StringProperty description = new StringProperty();

	/**
	 * An unique plugin ID in the following format:
	 * @c your_domain_or_full_name/internal_plugin_name
	 */
	public final StringProperty id = new StringProperty();

	/**
	 * A short, human readable plugin name.
	 */
	public final StringProperty name = new StringProperty();

	/**
	 * A full path to the plugin directory.
	 * @b NOTE: This property is not used in internal plugins.
	 */
	public final StringProperty path = new StringProperty();

	/**
	 * A plugin type (based on the class type).
	 */
	public final StringProperty type = new StringProperty();

	/**
	 * A plugin version.
	 */
	public final StringProperty version = new StringProperty();

	// private

	private boolean internal;
	private Icon icon;

	// public

	/**
	 * Constructs a new plugin info.
	 * @param internal @c true - internal plugin; @c false - external plugin
	 */
	public PluginInfo(final boolean internal) {
		this.internal = internal;
	}

	/**
	 * Compares two plugin infos by its names.
	 * @see toString
	 */
	public int compareTo(final PluginInfo info) {
		return toString().compareTo(info.toString());
	}

	/**
	 * Returns an icon associated with this plugin.
	 * The default icon is <code>"misc"</code>.
	 */
	public Icon getIcon() {
		if (icon == null)
			icon = FS.getIcon("misc");

		return icon;
	}

	/**
	 * Sets icon to @p value.
	 */
	public void setIcon(final Icon value) { icon = value; }

	/**
	 * Returns @c true if plugin is internal.
	 */
	public boolean isInternal() { return internal; }

	/**
	 * Returns a localized plugin name.
	 */
	@Override
	public String toString() {
		return name.get();
	}

}
