// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.plugins;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Collection;
import java.util.TreeSet;

import org.makagiga.common.*;
import org.makagiga.console.ConsolePlugin;
import org.makagiga.editors.EditorPlugin;
import org.makagiga.internetsearch.InternetSearchPlugin;

/** The plugin manager. */
public final class PluginManager {

	// private

	private static PluginManager _instance;
	private Table table;

	// public

	/**
	 * Returns a list of plugins of the specified type.
	 * Disabled plugins are excluded.
	 * @param type A plugin type (e.g. EditorPlugin.TYPE)
	 */
	public static Collection<PluginInfo> getInfos(final String type) {
		return getInfos(type, false);
	}

// !!!: includeDisabled -> flags [0.9.3]
	/**
	 * Returns a list of plugins of the specified type.
	 * @param type A plugin type (e.g. EditorPlugin.TYPE)
	 * @param includeDisabled @c true - include disabled plugins
	 */
	public static Collection<PluginInfo> getInfos(final String type, final boolean includeDisabled) {
		Table infos = new Table();
		for (PluginInfo i : getInstance().table) {
			if (i.type.equals(type) && (i.enabled.get() || includeDisabled))
				infos.add(i);
		}

		return infos;
	}

	/**
	 * Returns the plugin manager.
	 */
	public static PluginManager getInstance() {
		if (_instance == null)
			_instance = new PluginManager();

		return _instance;
	}

	/**
	 * Registers an external plugin.
	 * @param path A plugin directory
	 * @return @c false if registration failed; otherwise @c true
	 */
	public boolean registerExternal(final String path) {
		return register(path, false);
	}

	/**
	 * Registers an internal plugin.
	 * @param path A plugin path (e.g. "org/makagiga/console/commands/screenshot")
	 * @return @c false if registration failed; otherwise @c true
	 */
	public boolean registerInternal(final String path) {
		return register(path, true);
	}

	/**
	 * Unregisters a plugin.
	 * @param info A plugin to unregister
	 */
	public void unregister(final PluginInfo info) {
		table.remove(info);
	}

	// private

	private PluginManager() {
		table = new Table();
// !!!: display Web Start info
// !!!: OS.isWebStart
		if (!"true".equals(System.getProperty("makagiga.webstart"))) {
			// global
// !!!: absolute path [0.9.3]
			findExternalPlugins("plugins");
			// local/user
			findExternalPlugins(FS.makeConfigPath("plugins"));
		}
	}

	private boolean deletePluginData(final PluginInfo info, final String path) {
		String absolutePath;
		try {
			File f = new File(path);
			absolutePath = f.getCanonicalPath();
			if (f.isDirectory())
				absolutePath += File.separator;
		}
		catch (IOException exception) {
			return false;
		}

		if (absolutePath.toUpperCase().startsWith(info.path.get().toUpperCase() + File.separator)) {
			return FS.delete(absolutePath);
		}
		else {
			MLogger.error("PLUGIN ERROR: Cannot remove file from the outside plugin directory: \"%s\"", absolutePath);

			return false;
		}
	}

	private void findExternalPlugins(final String dir) {
		File[] plugins = FS.list(dir);

		if (TK.isEmpty(plugins))
			return;

		for (File i : plugins) {
			if (i.isDirectory()) {
				MLogger.info("EXTERNAL PLUGIN: \"%s\"", i.getPath());
				registerExternal(i.getPath());
			}
		}
	}

	private boolean register(final String path, final boolean internal) {
		try {
			// init plugin config
			String separator = internal ? "/" : File.separator;
			String configPath = path + separator + "plugin.properties";
			int flags = internal ? Config.RESOURCE : 0;
			Config config = new Config(
				configPath,
				flags | Config.READ_ONLY | Config.USE_I18N
			);
			if (config.isError()) {
				MLogger.error("PLUGIN ERROR: Properties file not found: \"%s\"", configPath);

				return false;
			}
// !!!: validate all plugin configs [0.9.3]
			// check config
			String className = config.read("className");
			if (className == null) {
				MLogger.error("PLUGIN ERROR: No \"String.className\" specified: %s", configPath);

				return false;
			}
			String name = config.read("name");
			if (name == null) {
				MLogger.error("PLUGIN ERROR: No \"String.name\" specified: %s", configPath);

				return false;
			}

			// init plugin info
			PluginInfo info = new PluginInfo(internal);
			info.copyright.set(config.read("copyright"));
			info.description.set(config.read("description"));
			info.id.set(config.read("id"));
			info.name.set(name);
			info.version.set(config.read("version", AppInfo.niceVersion.get()));
			boolean defaultEnabled = config.read("default.enabled", true);
			info.enabled.set(Config.getGlobal().read("plugin.enabled." + info.id, defaultEnabled));

			String classJar;
			String classPath;
			if (internal) {
				classJar = null;
				classPath = toClassPath(FS.makePath(path, className));
			}
			else {
				classJar = config.read("classJar");
				if (classJar == null) {
					MLogger.error("PLUGIN ERROR: No \"String.classJar\" specified: %s", configPath);

					return false;
				}
				classPath = toClassPath(className);
				// used to remove plugin from disk
				info.classJar.set(classJar);
				info.path.set(path);
			}

			// remove plugin
			Config globalConfig = Config.getGlobal();
			if (!internal && globalConfig.read("plugin.remove." + info.id, false)) {
// !!!: String.extraFiles=extrajar1.jar extraimage.png [0.9.3]
// list of extra files to remove
				MLogger.info("PLUGINS: Removing plugin: \"%s\"", info);
				boolean error = false;

				if (!deletePluginData(info, FS.makePath(info.path.get(), info.classJar.get())))
					error = true;
				if (!deletePluginData(info, FS.makePath(info.path.get(), "icon.png")))
					error = true;
				if (!deletePluginData(info, FS.makePath(info.path.get(), "plugin.properties")))
					error = true;
				// "File.separator" is required by the "deletePluginData"
				if (!deletePluginData(info, info.path.get() + File.separator))
					error = true;

				if (error)
					MLogger.error("PLUGIN ERROR: Could not remove all plugin files");

				globalConfig.removeBoolean("plugin.remove." + info.id);
				globalConfig.saveToFile();

				return true;
			}

			// load plugin
			Class pluginClass;
			Pluggable pluginObject;
			try {
				if (internal) {
					pluginClass = getClass().getClassLoader().loadClass(classPath);
				}
				else {
					URL[] jarPath = { new URL("file:" + path + "/" + classJar) };
// FIXME: 2.0: access denied (java.io.FilePermission) in Java Web Start
					pluginClass = URLClassLoader.newInstance(
						jarPath,
						getClass().getClassLoader()
					).loadClass(classPath);
				}
				Object instance = pluginClass.newInstance();

				if (!(instance instanceof Pluggable))
					throw new Exception("Plugin \"" + classPath + "\" does not implement Pluggable interface");

				pluginObject = Pluggable.class.cast(instance);
			}
			catch (Exception exception) {
				MLogger.exception(exception);

				return false;
			}
			info.plugin.set(pluginObject);
			String type;
			if (pluginObject instanceof ConsolePlugin) {
				type = ConsolePlugin.TYPE;
			}
			else if (pluginObject instanceof EditorPlugin) {
				type = EditorPlugin.TYPE;
			}
			else if (pluginObject instanceof InternetSearchPlugin) {
				type = InternetSearchPlugin.TYPE;
			}
			else {
				type = null;
				MLogger.error("PLUGIN ERROR: Unknown plugin type: \"%s\"", info);
			}
			info.type.set(type);
// TODO: 2.0: load icon on demand
			if (internal) {
				info.setIcon(FS.getCustomIcon(pluginObject, "icon.png", FS.iconSize.get()));
			}
			else {
				info.setIcon(FS.getCustomIcon(FS.makePath(info.path.get(), "icon.png"), FS.iconSize.get()));
			}

			// add plugin info
			MLogger.info(
				"PLUGINS: Registering plugin \"%s\" (classJar=%s, className=%s, classPath=%s, type=%s)",
				info,
				classJar,
				className,
				classPath,
				type
			);
			pluginObject.onInit(config, info);
			table.add(info);

			return true;
		}
		catch (PluginException exception) {
			MLogger.error("PLUGIN ERROR: " + exception.getMessage());

			return false;
		}
		catch (Exception exception) {
			MLogger.exception(exception);

			return false;
		}
	}

	private String toClassPath(final String path) {
		return path.replace("/", ".").replace("\\", ".");
	}

	// private classes

	private static final class Table extends TreeSet<PluginInfo> { }

}
