// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.search;

import static java.awt.event.KeyEvent.*;

import static org.makagiga.common.UI._;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.text.Collator;
import java.util.Comparator;
import javax.swing.Icon;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.makagiga.MainWindow;
import org.makagiga.MetaInfo;
import org.makagiga.Sidebar;
import org.makagiga.Tabs;
import org.makagiga.common.*;
import org.makagiga.tree.Tree;
import org.makagiga.tree.TreeList;

/** The @b Search panel. */
public final class Search extends MPanel {

	// private

	private Collator collator;
	private Comparator<MetaInfo> nameComparator;
	private MButton startSearchButton;
	private MComboBox<Sort> sort;
	private MList<MetaInfo> list;
	private MTextField text;
	private Query query;
	private static Search _instance;

	// public

	/**
	 * Creates and returns an instance of the Search panel.
	 */
	public static Search getInstance() {
		if (_instance == null)
			_instance = new Search();

		return _instance;
	}

	public void search(final String text) {
		query = new Query(text);
		list.clear();
		final TreeList treeList = TreeList.getInstance();
		try {
			UI.setWaitCursor(MainWindow.getInstance(), true);
			new MTreeScanner<MetaInfo>(treeList.getDataRoot()) {
				public void processItem(final MTreeItem<MetaInfo> item) {
// TODO: 2.0: search for color
// TODO: 2.0: search for label
// TODO: 2.0: search for date
					MetaInfo metaInfo = item.getData();
					if (query.matches(metaInfo))
						list.addItem(metaInfo);
				}
			};
		}
		finally {
			UI.setWaitCursor(MainWindow.getInstance(), false);
		}
		if (list.isEmpty()) {
			MStatusBar.setWarning(_("No item found: \"{0}\"", text));
		}
		else {
			list.sort(getComparator());
			MStatusBar.setText(
				_("{0} item(s) found: \"{1}\"", list.getItemCount(), text),
				MStatusBar.AUTOHIDE
			);
		}
	}

	/**
	 * Sets focus to the search text field,
	 * and selects all text.
	 */
	public void setFocus() {
		text.selectAll();
	}

	// private

	private Search() {
		super(NO_LAYOUT);
		setVBoxLayout();
		// sort by name; ignore case
		collator = Collator.getInstance();
		collator.setStrength(Collator.PRIMARY);
		nameComparator = new Comparator<MetaInfo>() {
			public int compare(MetaInfo o1, MetaInfo o2) {
				return collator.compare(o1.toString(), o2.toString());
			}
		};
		add(initQueryPanel());
		addGap();
		add(initSortPanel());
		addGap();
		add(initListPanel());

		// read last sort method
		sort.setSelectedIndex(Config.getGlobal().readInt("search.sortMethod", Sort.DEFAULT.ordinal()));
		// write last sort method
		sort.addActionListener(new ActionListener() {
			public void actionPerformed(final ActionEvent e) {
				Config.getGlobal().write("search.sortMethod", sort.getSelectedIndex());
			}
		} );
	}

	private Comparator<MetaInfo> getComparator() {
		switch (sort.getSelectedItem()) {
			case AGE:
				return new Comparator<MetaInfo>() {
					public int compare(MetaInfo o1, MetaInfo o2) {
						int i = Long.valueOf(o1.getAge()).compareTo(o2.getAge());

						return (i == 0) ? nameComparator.compare(o1, o2) : (i * -1);
					}
				};

			case COLOR:
				return new Comparator<MetaInfo>() {
					public int compare(MetaInfo o1, MetaInfo o2) {
						Color c1 = o1.getColor();
						Color c2 = o2.getColor();
						int rgb1 = (c1 == null) ? 0x000000 : c1.getRGB();
						int rgb2 = (c2 == null) ? 0x000000 : c2.getRGB();
						int i = Integer.valueOf(rgb1).compareTo(rgb2);

						return (i == 0) ? nameComparator.compare(o1, o2) : i;
					}
				};

			case DEFAULT:
				return new Comparator<MetaInfo>() {
					public int compare(MetaInfo o1, MetaInfo o2) {
						int i = Float.valueOf(o1.getRatio()).compareTo(Float.valueOf(o2.getRatio()));

						return (i == 0) ? nameComparator.compare(o1, o2) : (i * -1);
					}
				};

			case LABEL:
				return new Comparator<MetaInfo>() {
					public int compare(MetaInfo o1, MetaInfo o2) {
						Icon i1 = o1.getIcon();
						Icon i2 = o2.getIcon();

						// getIcon may return null if file has no plugin
						if (i1 == null)
							return 1;

						if (i2 == null)
							return -1;

						int i = i1.toString().compareTo(i2.toString());

						return (i == 0) ? (nameComparator.compare(o1, o2) * -1) : i;
					}
				};

			case NAME:
				return nameComparator;

			case USAGE:
				return new Comparator<MetaInfo>() {
					public int compare(MetaInfo o1, MetaInfo o2) {
						int i = Integer.valueOf(o1.getUsage()).compareTo(o2.getUsage());

						return (i == 0) ? nameComparator.compare(o1, o2) : (i * -1);
					}
				};
		}

		return null;
	}

	private MPanel initListPanel() {
		MPanel p = MPanel.createBorderPanel(5);
		p.setMargin(5);
		MLabel listLabel = new MLabel(_("Search &Results:"));
		p.addNorth(listLabel);
		list = new MList<MetaInfo>();
		list.setCellRenderer(new SimpleListCellRenderer<MetaInfo>() {
			public void onRender(final MetaInfo value) {
				value.setupRenderer(this, -1, selected, false, hasFocus);
			}
		} );
		list.setSingleSelectionMode();
		list.addMouseListener(new MMouseAdapter() {
			@Override
			public void mouseClicked(final MouseEvent e) {
				if (isLeft(e)) {
					e.consume();
					open(list.getItemAt(e.getPoint()));
				}
			}
			@Override
			public void popupTrigger(final MouseEvent e) {
				MetaInfo metaInfo = list.getItemAt(e.getPoint());

				if (metaInfo == null)
					return;

				TreeList.getInstance().focusItem(metaInfo);
				list.setSelectedItem(metaInfo);
				Tree.getInstance().createTreeMenu(null, true).showPopup(list);
// !!!!: context menu
			}
		} );
		listLabel.setLabelFor(list);
// !!!: make it common and move to the MList
// !!!: common code for actions
		UI.connect(
			list, UI.WHEN_FOCUSED, VK_ENTER,
			new SimpleAction("Search.open") {
				public void onAction() {
					open(list.getSelectedItem());
				}
			}
		);
		UI.connect(
			list, UI.WHEN_FOCUSED, VK_SPACE,
			new SimpleAction("Search.open") {
				public void onAction() {
					open(list.getSelectedItem());
				}
			}
		);
		p.addCenter(list);
		list.addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public void mouseMoved(final MouseEvent e) {
				UI.setHandCursor(list, list.getItemAt(e.getPoint()) != null);
			}
		} );
		// update bold font in list cells
		Tabs.getInstance().addChangeListener(new ChangeListener() {
			public void stateChanged(final ChangeEvent e) {
				list.repaint();
			}
		} );

		return p;
	}

	private MPanel initQueryPanel() {
		MPanel p = MPanel.createHBoxPanel();
		MLabel label = new MLabel(_("&Search for:"));
		p.addGap();
		p.add(label);
		p.addGap();
		text = new MTextField();
		text.setEnterPressAction(new SimpleAction() {
			public void onAction() {
				startSearchButton.doClick();
			}
		} );
		text.setToolTipText(UI.makeHTML(
			_("Search for file name, text, comment, ...")
			+ "<br><br>" +
			_("Find all items containing \"foo\": <b>foo</b>") + "<br>" +
			_("Find all items containing \"keyword1\" or \"keyword2\": <b>keyword1 keyword2</b>")
		));
		label.setLabelFor(text);
		p.add(text);
		p.addGap();
		startSearchButton = new MButton(FS.getIcon("ui/ok", 16)) {
			@Override
			protected void onClick() {
				search(text.getText());
			}
		};
		startSearchButton.setToolTipText(UI.makeHTML(_("Start search.")));
		p.add(startSearchButton);

		return p;
	}

	private MPanel initSortPanel() {
		MPanel p = MPanel.createHBoxPanel();

		MLabel label = new MLabel(_("Sort &By:"));
		p.addGap();
		p.add(label);
		p.addGap();
		sort = new MComboBox<Sort>() {
			@Override
			protected void onSelect() {
				if (list != null)
					list.sort(getComparator());
			}
		};
		for (Sort i : Sort.values())
			sort.addItem(i);
		p.add(sort);
		label.setLabelFor(sort);

		return p;
	}

	private void open(final MetaInfo metaInfo) {
		if (metaInfo == null)
			return;

		if (metaInfo.isFolder()) {
			Sidebar.goTo(Sidebar.Tab.TREE);
			TreeList.getInstance().focusItem(metaInfo);

			return;
		}

		Tabs.getInstance().openEditor(metaInfo);
	}

}
