// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.settings;

import static org.makagiga.common.UI._;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.makagiga.common.*;
import org.makagiga.console.ConsolePlugin;
import org.makagiga.editors.EditorPlugin;
import org.makagiga.internetsearch.InternetSearchPlugin;
import org.makagiga.plugins.PluginInfo;
import org.makagiga.plugins.PluginList;
import org.makagiga.plugins.PluginManager;

/** The "Plugins" dialog. */
public final class PluginSettings extends Settings {

	// public

	/**
	 * Constructs a new "Plugins" dialog.
	 * @param parent A parent frame
	 */
	public PluginSettings(final MFrame parent) {
		super(parent, _("Plugins"), MODAL | OK_BUTTON);

		// editor
		addPage(new PluginSettingsPage(EditorPlugin.TYPE, _("Editors"), "editor"));

		// internetsearch
		InternetSearchPlugin.registerBuiltIn();
		addPage(new PluginSettingsPage(InternetSearchPlugin.TYPE, _("Internet Search"), "find"));

		// console
		ConsolePlugin.registerBuiltIn();
		addPage(new PluginSettingsPage(ConsolePlugin.TYPE, _("Console Commands"), "console"));

		exec();
	}

	// private classes

	private final class PluginSettingsPage extends SettingsPage implements SettingsAction {

		// private

		private MButton enableDisableButton;
		private MButton removeButton;
		private PluginList list;

		// public

		public PluginSettingsPage(final String type, final String text, final String iconName) {
			// list
			this.iconName.set(iconName);
			list = new PluginList(
				type,
				PluginList.INCLUDE_DISABLED | PluginList.SHOW_COPYRIGHT | PluginList.SHOW_VERSION
			);
			list.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
				public void valueChanged(final ListSelectionEvent e) {
					PluginInfo info = list.getSelectedItem();
					if (info == null) {
						enableDisableButton.setEnabled(false);
						removeButton.setEnabled(false);
					}
					else {
						enableDisableButton.setEnabled(true);
						removeButton.setEnabled(!info.isInternal());
					}
					updateButtonText();
				}
			} );
			addCenter(list);
			this.text.set(text + " [" + list.getItemCount() + "]");

			// buttons
			MPanel buttonsPanel = MPanel.createHBoxPanel();
			enableDisableButton = new MButton(_("&Disable")) {
				@Override
				protected void onClick() {
					togglePlugin();
				}
			};
			enableDisableButton.setEnabled(false);
			buttonsPanel.add(enableDisableButton);
			buttonsPanel.addStretch();
			removeButton = new MButton(_("&Remove")) {
				@Override
				protected void onClick() {
					removePlugin();
				}
			};
			removeButton.setEnabled(false);
			buttonsPanel.add(removeButton);
			addSouth(buttonsPanel);
		}

		// SettingsAction

		public void onInit() { }

		public void onOK() {
// !!!: update when dialog closed [0.9.3]
			Config config = Config.getGlobal();
			for (PluginInfo i : list)
				config.write("plugin.enabled." + i.id, i.enabled);
			needRestart.yes();
		}

		// private

// !!!: add info that application restart is required
		private void removePlugin() {
			PluginInfo info = list.getSelectedItem();

			if (info == null)
				return;

			if (info.isInternal())
				return;

			if (!MMessage.confirm(
				PluginSettings.this,
				_("Remove plugin?\n\n\"{0}\""),
				info
			))
				return;

			// Plugin will be removed by the Plugin Manager
			// on next application startup.
			Config.getGlobal().write("plugin.remove." + info.id, true);
			Config.getGlobal().saveToFile();

			list.removeItem(info);
			PluginManager.getInstance().unregister(info);
			needRestart.yes();
		}

		private void togglePlugin() {
			PluginInfo info = list.getSelectedItem();

			if (info == null)
				return;

			info.enabled.toggle();
			updateButtonText();
			list.updateUI();
		}

		private void updateButtonText() {
			PluginInfo info = list.getSelectedItem();

			if (info == null)
				return;

			enableDisableButton.setText(info.enabled.get() ? _("&Disable") : _("&Enable"));
		}

	}

}
