// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.tools;

import static org.makagiga.common.UI._;

import java.io.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.makagiga.MainWindow;
import org.makagiga.Tabs;
import org.makagiga.common.*;

// TODO: 2.0: remember last backup time
// !!!: plugin
/** A simple backup utility. */
public final class Backup extends MDialog
	implements
		MDialogAction,
		Runnable
{

	// private

	private boolean cancelled;
	private int totalFiles;
	private MLabel status;
	private MProgressBar progress;
	private String defaultName;
	private ZipOutputStream output;

	// public

	/**
	 * Constructs a new "Backup" dialog.
	 * @param parent A parent frame
	 */
	public Backup(final MFrame parent) {
		super(parent, _("Backup"), STANDARD_DIALOG);
		getCancelButton().setEnabled(false);
		getOKButton().setEnabled(false);

		MPanel p = MPanel.createVBoxPanel();
		p.setMargin(5);
		status = new MLabel(_("Waiting for user..."));
		p.add(status);
		p.addGap();
		progress = new MProgressBar();
		progress.setStringPainted(true);
		progress.setString("");
		p.add(progress);
		addCenter(p);

		defaultName = String.format(
			"%s-backup-%s",
			AppInfo.internalName,
			DT.format(DT.now(), "yyyyMMdd-HHmm")
		);

		Thread thread = new Thread(this);
		thread.start();

		pack();
		setSize(640, getHeight());
		setResizable(false);
		exec();
	}

	/**
	 * The main thread.
	 */
	public void run() {
		String archivePath = getArchivePath();
		if (archivePath == null) {
			done(Code.REJECTED);

			return;
		}

		status.setText(_("Saving session..."));
		MainWindow.getInstance().saveSession(Tabs.NORMAL_SAVE);

		// count all files
		status.setText(_("Please wait..."));
		new FileScanner<Object>(FS.getConfigDir()) {
			public Object processDir(final Object parentItem, final File dir) {
				return null;
			}
			public void processFile(final Object parentItem, final File file) {
				totalFiles++;
			};
		};
		// setup progress bar
		progress.setMinimum(0);
		progress.setMaximum(totalFiles);

		getCancelButton().setEnabled(true);

		status.setText(_("Creating backup..."));
		try {
			MLogger.info("Creating archive: \"%s\"", archivePath);
			BufferedOutputStream outputStream = new BufferedOutputStream(new FileOutputStream(archivePath));
			output = new ZipOutputStream(outputStream);

			// create zip archive
			new FileScanner<Object>(FS.getConfigDir()) {
				public boolean isStopped() { return super.isStopped() || cancelled; };
				public Object processDir(final Object parentItem, final File dir) {
					return null;
				}
				public void processFile(final Object parentItem, final File file) {
					try {
						addFile(
							filePath.substring(FS.getConfigDir().length() + 1),
							filePath
						);
					}
					catch (IOException exception) {
						MMessage.fileError(
							Backup.this,
							_("Input/output error"),
							filePath
						);
						MLogger.exception(exception);
						stop();
						done(Code.REJECTED);
					}
					progress.step();
				}
			};
		}
		catch (FileNotFoundException exception) {
			MMessage.fileError(
				this,
				_("Could not save file"),
				archivePath
			);
			MLogger.exception(exception);
			done(Code.REJECTED);
		}
		catch (IOException exception) {
			MMessage.fileError(
				this,
				_("Input/output error"),
				archivePath
			);
			MLogger.exception(exception);
			done(Code.REJECTED);
		}
		finally {
			FS.close(output);
			if (cancelled) {
				MMessage.info(this, _("Backup cancelled!"));
				getCancelButton().setEnabled(false);
				getOKButton().setEnabled(true);
				// delete incomplete archive
				FS.delete(archivePath);
			}
			else {
				status.setText(_("Backup complete."));
				getCancelButton().setEnabled(false);
				getOKButton().setEnabled(true);
			}
		}
	}

	// MDialogAction

	public void accept() { }

	public void reject() {
		cancelled = true;
	}

	// private

	private void addFile(final String entryName, final String path) throws IOException {
		if (cancelled)
			return;

		print(entryName);
		byte[] buf = new byte[4096];
		int bufLen;
// FIXME: 2.0: local characters in zip entry names (JAVA BUG #4244499)
		ZipEntry zipEntry = null;
		BufferedInputStream input = null;
		try {
			zipEntry = new ZipEntry(defaultName + "/" + entryName);
			output.putNextEntry(zipEntry);
			input = new BufferedInputStream(new FileInputStream(path));

			// read and write
			while ((bufLen = input.read(buf)) != -1)
				output.write(buf, 0, bufLen);
		}
		finally {
			FS.close(input);
			if (zipEntry != null)
				output.closeEntry();
		}
	}

	/**
	 * Returns a new archive path (.zip file)
	 * selected from the file chooser dialog.
	 */
	private String getArchivePath() {
		MFileChooser fileChooser = new MFileChooser(this, _("Enter a new Zip file name..."));
		fileChooser.addFilter(".zip", _("Zip Archive"));
		fileChooser.setConfigKey("backup");
		fileChooser.setSelectedPath(defaultName + ".zip");
		// show dialog
		if (fileChooser.saveDialog()) {
// !!!: warn if user is trying to create backup in the ".makagiga" directory
			String path = fileChooser.getSelectedPath();

			// selected file exists?
			if (FS.exists(path) && !MMessage.confirmFileOverwrite(this, path))
				return null;

			return path;
		}

		return null;
	}

	private void print(final String text) {
		MLogger.info("BACKUP: " + text);
		progress.setString(text);
	}

}
