// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.tools;

import static java.awt.event.KeyEvent.*;

import static org.makagiga.common.UI._;

import java.awt.Dimension;
import java.awt.event.FocusListener;

import org.makagiga.common.*;
import org.makagiga.editors.Editor;
import org.makagiga.editors.EditorSearch;

/** Editor quick find panel. */
public final class EditorQuickFind extends MToolBar {

	// private

	private boolean reset = true;
	private static EditorQuickFind _instance;
	private MButton findNextButton;
	private MTextField text;
	private String lastSearch = "";

	// public

	public void findNext() {
		if (reset)
			resetSearch();
		startSearch();
	}

	/**
	 * Creates and returns the instance of the Editor quick find panel.
	 */
	public static EditorQuickFind getInstance() {
		if (_instance == null)
			_instance = new EditorQuickFind();

		return _instance;
	}

	public void setActive(final boolean active) {
		setVisible(active);
		if (active) {
			text.setDefault();
		}
		else {
			cancelSearch();
		}
	}

	// private

	private EditorQuickFind() {
		setVisible(false);

		CloseAction closeAction = new CloseAction();

		add(closeAction);
		addGap();

		text = new MTextField();
		// enter - start search
		text.setEnterPressAction(new SimpleAction() {
			public void onAction() {
				findNextButton.doClick();
			}
		} );
		MLabel textLabel = new MLabel(_("Find:"));
		textLabel.setLabelFor(text);
		add(textLabel);
		addGap();
		add(text);
		addGap();
		findNextButton = new MButton(FS.getIcon("ui/ok", 16)) {
			@Override
			protected void onClick() {
				startSearch();
			}
		};
		findNextButton.setToolTipText(UI.makeHTML(_("Find next.")));
		add(findNextButton);
		// esc - cancel search
		UI.connect(text, UI.WHEN_IN_FOCUSED_WINDOW, VK_ESCAPE, closeAction);
	}

	private void cancelSearch() {
		MStatusBar.ready();
		Editor editor = getEditor();
		if (editor != null)
			editor.setFocus();
	}

	private Editor getEditor() {
		Editor editor = Editor.currentEditor.get();

		return (editor instanceof EditorSearch) ? editor : null;
	}

	private void resetSearch() {
		reset = true;
		MStatusBar.ready();
		Editor editor = getEditor();
		if (editor == null) {
			text.setText(lastSearch);
		}
		else {
			String s = EditorSearch.class.cast(editor).getDefaultSearchText();
			text.setText((s == null) ? lastSearch : s);
		}
	}

	private void startSearch() {
		MStatusBar.ready();
		Editor editor = getEditor();

		if (editor == null)
			return;
		
		// nothing to search?
		if (text.isEmpty())
			return;

		String textToFind = text.getText();
		lastSearch = textToFind;

		EditorSearch.Result result;
		if (reset) {
			reset = false;
			result = EditorSearch.class.cast(editor).findText(textToFind);
		}
		else {
			result = EditorSearch.class.cast(editor).findNextText(textToFind);
		}
		switch (result) {
			case EMPTY:
				MStatusBar.setWarning(_("The editor is empty"));
				break;
			case FOUND:
				MStatusBar.ready();
				break;
			case NOT_FOUND:
				MStatusBar.setWarning(_("Text not found"));
				break;
		}
	}

	// private classes

	private final class CloseAction extends AppAction {

		// public

		public CloseAction() {
			super(_("Close"), "ui/close");
		}

		public void onAction() {
			setActive(false);
		}

	}

}
