// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.tree;

import static org.makagiga.common.UI._;

import java.io.File;
import java.io.IOException;

import org.makagiga.MetaInfo;
import org.makagiga.Tabs;
import org.makagiga.common.*;
import org.makagiga.editors.EditorPlugin;

public final class TreeFS {

	// public

	public static final int CONFIRM = 1 << 0; /**< Confirm operation. */
	public static final int UPDATE_SELECTION = 1 << 1; /**< Update selection after operation. */

	// private

	private static TreeListItem fileItem;

	// public

	public static void copyFileToFolder(final TreeListItem source, final TreeListItem destination) throws IOException {
		copyFileToFolder(false, source, destination);
	}

	public static void delete(final TreeListItem item) {
		delete(item, 0);
	}

	public static void delete(final TreeListItem item, final int flags) {
		if (item == null)
			return;

		TreeListItem parent = (TreeListItem)item.getParentItem();

		MetaInfo metaInfo = item.getData();
		String path = metaInfo.getPath();

		if (metaInfo.isDataRoot())
			return;

		// confirm delete
		if (((flags & CONFIRM) != 0) && !MMessage.confirmWithIcon(
			null,
			"ui/delete",
			UI.makeHTML(_("Delete file?") + "<h3>" + metaInfo.getNicePath() + "</h3>")
		))
			return;

		final TreeList treeList = TreeList.getInstance();

		// delete file
		if (metaInfo.isFile()) {

			// close editor (false - don't save)
			final Tabs tabs = Tabs.getInstance();
			tabs.closeEditorAt(tabs.findEditor(metaInfo), Tabs.REMOVE_TAB);

			// delete file
			if (!FS.delete(path)) {
				MMessage.fileError(
					null,
					_("Could not delete file"),
					path
				);

				return;
			}

			// delete item
			treeList.deleteItem(item);

			// delete meta info
			FS.delete(path + ".properties");
		}
		// delete folder
		else if (metaInfo.isFolder()) {
			// delete meta info
			FS.delete(FS.makePath(path, ".properties")); // file may not exists
			// delete empty folder
			if (!FS.delete(path)) {
				MMessage.fileError(
					null,
					_("Could not delete folder"),
					path
				);

				return;
			}
			// delete item
			treeList.deleteItem(item);
		}
		if ((flags & UPDATE_SELECTION) != 0) {
			if (treeList.getSelectedItem() == null)
				treeList.selectItem(parent);
		}
	}

	public static boolean isFileNameValid(final String name) {
		if (TK.isEmpty(name)) {
			MMessage.error(null, _("Name cannot be empty"));

			return false;
		}

		if (name.length() > 128) {
			MMessage.error(null, _("Name length cannot be greater than 128 characters"));

			return false;
		}

		if (name.startsWith(".")) {
			MMessage.error(null, _("Name cannot start with a dot character"));

			return false;
		}

		// for compatibility with shell
		final String[] INVALID = { "*" };
		for (String i : INVALID) {
			if (name.contains(i)) {
				MMessage.error(
					null,
					_("Name \"{0}\"\ncontains an illegal character \"{1}\"\n\nIllegal characters are: {2}"),
					name,
					i,
					TK.toString(INVALID)
				);

				return false;
			}
		}

		return true;
	}

	public static void moveFileToFolder(final TreeListItem source, final TreeListItem destination) throws IOException {
		copyFileToFolder(true, source, destination);
	}

	// private

	private static void copyFileToFolder(final boolean move, final TreeListItem source, final TreeListItem destination) throws IOException {
		MetaInfo sourceMetaInfo = source.getData();
		MetaInfo destinationMetaInfo = destination.getData();

		String sourcePath = sourceMetaInfo.getPath();
		String destinationPath = destinationMetaInfo.getPath();
		
		MLogger.XXX("sourcePath=%s, destinationPath=%s", sourcePath, destinationPath);

		File sourceFile = new File(sourcePath);
		File destinationFolder = new File(destinationPath);

		// check paths
		if (sourceMetaInfo.isFolder() && sourceFile.equals(destinationFolder))
			throw new IOException(_("Source and destination directories are identical"));

		final File destinationFile = new File(FS.makePath(destinationPath, sourceFile.getName()));

		MLogger.info(
			"%s file \"%s\" to file \"%s\"...",
			move ? "Moving" : "Copyging",
			sourceFile,
			destinationFile
		);

		// check paths
		if (sourceFile.equals(destinationFile))
			throw new IOException(_("Source and destination directories are identical"));

		// confirm overwrite
		boolean overwrite = destinationFile.exists();
// !!!: file names in "confirmFileOverwrite" are confusing
		if (overwrite && !MMessage.confirmFileOverwrite(null, destinationFile.getPath()))
			return;

		// save source file
		sourceMetaInfo.saveToFile();
		final Tabs tabs = Tabs.getInstance();
		tabs.saveEditor(tabs.getTabAt(tabs.findEditor(sourceMetaInfo)), Tabs.NORMAL_SAVE);
		// copy source file
//!!!folder dragging [0.9.2]
		if (FS.copyFile(sourceFile.getPath(), destinationFile.getPath())) {
			// copy properties file
			String propertiesFile = sourceFile.getPath() + ".properties";
			if (FS.exists(propertiesFile))
				FS.copyFile(propertiesFile, destinationFile.getPath() + ".properties");
			final TreeList treeList = TreeList.getInstance();
			if (overwrite) {
				tabs.closeEditorAt(tabs.findEditor(destinationFile.getPath()), Tabs.REMOVE_TAB);
			//!!! [0.9.2]
				new MTreeScanner<MetaInfo>(destination) {
					public void processItem(final MTreeItem<MetaInfo> item) {
						if (item.getData().getPath().equals(destinationFile.getPath())) {
							fileItem = (TreeListItem)item;
							stop();
						}
					}
				};
				//!!!cleanup and test [0.9.2]
				String path = destinationFile.getPath();
				MetaInfo metaInfo = new MetaInfo(path, MetaInfo.FILE);
				fileItem.setData(metaInfo);
// !!!: common code with "TreeListItem.createFileItem" [0.9.2]
				if (!metaInfo.isCustomIcon())
					metaInfo.setIcon(EditorPlugin.findIconForType(metaInfo.getType()));
			}
			else {
				String path = destinationFile.getPath();
				fileItem = TreeListItem.createFileItem(
					path,
					EditorPlugin.findPluginForType(FS.getType(path))
				);
				treeList.addItem(destination, fileItem);
			}

			// delete source file
// !!!: do not close editor or reopen after move [0.9.2]
			if (move)
				delete(source);

			treeList.focusItem(fileItem);
		}
	}

}
