// Copyright 2005 Konrad Twardowski
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.makagiga.tree;

import static java.awt.event.KeyEvent.*;

import static org.makagiga.common.UI._;

import java.awt.Point;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.io.File;
import java.io.IOException;
import java.util.TooManyListenersException;
import javax.swing.TransferHandler;
import javax.swing.tree.TreeSelectionModel;

import org.makagiga.MetaInfo;
import org.makagiga.Tabs;
import org.makagiga.common.*;
import org.makagiga.editors.EditorPlugin;
import org.makagiga.plugins.PluginInfo;

/** The tree list. */
public final class TreeList extends MTree<TreeListItem> implements DropTargetListener {

	// private

	private enum DnDInfo { COPY, MOVE, UNKNOWN };
	private static TreeList _instance;
	private TreeListItem dataRoot;

	// public

	/**
	 * Focuses item with @p metaInfo.
	 */
	public void focusItem(final MetaInfo metaInfo) {
		if (metaInfo == null)
			return;

		final String PATH = metaInfo.getPath();
		new MTreeScanner<MetaInfo>(getDataRoot()) {
			public void processItem(final MTreeItem<MetaInfo> item) {
				if (item.getData().getPath().equals(PATH)) {
					focusItem((TreeListItem)item);
					stop();
				}
			}
		};
	}

	public TreeListItem getDataRoot() { return dataRoot; }

	/**
	 * Constructs and returns a new instance of the tree list.
	 */
	public static TreeList getInstance() {
		if (_instance == null)
			_instance = new TreeList();

		return _instance;
	}

	/**
	 * Returns a meta info at row @p index, or @c null if meta info not found.
	 */
	public MetaInfo getMetaInfoAtRow(final int index) {
		TreeListItem item = getItemAtRow(index);

		if (item == null)
			return null;

		return item.getData();
	}

	/**
	 * Returns a selected meta info, or @c null if no selection.
	 */
	public MetaInfo getSelectedMetaInfo() {
		TreeListItem item = getSelectedItem();

		return (item == null) ? null : item.getData();
	}

	/**
	 * Restores the state of all items.
	 */
	public void restoreAll() {
		new MTreeScanner<MetaInfo>(getDataRoot()) {
			public void processItem(final MTreeItem<MetaInfo> item) {
				// reopen
				MetaInfo metaInfo = item.getData();
				if (metaInfo.isFile() && metaInfo.open.get())
					Tabs.getInstance().openEditor(metaInfo);
				// expand
				if (metaInfo.isFolder() && metaInfo.expanded.get())
					expandItem((TreeListItem)item);
			}
		};
		// select something
		TreeListItem item = getSelectedItem();
		if (item == null)
			selectItem(getDataRoot());
	}

	/**
	 * Saves the state of all items.
	 */
	public void saveAllItems() {
		new MTreeScanner<MetaInfo>(getDataRoot()) {
			public void processItem(final MTreeItem<MetaInfo> item) {
				MetaInfo metaInfo = item.getData();
				metaInfo.expanded.set(isExpanded(item.getTreePath()));
				metaInfo.saveToFile();
			}
		};
	}

	// drop target

	public void dragEnter(final DropTargetDragEvent e) {
		updateDnDInfo(e.getDropAction());
	}

	public void dragExit(final DropTargetEvent e) {
		updateDnDInfo(DnDInfo.UNKNOWN);
		
		MLogger.XXX("dragExit");
		//!!!MScrollPane.autoScroll(MTree.this, e.getX(), e.getY());
	}

	public void dragOver(final DropTargetDragEvent e) {
		TreeListItem item = getItemAtDropLocation(e.getLocation());

		if (item == null) {
			e.rejectDrag();

			return;
		}

		if (TreeListTransferHandler.draggedItem.isNull()) {
			e.rejectDrag();

			return;
		}

		MetaInfo draggedMetaInfo = TreeListTransferHandler.draggedItem.get().getData();
		if (draggedMetaInfo.isFolder() || draggedMetaInfo.isDataRoot()) {
			e.rejectDrag();

			return;
		}

		MetaInfo dropMetaInfo = item.getData();
		String draggedDir = draggedMetaInfo.getPath();
		draggedDir = draggedDir.substring(0, draggedDir.lastIndexOf(File.separator));
		String dropDir = dropMetaInfo.getPath();
		if (!dropMetaInfo.isFolder())
			dropDir = dropDir.substring(0, dropDir.lastIndexOf(File.separator));
		//MLogger.XXX("draggedDir: " + draggedDir);
		//MLogger.XXX("dropDir: " + dropDir);
		if (dropDir.equals(draggedDir)) {
			e.rejectDrag();

			return;
		}

		e.acceptDrag(e.getDropAction());

		focusItem(item);
		// open folder under mouse
		if ((dropMetaInfo.isFolder() || dropMetaInfo.isDataRoot())) {
			if (!isExpanded(item.getTreePath())) {
				// delay
				try {
					Thread.currentThread().sleep(500);
				}
				catch (InterruptedException exception) { }
				// expand
				Point p = getMousePosition();
				if (item == getItemAtDropLocation(p))
					expandItem(item);
			}
		}
	}

// TODO: 2.0: add support for import/export via drag and drop
	public void drop(final DropTargetDropEvent e) {
		updateDnDInfo(DnDInfo.UNKNOWN);

		TreeListItem destination = getItemAtDropLocation(e.getLocation());
		TreeListItem source = TreeListTransferHandler.draggedItem.get();
		TreeListTransferHandler.draggedItem.clear();

		if (destination == null)
			return;

		MetaInfo destinationMetaInfo = destination.getData();

		switch (e.getDropAction()) {
			case TransferHandler.COPY:
				try {
					TreeFS.copyFileToFolder(
						source,
						(destinationMetaInfo.isFolder() || destinationMetaInfo.isDataRoot())
							? destination
							: (TreeListItem)destination.getParentItem()
					);
				}
				catch (IOException exception) {
					MMessage.error(null, exception.getLocalizedMessage());
				}
				break;
			case TransferHandler.MOVE:
				try {
					TreeFS.moveFileToFolder(
						source,
						(destinationMetaInfo.isFolder() || destinationMetaInfo.isDataRoot())
							? destination
							: (TreeListItem)destination.getParentItem()
					);
				}
				catch (IOException exception) {
					MMessage.error(null, exception.getLocalizedMessage());
				}
				break;
		}
	}

	public void dropActionChanged(final DropTargetDragEvent e) {
		updateDnDInfo(e.getDropAction());
	}

	// private

	private TreeList() {
		// init main root
		TreeListItem mainRootItem = TreeListItem.createItem(
			null,
			MetaInfo.MAIN_ROOT
		);
		setRootItem(mainRootItem);
		setRootVisible(false);

		// init data
		String dataDir = FS.makeConfigPath(FS.makePath("vfs", "data"), FS.CREATE_DIR);
		if (!FS.exists(dataDir)) {
			MMessage.fileError(
				null,
				_("Could not initialize data directory"),
				dataDir
			);
			TK.quit(1);
		}
		MLogger.info("Data directory: \"%s\"", dataDir);

		dataRoot = TreeListItem.createItem(
			dataDir,
			MetaInfo.DATA_ROOT
		);
// TODO: 2.0: vfs
		addItem(mainRootItem, dataRoot);
// !!!: setEditable(true); [0.9.3]
		// build tree
		new FileScanner<TreeListItem>(getDataRoot(), dataDir) {
			public TreeListItem processDir(final TreeListItem parentItem, final File dir) {
				TreeListItem folderItem = TreeListItem.createFolderItem(filePath);
				parentItem.add(folderItem);

				return folderItem;
			}
			public void processFile(final TreeListItem parentItem, final File file) {
				String type = FS.getType(fileName);

				// skip ".properties"
				if (type == null)
					return;

				PluginInfo info = EditorPlugin.findPluginForType(type);
				if (info != null) {
					parentItem.add(TreeListItem.createFileItem(
						filePath,
						info
					));
				}
				else {
					if (!type.equals(".properties"))
						parentItem.add(TreeListItem.createItem(
							filePath,
							MetaInfo.FILE | MetaInfo.NOT_SUPPORTED
						));
				}
			}
		};
/* TODO: 2.0: trash
		String trashDir = FS.makeConfigPath(FS.makePath("vfs", "trash"), FS.CREATE_DIR);
		TreeListItem trashRoot = TreeListItem.createItem(
			trashDir,
			MetaInfo.TRASH_ROOT
		);
		addItem(mainRootItem, trashRoot);
*/
		setCellRenderer(new TreeListCellRenderer());
		setExpandedState(getDataRoot().getTreePath(), true);
		getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
// TODO: 2.0: multiselection

		initDragAndDrop();

		// add mouse event handlers
		addMouseListener(new MMouseAdapter() {
			@Override
			public void mouseClicked(final MouseEvent e) {
				if (
					isLeft(e) &&
					// ctrl, shift - selection mode
					!e.isControlDown() && !e.isShiftDown()
				) {
					e.consume();
					openFileAtRow(getRowForLocation(e.getX(), e.getY()));
				}
			}
			@Override
			public void popupTrigger(final MouseEvent e) {
				popupContextMenu(e.getX(), e.getY());
			}
		} );
		addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public void mouseDragged(final MouseEvent e) {
				if (MMouseAdapter.isLeft(e)) {
					e.consume();
					TreeListTransferHandler.draggedItem.set(getSelectedItem());

					if (
						TreeListTransferHandler.draggedItem.isNull() ||
						TreeListTransferHandler.draggedItem.get().getData().isDataRoot()
					)
						return;

					getTransferHandler().exportAsDrag(TreeList.this, e, TransferHandler.MOVE);
				}
			}
			@Override
			public void mouseMoved(final MouseEvent e) {
				UI.setHandCursor(TreeList.this, getRowForLocation(e.getX(), e.getY()) != -1);
			}
		} );

		// add key event handler
		addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(final KeyEvent e) {
				if (
					((e.getKeyCode() == VK_ENTER) || (e.getKeyCode() == VK_SPACE)) &&
					(e.getModifiers() == 0)
				) {
					e.consume();
					openFileAtRow(getLeadSelectionRow());
				}
				else if (e.getKeyCode() == VK_F2) {
					e.consume();
					UI.fire(Tree.getInstance().getRenameAction(), TreeList.this);
				}
			}
		} );

/* TODO: 2.0: expand/collapse listener
		// add expand/collapse event handler
		addTreeWillExpandListener(new TreeWillExpandListener() {
			public void treeWillCollapse(final TreeExpansionEvent e) {
				MLogger.info("treeWillCollapse: " + e.getPath());
			}
			public void treeWillExpand(final TreeExpansionEvent e) {
				MLogger.info("treeWillExpand: " + e.getPath());
			}
		} );
*/
	}

	private TreeListItem getItemAtDropLocation(final Point location) {
		if (location == null)
			return null;
		
		int row = getRowForLocation(location.x, location.y);

		if (row == -1)
			return null;

		TreeListItem item = getItemAtRow(row);

		if (item == null)
			return null;

		return item;
	}

	private void initDragAndDrop() {
// !!!: tree drag and drop autoscroll [0.9.2]
		setAutoscrolls(true);
		setDragEnabled(true);
		DropTarget dropTarget = new DropTarget();
		try {
			dropTarget.addDropTargetListener(this);
			setDropTarget(dropTarget);
		}
		catch (TooManyListenersException exception) {
			MLogger.exception(exception);
		}
		setTransferHandler(new TreeListTransferHandler());
	}

	private void openFileAtRow(final int index) {
		if (index == -1)
			return;

		TreeListItem item = getItemAtRow(index);
		if (item.getData().isFile()) {
			Tabs.getInstance().openEditor(item.getData());
		}
		else {
			if (isExpanded(index))
				collapseRow(index);
			else
				expandRow(index);
		}
	}

	private void popupContextMenu(final int x, final int y) {
		// select row under mouse
		setSelectionRow(getRowForLocation(x, y));
		// show popup menu
		Tree.getInstance().createTreeMenu(null, false).showPopup(this, x, y);

	}

	private void updateDnDInfo(final DnDInfo info) {
		switch (info) {
			case COPY:
				MStatusBar.setText(_("Copy"));
				break;
			case MOVE:
				MStatusBar.setText(_("Move (press Ctrl key to copy)"));
				break;
			case UNKNOWN:
				MStatusBar.ready();
				break;
			default:
				MStatusBar.ready();
		}
	}

	private void updateDnDInfo(final int info) {
		switch (info) {
			case TransferHandler.COPY:
				updateDnDInfo(DnDInfo.COPY);
				break;
			case TransferHandler.MOVE:
				updateDnDInfo(DnDInfo.MOVE);
				break;
			default:
				updateDnDInfo(DnDInfo.UNKNOWN);
		}
	}

}
